/*
 * mdh (MailDooHicky)
 *
 * Copyright (c) 2003 Mike Hokenson <mdh at gozer dot org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifdef __sun__
#include <kstat.h>
#endif

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "confuse.h"

#include "mdh.h"
#include "util.h"

#include "mdh_gen.h"
#include "mdh_pad.h"
#include "mdh_run.h"

#include "mdh_toolbar.h" 
#include "mdh_panel_time.h"
#include "mdh_panel_mail.h"
#include "mdh_panel_cpu.h"
#include "mdh_panel_mem.h"
#include "mdh_panel_disk.h"  
#include "mdh_panel_net.h"
#include "mdh_panel_temp.h"  
#include "mdh_panel_uptime.h"

#ifdef __sun__
kstat_ctl_t *kc;
#endif

/* tooltips for everything (config window) except the main panel */
GtkTooltips *tool_tips;

static gchar *mdh_rc, *mdh_lock, *mdh_hist;
gchar *mdh_cache;

static void mdh_quit(void);

static void mdh_toolbar_init(void);

static void mdh_toolbar_setup(void);

static void mdh_toolbar_config_load(void);
static void mdh_toolbar_config_save(void);

static void mdh_toolbar_history_load(void);
static void mdh_toolbar_history_save(void);

static void mdh_toolbar_lock(void);

static void mdh_toolbar_create(void);
static void mdh_toolbar_show(void);

static void mdh_init(gint argc, gchar **argv);

static void mdh_toolbar_set_winops(void);

RETSIGTYPE mdh_signal_quit()
{
    exit(1);
}

RETSIGTYPE mdh_signal_reconfigure()
{
    /* re-read config */
    mdh_toolbar_config_load();

    /* show/hide panel widgets, reset timers, etc */
    mdh_toolbar_reset();

    set_signal(SIGUSR1, mdh_signal_reconfigure);
}

gint main(gint argc, gchar **argv)
{
    mdh_init(argc, argv);             /* cli opts, sig handlers, paths */

    mdh_toolbar_lock();               /* create <rc>.lock */

    mdh_toolbar_init();               /* init vars, tooltip widgets */
    mdh_toolbar_setup();              /* setup config related variables */

    mdh_toolbar_config_load();        /* load configuration */
    mdh_toolbar_history_load();       /* load run history */

    mdh_toolbar_create();             /* create widgets, add panels */
    mdh_toolbar_reset();              /* show widgets, start timers, etc */
    mdh_toolbar_show();               /* show the main toolbar window */

    g_atexit(mdh_quit);

    gtk_main();
    
    exit(0);
}

static void mdh_toolbar_custom_free(struct custom_t *obj)
{
    g_return_if_fail(obj != NULL);

    g_free(obj->value);
    g_free(obj->command);
    g_free(obj->tooltip);
}

static void mdh_toolbar_panel_free(struct panel_t *obj)
{
    g_return_if_fail(obj != NULL);

    g_free(obj->value);
    g_free(obj->command);
    g_free(obj->tooltip);

    g_free(obj->data);
}

static void mdh_quit(void)
{
    g_slist_foreach(Toolbar.panels, (GFunc) mdh_toolbar_panel_stop, NULL);

    mdh_toolbar_config_save();

    mdh_toolbar_history_save();

    if(IS_LINK(mdh_lock) == TRUE)
        mdh_unlink(mdh_lock);

    g_free(Toolbar.run_item);
    g_free(Toolbar.font);

    g_slist_free(Toolbar.buttons);

    g_slist_foreach(Toolbar.custom, (GFunc) mdh_toolbar_custom_free, NULL);
    g_slist_free(Toolbar.custom);

    g_slist_foreach(Toolbar.panels, (GFunc) mdh_toolbar_panel_free, NULL);
    g_slist_free(Toolbar.panels);

    g_slist_foreach(Toolbar.errors, (GFunc) g_free, NULL);
    g_slist_free(Toolbar.errors);

    g_list_foreach(Toolbar.history, (GFunc) g_free, NULL);
    g_list_free(Toolbar.history);

    g_free(mdh_rc);
    g_free(mdh_lock);
    g_free(mdh_cache);
    g_free(mdh_hist);
}

static void mdh_toolbar_init(void)
{
    Toolbar.hidden = FALSE;

    /* only allow certain windows to be open one at a time */
    Toolbar.config_open = FALSE;
    Toolbar.error_open  = FALSE;
    Toolbar.custom_open = FALSE;
    Toolbar.pad_open    = FALSE;
    Toolbar.run_open    = FALSE;

    Toolbar.tips = gtk_tooltips_new(); /* panel only, can be disabled */

    tool_tips    = gtk_tooltips_new(); /* everything else (configuration) */

    mdh_scratch_pad_init();

    ADD_LOG(NULL, 0, "MailDooHicky v%s started", mdh_toolbar_version());
}

static void mdh_toolbar_button_setup(struct button_t *obj,
                                     const gchar *label,
                                     const gchar *tooltip,
                                     gboolean can_toggle,
                                     gboolean enabled,
                                     gpointer function)
{
    obj->name       = label;
    obj->tooltip    = tooltip;
    obj->can_toggle = can_toggle; /* allow mdh_toolbar_toggle_all() to hide? */
    obj->enabled    = enabled;
    obj->callback   = (GtkFunction) function;
}

static void mdh_toolbar_custom_setup(struct custom_t *obj,
                                     const gchar *command,
                                     const gchar *name,
                                     const gchar *tooltip)
{
    obj->command     = (command) ? g_strdup(command) : g_strdup("command");
    obj->value       = (name)    ? g_strdup(name)    : g_strdup("Name");
    obj->tooltip     = (tooltip) ? g_strdup(tooltip) : g_strdup("Tooltip");
    obj->enabled     = FALSE;
    obj->callback    = (GtkFunction) mdh_toolbar_custom_command_execute;
}

#define cfg_getBool(sec, str) ((!!cfg_getint(sec, str)))
#define cfg_getPNum(str)      ((ATOI(cfg_title(str))))

static void mdh_toolbar_setup(void)
{
    gint i;

    Toolbar.geom_x       = 0;
    Toolbar.geom_y       = 0;

    Toolbar.on_top       = TRUE;

    Toolbar.run_item     = g_strdup("xterm");

    Toolbar.run_save     = FALSE;
    Toolbar.run_close    = TRUE;

    Toolbar.tooltips     = TRUE;

    Toolbar.font         = g_strdup("");

    mdh_toolbar_button_setup(&B_Menu, "+",    "MailDooHicky Menu",
    /* disable can_toggle */ FALSE, TRUE,  mdh_toolbar_menu_show);

    mdh_toolbar_button_setup(&B_Hide, "<",    "Toggle Toolbar",
    /* disable can_toggle */ FALSE, TRUE,  mdh_toolbar_toggle_all);

    mdh_toolbar_button_setup(&B_Run,  "Run",  "Run...",
                             TRUE,  FALSE, mdh_toolbar_run);

    mdh_toolbar_button_setup(&B_Pad,  "Pad",  "Scratch-Pad",
                             TRUE,  FALSE, mdh_toolbar_scratch_pad);

    mdh_toolbar_panel_time_setup(&P_Time, "%m/%d - %l:%M %p");

    P_Mail_opt.labels    = FALSE;
    P_Mail_opt.use_bold  = TRUE;

    mdh_toolbar_panel_mail_setup(&P_Mail[0], g_getenv("MAIL"));
    mdh_toolbar_panel_mail_setup(&P_Mail[1], g_getenv("MAILDIR"));

    for(i = 2; i <= _MDH_MAX_MAIL; i++)
        mdh_toolbar_panel_mail_setup(&P_Mail[i], NULL);

#if _MDH_HAS_CPU
    P_Cpu_opt.use_bold   = TRUE;

    mdh_toolbar_panel_cpu_setup(&P_Cpu);
#endif

#if _MDH_HAS_MEM
    P_Mem_opt.use_bold   = TRUE;

    mdh_toolbar_panel_mem_setup(&P_Mem);
#endif

#if _MDH_HAS_DISK
    P_Disk_opt.labels    = FALSE;
    P_Disk_opt.use_bold  = TRUE;

    mdh_toolbar_panel_disk_setup(&P_Disk[0], "/",    "gmc /",    TRUE);
    mdh_toolbar_panel_disk_setup(&P_Disk[1], "/usr", "gmc /usr", FALSE);
    mdh_toolbar_panel_disk_setup(&P_Disk[2], "/var", "gmc /var", FALSE);

    /* setup defaults for the rest */
    for(i = 3; i <= _MDH_MAX_DISK; i++)
        mdh_toolbar_panel_disk_setup(&P_Disk[i], "/", "gmc /", FALSE);
#endif

#if _MDH_HAS_NET
# ifndef __sun__
    P_Net_opt.labels     = FALSE;
# endif

    mdh_toolbar_panel_net_setup(&P_Net[0], "eth0", TRUE);
# ifndef __sun__
    mdh_toolbar_panel_net_setup(&P_Net[1], "eth1", FALSE);
    mdh_toolbar_panel_net_setup(&P_Net[2], "ppp0", FALSE);

    for(i = 3; i <= _MDH_MAX_NET; i++)
        mdh_toolbar_panel_net_setup(&P_Net[i], "eth0", FALSE);
# endif
#endif

    P_Temp_opt.celcius   = FALSE;    /* FALSE = Ferenheight, TRUE = Celcius */

    mdh_toolbar_panel_temp_setup(&P_Temp, "KGRB");

    mdh_toolbar_panel_temp_proxy_setup();

#if _MDH_HAS_UPTIME
    mdh_toolbar_panel_uptime_setup(&P_Uptime);
#endif

    mdh_toolbar_custom_setup(&C_User[0], "mozilla", "Web", "Mozilla");
    mdh_toolbar_custom_setup(&C_User[1], "mozilla -mail","Mail","Mozilla Mail");
    mdh_toolbar_custom_setup(&C_User[2], "mozilla -news","News","Mozilla News");
    mdh_toolbar_custom_setup(&C_User[3], "ooffice", "OO", "Open Office");
    mdh_toolbar_custom_setup(&C_User[4], "gftp", "gFTP", "gFTP");
    mdh_toolbar_custom_setup(&C_User[5], "gmc", "GMC", "GNOME File Manager");
    mdh_toolbar_custom_setup(&C_User[6], "xterm", "Term", "X Terminal");
    mdh_toolbar_custom_setup(&C_User[7], "xqf", "XQF", "Quake Server Browser");
    mdh_toolbar_custom_setup(&C_User[8], "aumix", "Mix", "Audio Mixer");
    mdh_toolbar_custom_setup(&C_User[9], "xlock -mode blank", "Lock", "Lock Workstation");

    /* assign any additional custom buttons generic values */
    for(i = 10; i <= _MDH_MAX_CUST; i++)
        mdh_toolbar_custom_setup(&C_User[i], "command", "Name", "Tooltip");
}

static void mdh_toolbar_config_load(void)
{
    gint i, n;
    gchar *p;

    cfg_t *cfg, *sec;

    static cfg_opt_t panel_opts[] = {
        CFG_INT("enabled", FALSE, CFGF_NONE),
        CFG_INT("interval", 3, CFGF_NONE),
        CFG_STR("value", NULL, CFGF_NONE),
        CFG_STR("command", NULL, CFGF_NONE),
        CFG_STR("name", NULL, CFGF_NONE),
        CFG_STR("tooltip", NULL, CFGF_NONE),

        CFG_END()
    };

    static cfg_opt_t opts[] = {
        CFG_INT("Toolbar.geom_x", 0, CFGF_NONE),
        CFG_INT("Toolbar.geom_y", 0, CFGF_NONE),

        CFG_INT("Toolbar.on_top", TRUE, CFGF_NONE),

        CFG_STR("Toolbar.run_item", "xterm", CFGF_NONE),
        CFG_INT("Toolbar.run_save", FALSE, CFGF_NONE),
        CFG_INT("Toolbar.run_close", TRUE, CFGF_NONE),

        CFG_INT("Toolbar.tooltips", TRUE, CFGF_NONE),

        CFG_STR("Toolbar.font", NULL, CFGF_NONE),

        CFG_INT("B_Menu.enabled", TRUE, CFGF_NONE),
        CFG_INT("B_Hide.enabled", TRUE, CFGF_NONE),
        CFG_INT("B_Run.enabled", FALSE, CFGF_NONE),
        CFG_INT("B_Pad.enabled", FALSE, CFGF_NONE),

        CFG_SEC("User", panel_opts, CFGF_MULTI | CFGF_TITLE),

        CFG_SEC("Time", panel_opts, CFGF_NONE),

        CFG_INT("Mail.labels", FALSE, CFGF_NONE),
        CFG_INT("Mail.use_bold", TRUE, CFGF_NONE),
        CFG_SEC("Mail", panel_opts, CFGF_MULTI | CFGF_TITLE),

#if _MDH_HAS_CPU
        CFG_INT("Cpu.use_bold", TRUE, CFGF_NONE),
        CFG_SEC("Cpu", panel_opts, CFGF_NONE),
#endif /* _MDH_HAS_CPU */

#if _MDH_HAS_MEM
        CFG_INT("Mem.use_bold", TRUE, CFGF_NONE),
        CFG_SEC("Mem", panel_opts, CFGF_NONE),
#endif /* _MDH_HAS_MEM */

#if _MDH_HAS_DISK
        CFG_INT("Disk.labels", FALSE, CFGF_NONE),
        CFG_INT("Disk.use_bold", TRUE, CFGF_NONE),
        CFG_SEC("Disk", panel_opts, CFGF_MULTI | CFGF_TITLE),
#endif /* _MDH_HAS_DISK */

#if _MDH_HAS_NET
        CFG_INT("Net.labels", FALSE, CFGF_NONE),
        CFG_SEC("Net", panel_opts, CFGF_MULTI | CFGF_TITLE),
#endif /* _MDH_HAS_NET */

        CFG_INT("Temp.celcius", FALSE, CFGF_NONE),
        CFG_INT("Temp.proxy_enable", FALSE, CFGF_NONE),
        CFG_STR("Temp.proxy_host", _PROXY_HOST, CFGF_NONE),
        CFG_INT("Temp.proxy_port", _PROXY_PORT, CFGF_NONE),
        CFG_SEC("Temp", panel_opts, CFGF_NONE),

#if _MDH_HAS_UPTIME
        CFG_SEC("Uptime", panel_opts, CFGF_NONE),
#endif /* _MDH_HAS_UPTIME */

        CFG_END()
    };

    if(IS_FILE(mdh_rc) == FALSE)
        return;

    cfg = cfg_init(opts, 0);

    {
        gint ret = cfg_parse(cfg, mdh_rc);

        if(ret != CFG_SUCCESS || cfg->line < 2) {
            if(ret == CFG_FILE_ERROR)
                PERROR(NULL, 0, "mdh_toolbar_config_load: cfg_parse: %s",
                       mdh_rc);

            cfg_free(cfg);

            return;
        }
    }

    Toolbar.geom_x = cfg_getint(cfg, "Toolbar.geom_x");
    Toolbar.geom_y = cfg_getint(cfg, "Toolbar.geom_y");

    Toolbar.on_top = cfg_getBool(cfg, "Toolbar.on_top");

    if((p = cfg_getstr(cfg, "Toolbar.run_item")) != NULL) {
        g_free(Toolbar.run_item);
        Toolbar.run_item = g_strdup(g_strstrip(p));
    }

    Toolbar.run_save = cfg_getBool(cfg, "Toolbar.run_save");
    Toolbar.run_close = cfg_getBool(cfg, "Toolbar.run_close");

    Toolbar.tooltips = cfg_getBool(cfg, "Toolbar.tooltips");

    if((p = cfg_getstr(cfg, "Toolbar.font")) != NULL) {
        g_free(Toolbar.font);
        Toolbar.font = g_strdup(g_strstrip(p));
    }

    B_Menu.enabled = cfg_getBool(cfg, "B_Menu.enabled");
    B_Hide.enabled = cfg_getBool(cfg, "B_Hide.enabled");
    B_Run.enabled  = cfg_getBool(cfg, "B_Run.enabled");
    B_Pad.enabled  = cfg_getBool(cfg, "B_Pad.enabled");

    n = cfg_size(cfg, "User");

    for(i = 0; i < n; i++) {
        gint num;

        if((sec = cfg_getnsec(cfg, "User", i)) == NULL)
            break;

        num = cfg_getPNum(sec);

        if(num < 0 || num > _MDH_MAX_CUST)
            continue;

        C_User[num].enabled = cfg_getBool(sec, "enabled");

        if((p = cfg_getstr(sec, "command")) != NULL) {
            g_free(C_User[num].command);
            C_User[num].command = g_strdup(p);
        }

        if((p = cfg_getstr(sec, "name")) != NULL) {
            g_free(C_User[num].value);
            C_User[num].value = g_strdup(p);
        }

        if((p = cfg_getstr(sec, "tooltip")) != NULL) {
            g_free(C_User[num].tooltip);
            C_User[num].tooltip = g_strdup(p);
        }
    }

    if((sec = cfg_getsec(cfg, "Time")) != NULL) {
        P_Time.enabled = cfg_getBool(sec, "enabled");

        if((p = cfg_getstr(sec, "value")) != NULL) {
            g_free(P_Time.value);
            P_Time.value = g_strdup(p);
        }

        if((p = cfg_getstr(sec, "command")) != NULL) {
            g_free(P_Time.command);
            P_Time.command = g_strdup(p);
        }
    }

    P_Mail_opt.labels   = cfg_getBool(cfg, "Mail.labels");
    P_Mail_opt.use_bold = cfg_getBool(cfg, "Mail.use_bold");

    n = cfg_size(cfg, "Mail");

    for(i = 0; i < n; i++) {
        gint num;

        if((sec = cfg_getnsec(cfg, "Mail", i)) == NULL)
            break;

        num = cfg_getPNum(sec);

        if(num < 0 || num > _MDH_MAX_MAIL)
            continue;

        P_Mail[num].enabled  = cfg_getBool(sec, "enabled");

        if((p = cfg_getstr(sec, "value")) != NULL) {
            g_free(P_Mail[num].value);
            P_Mail[num].value = mdh_expand_tilde(p);
        }

        P_Mail[num].interval = cfg_getint(sec, "interval");

        if((p = cfg_getstr(sec, "command")) != NULL) {
            g_free(P_Mail[num].command);
            P_Mail[num].command = g_strdup(p);
        }

        g_free(P_Mail[num].tooltip);
        P_Mail[num].tooltip  = g_strdup_printf("Mail: %s", P_Mail[num].value);
    }

#if _MDH_HAS_CPU
    P_Cpu_opt.use_bold = cfg_getBool(cfg, "Cpu.use_bold");

    if((sec = cfg_getsec(cfg, "Cpu")) != NULL) {
        P_Cpu.enabled  = cfg_getBool(sec, "enabled");
        P_Cpu.interval = cfg_getint(sec, "interval");

        if((p = cfg_getstr(sec, "command")) != NULL) {
            g_free(P_Cpu.command);
            P_Cpu.command = g_strdup(p);
        }
    }
#endif /* _MDH_HAS_CPU */

#if _MDH_HAS_MEM
    P_Mem_opt.use_bold = cfg_getBool(cfg, "Mem.use_bold");

    if((sec = cfg_getsec(cfg, "Mem")) != NULL) {
        P_Mem.enabled  = cfg_getBool(sec, "enabled");
        P_Mem.interval = cfg_getint(sec, "interval");

        if((p = cfg_getstr(sec, "command")) != NULL) {
            g_free(P_Mem.command);
            P_Mem.command = g_strdup(p);
        }
    }
#endif /* _MDH_HAS_MEM */

#if _MDH_HAS_DISK
    P_Disk_opt.labels   = cfg_getBool(cfg, "Disk.labels");
    P_Disk_opt.use_bold = cfg_getBool(cfg, "Disk.use_bold");

    n = cfg_size(cfg, "Disk");

    for(i = 0; i < n; i++) {
        gint num;

        if((sec = cfg_getnsec(cfg, "Disk", i)) == NULL)
            break;

        num = cfg_getPNum(sec);

        if(num < 0 || num > _MDH_MAX_DISK)
            continue;

        P_Disk[num].enabled = cfg_getBool(sec, "enabled");

        if((p = cfg_getstr(sec, "value")) != NULL) {
            g_free(P_Disk[num].value);
            P_Disk[num].value = g_strdup(g_strstrip(p));
        }

        if((p = cfg_getstr(sec, "command")) != NULL) {
            g_free(P_Disk[num].command);
            P_Disk[num].command = g_strdup(p);
        }

        g_free(P_Disk[num].tooltip);
        P_Disk[num].tooltip = g_strdup_printf("Disk: %s", P_Disk[num].value);
    }
#endif /* _MDH_HAS_DISK */

#if _MDH_HAS_NET
# ifndef __sun__
    P_Net_opt.labels = cfg_getBool(cfg, "Net.labels");
# endif

    n = cfg_size(cfg, "Net");

    for(i = 0; i < n; i++) {
        gint num;

        if((sec = cfg_getnsec(cfg, "Net", i)) == NULL)
            break;

        num = cfg_getPNum(sec);

        if(num < 0 || num > _MDH_MAX_NET)
            continue;

        P_Net[num].enabled  = cfg_getBool(sec, "enabled");
        P_Net[num].interval = cfg_getint(sec, "interval");

#ifndef __sun__
        if((p = cfg_getstr(sec, "value")) != NULL) {
            g_free(P_Net[num].value);
            P_Net[num].value = g_strdup(g_strstrip(p));
        }
#endif

        if((p = cfg_getstr(sec, "command")) != NULL) {
            g_free(P_Net[num].command);
            P_Net[num].command = g_strdup(p);
        }

#ifndef __sun__
        g_free(P_Net[num].tooltip);
        P_Net[num].tooltip = g_strdup_printf("Net: %s", P_Net[num].value);
#endif
    }
#endif /* _MDH_HAS_NET */

    P_Temp_opt.celcius      = cfg_getBool(cfg, "Temp.celcius");
    P_Temp_opt.proxy_enable = cfg_getBool(cfg, "Temp.proxy_enable");

    if((p = cfg_getstr(cfg, "Temp.proxy_host")) != NULL) {
        g_free(P_Temp_opt.proxy_host);
        P_Temp_opt.proxy_host = g_strdup(g_strstrip(p));
    }

    P_Temp_opt.proxy_port   = cfg_getint(cfg,  "Temp.proxy_port");

    if((sec = cfg_getsec(cfg, "Temp")) != NULL) {
        P_Temp.enabled  = cfg_getBool(sec, "enabled");
        P_Temp.interval = cfg_getint(sec, "interval");

        if((p = cfg_getstr(sec, "value")) != NULL) {
            g_free(P_Temp.value);
            P_Temp.value = g_ascii_strup(p, -1);
        }

        if((p = cfg_getstr(sec, "command")) != NULL) {
            g_free(P_Temp.command);
            P_Temp.command = g_strdup(p);
        }

        g_free(P_Temp.tooltip);
        P_Temp.tooltip = g_strdup_printf("Temp: %s", P_Temp.value);
    }

#if _MDH_HAS_UPTIME
    if((sec = cfg_getsec(cfg, "Uptime")) != NULL) {
        P_Uptime.enabled = cfg_getBool(sec, "enabled");

        if((p = cfg_getstr(sec, "command")) != NULL) {
            g_free(P_Uptime.command);
            P_Uptime.command = g_strdup(p);
        }
    }
#endif /* _MDH_HAS_UPTIME */

    cfg_free(cfg);

    /* some sanity checks */

    for(i = 0; i <= _MDH_MAX_MAIL; i++)
        if(P_Mail[i].interval < _MAIL_I_MIN || P_Mail[i].interval > _MAIL_I_MAX)
            P_Mail[i].interval = _MAIL_I_DEF;

#if _MDH_HAS_CPU
    if(P_Cpu.interval < 1)
        P_Cpu.interval = 3;
#endif

#if _MDH_HAS_MEM
    if(P_Mem.interval < 1)
        P_Mem.interval = 3;
#endif

#if _MDH_HAS_DISK
    for(i = 0; i <= _MDH_MAX_DISK; i++)
        if(P_Disk[i].interval < 1)
            P_Disk[i].interval = 60;
#endif

#if _MDH_HAS_NET
    for(i = 0; i <= _MDH_MAX_NET; i++)
        if(P_Net[i].interval < 1)
            P_Net[i].interval = 3;
#endif

    if(P_Temp.interval < _WEATHER_MIN) {
        ADD_LOG(NULL, 0,
            "mdh_toolbar_config_load: temp interval too low, reset to %d",
            _WEATHER_MIN);

        P_Temp.interval = _WEATHER_MIN;
    }

    if(P_Temp_opt.proxy_port < _TCP_PORT_MIN ||
      P_Temp_opt.proxy_port > _TCP_PORT_MAX) {
        ADD_LOG(NULL, 0,
            "mdh_toolbar_config_load: proxy port (%d) out of range, set to %d",
            P_Temp_opt.proxy_port, _PROXY_PORT);

        P_Temp_opt.proxy_port = _PROXY_PORT;
    }

    if(STRLEN(P_Temp.value) != _WEATHER_LEN) {
        g_free(P_Temp.value);

        P_Temp.value = g_strdup("KGRB");

        mdh_window_notice("Invalid Station-ID.",
        "Station-ID's should be 4 characters in length. Value set to default.");

        ADD_LOG(NULL, 0,
            "mdh_toolbar_config_load: invalid station-id, set to %s",
            P_Temp.value);
    }
}

static const gchar *mdh_strescape(const gchar *str)
{
    static gchar buf[LINE_MAX];
    gchar *p;

    if(str == NULL || str[0] == '\0')
        return(NULL);

    p = g_strescape(str, NULL);
    g_strlcpy(buf, p, sizeof(buf));
    g_free(p);

    return(buf);
}

static void mdh_toolbar_config_save(void)
{
    gchar *mdh_rc_tmp;

    gint i;

    FILE *fp;

    mdh_rc_tmp = g_strdup_printf("%s.tmp", mdh_rc);

    if((fp = fopen(mdh_rc_tmp, "w")) == NULL) {
        PERROR(NULL, 0, "mdh_toolbar_config_save: fopen: %s", mdh_rc);
        g_free(mdh_rc_tmp);
        return;
    }

    fprintf(fp, "# MailDooHicky v%s\n", mdh_toolbar_version());
    fprintf(fp, "#   %s: automatically generated, do not edit.\n\n", mdh_rc);

    fprintf(fp, "# Toolbar geometry\n");
    fprintf(fp, "Toolbar.geom_x = %d\n", Toolbar.geom_x);
    fprintf(fp, "Toolbar.geom_y = %d\n\n", Toolbar.geom_y);

    fprintf(fp, "# Try to keep the toolbar on top of other windows\n");
    fprintf(fp, "Toolbar.on_top = %d\n\n", Toolbar.on_top);

    fprintf(fp, "# Run window options:\n");
    fprintf(fp, "#   run_item:  xterm for 'Run in terminal' (-e assumed)\n");
    fprintf(fp, "#   run_save:  save run history\n");
    fprintf(fp, "#   run_close: close window after command execution\n");
    fprintf(fp, "Toolbar.run_item = \"%s\"\n", mdh_strescape(Toolbar.run_item));
    fprintf(fp, "Toolbar.run_save = %d\n", Toolbar.run_save);
    fprintf(fp, "Toolbar.run_close = %d\n\n", Toolbar.run_close);

    fprintf(fp, "# Toggle toolbar tooltips\n");
    fprintf(fp, "Toolbar.tooltips = %d\n\n", Toolbar.tooltips);

    fprintf(fp, "# Custom font for main toolbar (eg. \"Nimbus Sans L, 8\")\n");
    fprintf(fp, "Toolbar.font = \"%s\"\n\n",
        (ISSET(Toolbar.font)) ? mdh_strescape(Toolbar.font) : "");

    fprintf(fp, "# Toolbar buttons\n");
    fprintf(fp, "B_Menu.enabled = %d\n", B_Menu.enabled);
    fprintf(fp, "B_Hide.enabled = %d\n", B_Hide.enabled);
    fprintf(fp, "B_Run.enabled = %d\n", B_Run.enabled);
    fprintf(fp, "B_Pad.enabled = %d\n\n", B_Pad.enabled);

    fprintf(fp, "# User-definable buttons:\n");
    fprintf(fp, "#   command: command to run\n");
    fprintf(fp, "#   name:    button text\n");
    fprintf(fp, "#   tooltip: button tooltip, none to disable\n\n");

    for(i = 0; i <= _MDH_MAX_CUST; i++) {
        fprintf(fp, "User %d {\n", i);
        fprintf(fp, "    enabled = %d\n", C_User[i].enabled);
        fprintf(fp, "    command = \"%s\"\n", mdh_strescape(C_User[i].command));
        fprintf(fp, "    name = \"%s\"\n", mdh_strescape(C_User[i].value));
        fprintf(fp, "    tooltip = \"%s\"\n", mdh_strescape(C_User[i].tooltip));
        fprintf(fp, "}\n\n");
    }

    fprintf(fp, "# Common panel options:\n");
    fprintf(fp, "#   value:    data used by functions (time format, net ");
    fprintf(fp, "interface, etc).\n");
    fprintf(fp, "#   command:  command to run when clicked, none to disable\n");
    fprintf(fp, "#   interval: interval at which panel data is refreshed\n\n");

    fprintf(fp, "# Time panel: displays current time. see strftime(3)\n");
    fprintf(fp, "Time {\n");
    fprintf(fp, "    enabled = %d\n", P_Time.enabled);
    fprintf(fp, "    value = \"%s\"\n", mdh_strescape(P_Time.value));
    fprintf(fp, "    command = \"%s\"\n", mdh_strescape(P_Time.command));
    fprintf(fp, "}\n\n");

    fprintf(fp, "# Common mail options\n");
    fprintf(fp, "Mail.labels = %d\n", P_Mail_opt.labels);
    fprintf(fp, "Mail.use_bold = %d\n\n", P_Mail_opt.use_bold);

    fprintf(fp, "# Mail panels: displays read/total counts\n\n");

    for(i = 0; i <= _MDH_MAX_MAIL; i++) {
        fprintf(fp, "Mail %d {\n", i);
        fprintf(fp, "    enabled = %d\n", P_Mail[i].enabled);
        fprintf(fp, "    value = \"%s\"\n", mdh_strescape(P_Mail[i].value));
        fprintf(fp, "    command = \"%s\"\n", mdh_strescape(P_Mail[i].command));
        fprintf(fp, "    interval = %d\n", P_Mail[i].interval);
        fprintf(fp, "}\n\n");
    }

#if _MDH_HAS_CPU
    fprintf(fp, "# Common cpu options\n");
    fprintf(fp, "Cpu.use_bold = %d\n\n", P_Cpu_opt.use_bold);

    fprintf(fp, "# CPU panel: displays cpu load\n");
    fprintf(fp, "Cpu {\n");
    fprintf(fp, "    enabled = %d\n", P_Cpu.enabled);
    fprintf(fp, "    command = \"%s\"\n", mdh_strescape(P_Cpu.command));
    fprintf(fp, "    interval = %d\n", P_Cpu.interval);
    fprintf(fp, "}\n\n");
#endif /* _MDH_HAS_CPU */

#if _MDH_HAS_MEM
    fprintf(fp, "# Common mem options\n");
    fprintf(fp, "Mem.use_bold = %d\n\n", P_Mem_opt.use_bold);

    fprintf(fp, "# Mem panel: displays memory usage\n");
    fprintf(fp, "Mem {\n");
    fprintf(fp, "    enabled = %d\n", P_Mem.enabled);
    fprintf(fp, "    command = \"%s\"\n", mdh_strescape(P_Mem.command));
    fprintf(fp, "    interval = %d\n", P_Mem.interval);
    fprintf(fp, "}\n\n");
#endif /* _MDH_HAS_MEM */

#if _MDH_HAS_DISK
    fprintf(fp, "# Common disk options\n");
    fprintf(fp, "Disk.labels = %d\n", P_Disk_opt.labels);
    fprintf(fp, "Disk.use_bold = %d\n\n", P_Disk_opt.use_bold);

    fprintf(fp, "# Disk panels: displays disk usage\n");
    fprintf(fp, "#   value: filesystem to monitor\n\n");

    for(i = 0; i <= _MDH_MAX_DISK; i++) {
        fprintf(fp, "Disk %d {\n", i);
        fprintf(fp, "    enabled = %d\n", P_Disk[i].enabled);
        fprintf(fp, "    value = \"%s\"\n", mdh_strescape(P_Disk[i].value));
        fprintf(fp, "    command = \"%s\"\n", mdh_strescape(P_Disk[i].command));
        fprintf(fp, "}\n\n");
    }
#endif /* _MDH_HAS_DISK */

#if _MDH_HAS_NET
# ifndef __sun__
    fprintf(fp, "# Common net options\n");
    fprintf(fp, "Net.labels = %d\n\n", P_Net_opt.labels);
# endif

    fprintf(fp, "# Net panel: displays network utilization\n");
    fprintf(fp, "#   value: interface to monitor\n\n");

    for(i = 0; i <= _MDH_MAX_NET; i++) {
        fprintf(fp, "Net %d {\n", i);
        fprintf(fp, "    enabled = %d\n", P_Net[i].enabled);
#ifndef __sun__
        fprintf(fp, "    value = \"%s\"\n", mdh_strescape(P_Net[i].value));
#endif
        fprintf(fp, "    command = \"%s\"\n", mdh_strescape(P_Net[i].command));
        fprintf(fp, "    interval = %d\n", P_Net[i].interval);
        fprintf(fp, "}\n\n");
    }
#endif /* _MDH_HAS_NET */

    fprintf(fp, "# Common temp options\n\n");

    fprintf(fp, "# Display temperature in Celcius vs Ferenheight\n");
    fprintf(fp, "Temp.celcius = %d\n\n", P_Temp_opt.celcius);

    fprintf(fp, "# Proxy options\n");
    fprintf(fp, "Temp.proxy_enable = %d\n", P_Temp_opt.proxy_enable);
    fprintf(fp, "Temp.proxy_host = \"%s\"\n",
            mdh_strescape(P_Temp_opt.proxy_host));
    fprintf(fp, "Temp.proxy_port = %d\n\n", P_Temp_opt.proxy_port);

    fprintf(fp, "# Temp panel: displays temperature of selected station-id\n");
    fprintf(fp, "#   See http://www.nws.noaa.gov/tg/siteloc.shtml\n");

    fprintf(fp, "Temp {\n");
    fprintf(fp, "    enabled = %d\n", P_Temp.enabled);
    fprintf(fp, "    value = \"%s\"\n", mdh_strescape(P_Temp.value));
    fprintf(fp, "    command = \"%s\"\n", mdh_strescape(P_Temp.command));
    fprintf(fp, "    interval = %d\n", P_Temp.interval);
    fprintf(fp, "}\n\n");

#if _MDH_HAS_UPTIME
    fprintf(fp, "# Uptime panel: displays current uptime\n");
    fprintf(fp, "Uptime {\n");
    fprintf(fp, "    enabled = %d\n", P_Uptime.enabled);
    fprintf(fp, "    command = \"%s\"\n", mdh_strescape(P_Uptime.command));
    fprintf(fp, "}\n");
#endif /* _MDH_HAS_UPTIME */

    fclose(fp);

    if(IS_FILE(mdh_rc) == TRUE)
        mdh_unlink(mdh_rc);

    mdh_rename(mdh_rc_tmp, mdh_rc);

    g_free(mdh_rc_tmp);
}

/*
 * maximum run history size (used when loaded from history file). should
 * this be configurable? if so, perhaps a pre-defined radio-button type
 * setup on the run window's Option menu? 25, 50, 100, 250 ...
 */
#define _RUN_HIST_MAX     50

static void mdh_toolbar_history_load(void)
{
    FILE *fp;

    /* should the run history be loaded if saving is disabled? */

    /* if(Toolbar.run_save == FALSE)
        return; */

    if(IS_FILE(mdh_hist) == FALSE)
        return;

    if((fp = fopen(mdh_hist, "r")) == NULL) {
        PERROR(NULL, 0, "mdh_toolbar_history_load: fopen: %s", mdh_hist);
        return;
    }

    {
        gint count = 0;

        gchar buf[LINE_MAX];

        while((fgets(buf, sizeof(buf), fp))) {
            buf[STRLEN(buf) - 1] = 0;

            if(!ISSET(buf) || buf[0] == '#')
                continue;

            if(count++ >= _RUN_HIST_MAX)
                break;

            Toolbar.history = g_list_append(Toolbar.history, g_strdup(buf));
        }
    }

    fclose(fp);
}

static void mdh_toolbar_history_save(void)
{
    gchar *mdh_hist_tmp;

    FILE *fp;

    if(Toolbar.run_save == FALSE || Toolbar.history == NULL)
        return;

    mdh_hist_tmp = g_strdup_printf("%s.tmp", mdh_hist);

    if((fp = fopen(mdh_hist_tmp, "w")) == NULL) {
        PERROR(NULL, 0, "mdh_toolbar_history_save: fopen: %s", mdh_hist_tmp);
        g_free(mdh_hist_tmp);
        return;
    }

    {
        GList *p;

        for(p = Toolbar.history; p != NULL; p = g_list_next(p))
            fprintf(fp, "%s\n", (gchar *) p->data);
    }

    fclose(fp);

    if(IS_FILE(mdh_hist) == TRUE)
        mdh_unlink(mdh_hist);

    mdh_rename(mdh_hist_tmp, mdh_hist);

    g_free(mdh_hist_tmp);
}

static GtkFunction mdh_window_toolbar_configure(GtkWidget *widget,
                                                GdkEventConfigure *event)
{
    Toolbar.geom_x = event->x;
    Toolbar.geom_y = event->y;

    return((GtkFunction) FALSE);
}

static GtkFunction mdh_window_toolbar_map_event(GtkWidget *widget,
                                                GdkEventConfigure *event)
{
    mdh_toolbar_set_layer_on_top(Toolbar.on_top);

    return((GtkFunction) FALSE);
}

static gchar *mdh_readlink(const gchar *symlink)
{
    static gchar buf[PATH_MAX];

    gint len;

    if((len = readlink(symlink, buf, sizeof(buf))) == -1) {
        PERROR(NULL, 0, "mdh_readlink: readlink: %s", symlink);
        return(NULL);
    }

    buf[len] = '\0';

    return(buf);
}

/* checks to see if specified pid is valid and running (may not be mdh) */
static gboolean mdh_is_pid(pid_t pid)
{
    if(pid < 1 || kill(pid, 0) == -1)
        return(FALSE);

    return(TRUE);
}

static void mdh_toolbar_lock(void)
{
    gchar *p;

    if(IS_LINK(mdh_lock) == TRUE) {
        pid_t pid;

        if((p = mdh_readlink(mdh_lock)) == NULL)
            exit(1);

        pid = (pid_t) ATOI(p);

        /* see if the pid is still running, if not unlink() and continue on */
	if(mdh_is_pid(pid) == TRUE) {
            gchar *text = g_strdup_printf("Existing mdh lock-file (%s) found. Continue anyway?", mdh_lock);

            if(mdh_window_yes_no_prompt(text) == FALSE)
                exit(1);

            g_free(text);
        }

        mdh_unlink(mdh_lock);
    }

    p = g_strdup_printf("%ld", (glong) getpid());

    /*
     * this shouldn't be a fatal error, just would rather not have multiple
     * processes updating the same config file.
     */

    if(symlink(p, mdh_lock) == -1)
        PERROR(NULL, 0, "mdh_toolbar_lock: symlink: %s", mdh_lock);

    g_free(p);
}

static void mdh_toolbar_create(void)
{
    gint i, id = 0;

    GtkWidget *window, *hbox;

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

    gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, TRUE);

    gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DOCK);

    gtk_window_set_decorated(GTK_WINDOW(window), FALSE);
    gtk_window_stick(GTK_WINDOW(window));

    gtk_window_set_skip_taskbar_hint(GTK_WINDOW(window), TRUE);
    gtk_window_set_skip_pager_hint(GTK_WINDOW(window), TRUE);

    /* allow the toolbar window to resize according to it's contents */
    gtk_window_set_resizable(GTK_WINDOW(window), FALSE);

    g_signal_connect(G_OBJECT(window), "map_event",
        G_CALLBACK(mdh_window_toolbar_map_event), NULL);

    g_signal_connect(G_OBJECT(window), "configure_event",
        G_CALLBACK(mdh_window_toolbar_configure), NULL);

    g_signal_connect(G_OBJECT(window), "destroy",
        G_CALLBACK(gtk_main_quit), NULL);

    gtk_container_set_border_width(GTK_CONTAINER(window), 1);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);

    gtk_container_add(GTK_CONTAINER(window), hbox);

    mdh_toolbar_button_add(hbox, &B_Menu);
    mdh_toolbar_button_add(hbox, &B_Hide);
    mdh_toolbar_button_add(hbox, &B_Run);
    mdh_toolbar_button_add(hbox, &B_Pad);

    /* initialize custom buttons */
    for(i = 0; i <= _MDH_MAX_CUST; i++)
        mdh_toolbar_custom_add(hbox, &C_User[i]);

    mdh_toolbar_panel_add(hbox, &P_Time, _PANEL_TYPE_TIME, id++);

    for(i = 0; i <= _MDH_MAX_MAIL; i++)
        mdh_toolbar_panel_add(hbox, &P_Mail[i], _PANEL_TYPE_MAIL, id++);

#if _MDH_HAS_CPU
    mdh_toolbar_panel_add(hbox, &P_Cpu, _PANEL_TYPE_CPU, id++);
#endif

#if _MDH_HAS_MEM
    mdh_toolbar_panel_add(hbox, &P_Mem, _PANEL_TYPE_MEM, id++);
#endif

#if _MDH_HAS_DISK
    for(i = 0; i <= _MDH_MAX_DISK; i++)
        mdh_toolbar_panel_add(hbox, &P_Disk[i], _PANEL_TYPE_DISK, id++);
#endif

#if _MDH_HAS_NET
    for(i = 0; i <= _MDH_MAX_NET; i++)
        mdh_toolbar_panel_add(hbox, &P_Net[i], _PANEL_TYPE_NET, id++);
#endif

    mdh_toolbar_panel_add(hbox, &P_Temp, _PANEL_TYPE_TEMP, id++);

#if _MDH_HAS_UPTIME
    mdh_toolbar_panel_add(hbox, &P_Uptime, _PANEL_TYPE_UPTIME, id++);
#endif

    gtk_widget_realize(window);

    Toolbar.window = window;

    mdh_toolbar_set_winops();
}

static void mdh_toolbar_show(void)
{
    gtk_widget_show(Toolbar.window);
}

static void mdh_usage(const gchar *argv0)
{
    g_print("Usage: %s [options] [rc]\n", argv0);
    g_print("  options:\n");
    g_print("    -d/--debug     enable debug\n");
    g_print("    -h/--help      help (this message)\n");
    g_print("    -v/--version   display version\n");
}

static void mdh_mkdir(const gchar *path, gint mode)
{
    g_return_if_fail(path != NULL && path[0] != '\0');

    if(IS_DIR(path) == TRUE)
        return;

    if(mkdir(path, mode) == -1) {
        PERROR(NULL, 0, "mdh_mkdir: %s", path, g_strerror(errno));
        exit(1);
    }
}

static void mdh_mkdir_path(const gchar *path, gint mode)
{
    gchar *c, *e;

    g_return_if_fail(path != NULL && path[0] != '\0');

    if(IS_DIR(path) == TRUE)
        return;

    c = g_strdup(path);
    e = c;
    while((e = strchr(e + 1, '/'))) {
        *e = '\0';
        mdh_mkdir(c, mode);
        *e = '/';
    }
    mdh_mkdir(c, mode);
    g_free(c);
}

#define MDH_CLI_ARG(str, l, s) (STREQ(str, l) || STREQ(str, s))

static void mdh_init(gint argc, gchar **argv)
{
    gint i;

    gchar *rc = NULL;

    memset(&Toolbar, '\0', sizeof(struct toolbar));

    Toolbar.debug = FALSE;

    for(i = 1; i < argc; i++) {
        if(MDH_CLI_ARG(argv[i], "--help", "-h")) {
            mdh_usage(argv[0]);
            exit(0);
        } else if(MDH_CLI_ARG(argv[i], "--debug", "-d")) {
            Toolbar.debug = TRUE;
        } else if(MDH_CLI_ARG(argv[i], "--version", "-v")) {
            g_print("MailDooHicky v%s\n", mdh_toolbar_version());
            exit(0);
        } else if(!STRNEQ(argv[i], "-", 1)) {
            rc = argv[i];
        } else {
            g_print("unknown option '%s'\n\n", argv[i]);
            mdh_usage(argv[0]);
            exit(1);
        }
    }

    set_signal(SIGTERM, mdh_signal_quit);
    set_signal(SIGQUIT, mdh_signal_quit);
    set_signal(SIGINT,  mdh_signal_quit);
    set_signal(SIGHUP,  mdh_signal_quit);

    set_signal(SIGUSR1, mdh_signal_reconfigure);
    set_signal(SIGUSR2, SIG_IGN);

    set_signal(SIGPIPE, SIG_IGN);

    gtk_init(&argc, &argv);

    {
        const gchar *path = g_getenv("XDG_CONFIG_HOME");

        gchar *p, *s;

        if(ISSET(path))
            p = g_strdup(path);
        else
            p = g_build_filename(g_get_home_dir(), ".config", NULL);

        mdh_mkdir_path(p, 0755);

        s = g_build_filename(p, "mdh", NULL);

        g_free(p);

        mdh_mkdir(s, 0755);

        if(rc == NULL)
            mdh_rc = g_build_filename(s, "rc", NULL);
        else
            mdh_rc = g_strdup(rc);

        mdh_lock  = g_strdup_printf("%s.lock", mdh_rc);

        mdh_cache = g_build_filename(s, "cache", NULL);
        mdh_hist  = g_build_filename(s, "hist",  NULL);

        g_free(s);
    }

#if 0
    if(chdir(g_get_home_dir()) == -1)
        ADD_LOG(NULL, 0, "Unable to change to home directory: %s",
                g_strerror(errno));
#endif
}

static void mdh_toolbar_set_winops(void)
{
    GdkAtom atoms[5];
    gint i = 0;

    atoms[0] = gdk_atom_intern("_NET_WM_WINDOW_TYPE_DOCK", FALSE);

    gdk_property_change(Toolbar.window->window,
                        gdk_atom_intern("_NET_WM_WINDOW_TYPE", FALSE),
                        gdk_atom_intern("ATOM", FALSE), 32,
                        GDK_PROP_MODE_REPLACE, (guchar *) atoms, 1);

    atoms[i++] = gdk_atom_intern("_NET_WM_STATE_SKIP_TASKBAR", FALSE);
    atoms[i++] = gdk_atom_intern("_NET_WM_STATE_SKIP_PAGER", FALSE);

    if(Toolbar.on_top) {
        atoms[i++] = gdk_atom_intern("_NET_WM_STATE_ABOVE", FALSE);
        atoms[i++] = gdk_atom_intern("_NET_WM_STATE_STAYS_ON_TOP", FALSE);
    } else {
        atoms[i++] = gdk_atom_intern("_NET_WM_STATE_BELOW", FALSE);
    }

    gdk_property_change(Toolbar.window->window,
                        gdk_atom_intern("_NET_WM_STATE", FALSE),
                        gdk_atom_intern("ATOM", FALSE), 32,
                        GDK_PROP_MODE_REPLACE, (guchar *) atoms, i);
}

#define _NET_WM_STATE_REMOVE 0
#define _NET_WM_STATE_ADD    1
#define _NET_WM_STATE_TOGGLE 2

static void mdh_toolbar_set_net_wm_state(const gchar *hint, gboolean state)
{
    XEvent xev;

    xev.type                 = ClientMessage;
    xev.xclient.type         = ClientMessage;
    xev.xclient.window       = GDK_WINDOW_XWINDOW(Toolbar.window->window);
    xev.xclient.message_type = gdk_x11_get_xatom_by_name("_NET_WM_STATE");
    xev.xclient.format       = 32;
    xev.xclient.data.l[0]    = state ? _NET_WM_STATE_ADD : _NET_WM_STATE_REMOVE;
    xev.xclient.data.l[1]    = gdk_x11_get_xatom_by_name(hint);
    xev.xclient.data.l[2]    = 0;

    XSendEvent(GDK_DISPLAY(), GDK_ROOT_WINDOW(),
               False, SubstructureNotifyMask, &xev);
}

void mdh_toolbar_set_layer_on_top(gboolean state)
{
    mdh_toolbar_set_net_wm_state("_NET_WM_STATE_BELOW",       !state);

    mdh_toolbar_set_net_wm_state("_NET_WM_STATE_ABOVE",        state);
    mdh_toolbar_set_net_wm_state("_NET_WM_STATE_STAYS_ON_TOP", state);
}
