/*
 * mdh (MailDooHicky)
 *
 * Copyright (c) 2003 Mike Hokenson <mdh at gozer dot org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>

#include <gtk/gtk.h>

#include "mdh.h"
#include "util.h"

#include "mdh_gen.h"
#include "mdh_cfg.h"

#include "mdh_toolbar.h"
#include "mdh_panel_mail.h"
#include "mdh_panel_temp.h"

/* configuration widgets */

enum {
    CFG_TIME_ENABLED,
    CFG_TIME_VALUE,
    CFG_TIME_COMMAND
};

enum {
    CFG_MAIL_ENABLED,
    CFG_MAIL_VALUE,
    CFG_MAIL_COMMAND,
    CFG_MAIL_SPINNER
};

#if _MDH_HAS_CPU
enum {
    CFG_CPU_ENABLED,
    CFG_CPU_SPINNER,
    CFG_CPU_COMMAND,
    CFG_CPU_BOLD
};
#endif

#if _MDH_HAS_MEM
enum {
    CFG_MEM_ENABLED,
    CFG_MEM_SPINNER,
    CFG_MEM_COMMAND,
    CFG_MEM_BOLD
};
#endif

#if _MDH_HAS_DISK
enum {
    CFG_DISK_ENABLED,
    CFG_DISK_VALUE,
    CFG_DISK_COMMAND
};
#endif

#if _MDH_HAS_NET
enum {
    CFG_NET_ENABLED,
    CFG_NET_SPINNER,
    CFG_NET_VALUE,
    CFG_NET_COMMAND
};
#endif

enum {
    CFG_TEMP_ENABLED,
    CFG_TEMP_SPINNER,
    CFG_TEMP_VALUE,
    CFG_TEMP_COMMAND,
    CFG_TEMP_CELCIUS,
    CFG_TEMP_PROXY_ENABLE,
    CFG_TEMP_PROXY_HOST,
    CFG_TEMP_PROXY_PORT
};

#if _MDH_HAS_UPTIME
enum {
    CFG_UPTIME_ENABLED,
    CFG_UPTIME_COMMAND
};
#endif

struct cfg_common {
    gboolean     enabled;
    const gchar *value;
    const gchar *command;
    gint         interval;
};

static GtkWidget *p_font_w, *p_tooltips_w, *p_on_top_w, *p_run_item_w;

static GtkWidget *p_mail_labels_w, *p_mail_use_bold_w;

#ifdef _MDH_HAS_DISK
static GtkWidget *p_disk_labels_w, *p_disk_use_bold_w;
#endif

#ifdef _MDH_HAS_NET
# ifndef __sun__
static GtkWidget *p_net_labels_w;
# endif
#endif

static gboolean p_temp_proxy_enable;

static gint geom_x, geom_y;

extern GtkTooltips *tool_tips;

static void mdh_window_configuration_close(GtkWidget *widget,
                                           GtkWidget *window);

static void mdh_window_configuration_cancel(GtkWidget *widget,
                                            GtkWidget *window);

static gboolean mdh_window_configuration_apply(void);

static void mdh_spinner_update_geom_x(GtkWidget *widget, GtkWidget *spinner);
static void mdh_spinner_update_geom_y(GtkWidget *widget, GtkWidget *spinner);

static void mdh_button_toggle_temp_proxy_value(GtkWidget *widget,
                                               gpointer data);

static void mdh_button_toggle_panel(GtkWidget *widget, struct panel_t *obj);
static void mdh_button_toggle_panel_temp(GtkWidget *widget,
                                         struct panel_t *obj);

static void mdh_select_font(void);
static void mdh_unselect_font(void);
static void mdh_font_selection_ok(GtkWidget *widget, GtkWidget *window);

static GtkWidget *mdh_frame_new_with_title(const gchar *title);

static void mdh_toolbar_get_new(void);
static void mdh_toolbar_button_get_new(struct button_t *obj);
static void mdh_toolbar_panel_get_new(struct panel_t *obj);

static void mdh_toolbar_get_cfg_common(struct panel_t *obj,
                                       struct cfg_common *cfg);

void mdh_window_configuration(void)
{
    GtkWidget *window;
    GtkWidget *vbox, *vbox3, *vbox4, *vbox5;
    GtkWidget *hbox, *hbox2;
    GtkWidget *frame;
    GtkWidget *notebook;
    GtkWidget *label;
    GtkWidget *entry;
    GtkWidget *button, *p_button;
    GtkWidget *hsep, *vsep;

    GtkObject *adj;
    GtkWidget *spinner;

    gint i, max_geom_x, max_geom_y;

    if(Toolbar.config_open == TRUE)
        return;

    max_geom_x = gdk_screen_width()  - 5;
    max_geom_y = gdk_screen_height() - 5;

    /* allow toolbar geometry to be previewed, reset if canceled */
    geom_x = Toolbar.geom_x;
    geom_y = Toolbar.geom_y;

    /* expand panel */
    if(Toolbar.hidden == TRUE)
        mdh_toolbar_toggle_all();

    /* set after mdh_toolbar_toggle_all() otherwise the toolbar won't expand */
    mdh_toolbar_window_toggle_open(&Toolbar.config_open);

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), "MailDooHicky: Configuration");
    gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);

    /* ignores the window managers's close request */
    gtk_signal_connect(GTK_OBJECT(window), "delete_event",
        GTK_SIGNAL_FUNC(gtk_true), NULL);

    gtk_container_set_border_width(GTK_CONTAINER(window), 5);

    vbox = gtk_vbox_new(FALSE, 5);
    gtk_widget_show(vbox);

    gtk_container_add(GTK_CONTAINER(window), vbox);

    notebook = gtk_notebook_new();
    gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(notebook), TRUE);
    gtk_notebook_set_homogeneous_tabs(GTK_NOTEBOOK(notebook), TRUE);
    gtk_widget_show(notebook);

    gtk_box_pack_start(GTK_BOX(vbox), notebook, TRUE, TRUE, 0);

    /* General tab */

    vbox3 = gtk_vbox_new(FALSE, 3);
    gtk_widget_show(vbox3);

    gtk_container_set_border_width(GTK_CONTAINER(vbox3), 5);

    hbox2 = gtk_hbox_new(FALSE, 3);
    gtk_box_pack_start(GTK_BOX(vbox3), hbox2, TRUE, TRUE, 0);
    gtk_widget_show(hbox2);

    frame = mdh_frame_new_with_title("Location");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(hbox2), frame, FALSE, TRUE, 0);

    hbox = gtk_hbox_new(TRUE, 3);
    gtk_widget_show(hbox);

    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);

    gtk_container_add(GTK_CONTAINER(frame), hbox);

    adj = gtk_adjustment_new(Toolbar.geom_x, 0.0, max_geom_x, 1.0, 50.0, 0.0);
    spinner = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 0);
    gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, TRUE, 2);
    gtk_widget_show(spinner);

    mdh_tooltips_set_tip(tool_tips, spinner,
        "Adjust the toolbars horizontal location");

    g_signal_connect(G_OBJECT(adj), "value_changed",
        G_CALLBACK(mdh_spinner_update_geom_x), spinner);

    adj = gtk_adjustment_new(Toolbar.geom_y, 0.0, max_geom_y, 1.0, 50.0, 0.0);
    spinner = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 0);
    gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, TRUE, 2);
    gtk_widget_show(spinner);

    mdh_tooltips_set_tip(tool_tips, spinner,
        "Adjust the toolbars vertical location");

    g_signal_connect(G_OBJECT(adj), "value_changed",
        G_CALLBACK(mdh_spinner_update_geom_y), spinner);

    vsep = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(hbox2), vsep, FALSE, FALSE, 5);
    gtk_widget_show(vsep);

    frame = mdh_frame_new_with_title("Font");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(hbox2), frame, TRUE, TRUE, 0);

    hbox = gtk_hbox_new(FALSE, 3);
    gtk_widget_show(hbox);

    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);

    gtk_container_add(GTK_CONTAINER(frame), hbox);

    button = gtk_button_new_with_label("Select");
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button,
        "Select a font for the main toolbar");

    g_signal_connect(G_OBJECT(button), "clicked",
        G_CALLBACK(mdh_select_font), NULL);

    entry = gtk_entry_new();
    gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    gtk_widget_show(entry);

    if(ISSET(Toolbar.font))
        gtk_entry_set_text(GTK_ENTRY(entry), Toolbar.font);

    p_font_w = entry;

    button = gtk_button_new_with_label("Reset");
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button, "Use the default (GTK) font");

    g_signal_connect(G_OBJECT(button), "clicked",
        G_CALLBACK(mdh_unselect_font), NULL);

    vsep = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(hbox2), vsep, FALSE, FALSE, 5);
    gtk_widget_show(vsep);

    frame = mdh_frame_new_with_title("Tooltips");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(hbox2), frame, FALSE, TRUE, 0);

    hbox = gtk_hbox_new(FALSE, 3);
    gtk_widget_show(hbox);

    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);

    gtk_container_add(GTK_CONTAINER(frame), hbox);

    button = gtk_check_button_new_with_label("Enable");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), Toolbar.tooltips);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button,
        "Enable tooltips for the main toolbar buttons");

    p_tooltips_w = button;

    vsep = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(hbox2), vsep, FALSE, FALSE, 5);
    gtk_widget_show(vsep);

    frame = mdh_frame_new_with_title("On Top");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(hbox2), frame, FALSE, TRUE, 0);

    hbox = gtk_hbox_new(FALSE, 3);
    gtk_widget_show(hbox);

    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);

    gtk_container_add(GTK_CONTAINER(frame), hbox);

    button = gtk_check_button_new_with_label("Enable");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), Toolbar.on_top);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button,
        "Try to keep the toolbar above other windows");

    p_on_top_w = button;

    hsep = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox3), hsep, TRUE, TRUE, 0);
    gtk_widget_show(hsep);

    frame = mdh_frame_new_with_title("Buttons");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(vbox3), frame, TRUE, TRUE, 0);

    hbox = gtk_hbox_new(TRUE, 3);
    gtk_widget_show(hbox);

    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);

    gtk_container_add(GTK_CONTAINER(frame), hbox);

    button = gtk_check_button_new_with_label("Menu");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), B_Menu.enabled);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button,
        "Display the Menu button (recommended)");

    B_Menu.cfg_w = button;

    button = gtk_check_button_new_with_label("Hide");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), B_Hide.enabled);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button, "Display the Hide button");

    B_Hide.cfg_w = button;

    button = gtk_check_button_new_with_label("Run");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), B_Run.enabled);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button, "Display the Run button");

    B_Run.cfg_w = button;

    button = gtk_check_button_new_with_label("Pad");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), B_Pad.enabled);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button, "Display the Scratch-Pad button");

    B_Pad.cfg_w = button;

    hsep = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox3), hsep, TRUE, TRUE, 0);
    gtk_widget_show(hsep);

    hbox = gtk_hbox_new(TRUE, 3);
    gtk_box_pack_start(GTK_BOX(vbox3), hbox, TRUE, TRUE, 0);
    gtk_widget_show(hbox);

    frame = mdh_frame_new_with_title("Run xterm");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(hbox), frame, TRUE, TRUE, 0);

    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);

    gtk_container_set_border_width(GTK_CONTAINER(vbox4), 5);

    gtk_container_add(GTK_CONTAINER(frame), vbox4);

    entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(entry), Toolbar.run_item);
    gtk_box_pack_start(GTK_BOX(vbox4), entry, FALSE, FALSE, 0);
    gtk_widget_show(entry);

    mdh_tooltips_set_tip(tool_tips, entry,
        "XTerm to use for Run... (-e assumed)");

    p_run_item_w = entry;

    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
        vbox3, gtk_label_new("General"));

    /* Time tab */

    vbox3 = gtk_vbox_new(FALSE, 3);
    gtk_widget_show(vbox3);

    gtk_container_set_border_width(GTK_CONTAINER(vbox3), 5);

    frame = mdh_frame_new_with_title("Toolbar: Time");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(vbox3), frame, TRUE, TRUE, 0);

    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);

    gtk_container_set_border_width(GTK_CONTAINER(vbox4), 5);

    gtk_container_add(GTK_CONTAINER(frame), vbox4);

    /* ugh! can't seem to get this to line up with the other Enable's */

    p_button = gtk_check_button_new_with_label("Enable");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p_button), P_Time.enabled);
    gtk_box_pack_start(GTK_BOX(vbox4), p_button, FALSE, FALSE, 10);
    gtk_widget_show(p_button);

    P_Time.cfg_w[CFG_TIME_ENABLED] = p_button;

    mdh_tooltips_set_tip(tool_tips, p_button, "Display the Time panel");

    hbox = gtk_hbox_new(TRUE, 5);
    gtk_box_pack_start(GTK_BOX(vbox4), hbox, TRUE, TRUE, 0);
    gtk_widget_show(hbox);

    vbox5 = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), vbox5, FALSE, TRUE, 0);
    gtk_widget_show(vbox5);

    label = gtk_label_new("strftime format:");
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox5), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(entry), P_Time.value);
    gtk_box_pack_start(GTK_BOX(vbox5), entry, FALSE, TRUE, 0);
    gtk_widget_show(entry);

    P_Time.cfg_w[CFG_TIME_VALUE] = entry;

    gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Time.enabled);

    mdh_tooltips_set_tip(tool_tips, entry,
        "strftime format to use (default: %m/%d - %l:%M %p)");

    vbox5 = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), vbox5, FALSE, TRUE, 0);
    gtk_widget_show(vbox5);

    label = gtk_label_new("Command:");
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox5), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(entry), P_Time.command);
    gtk_box_pack_start(GTK_BOX(vbox5), entry, FALSE, TRUE, 0);
    gtk_widget_show(entry);

    P_Time.cfg_w[CFG_TIME_COMMAND] = entry;

    gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Time.enabled);

    mdh_tooltips_set_tip(tool_tips, entry,
        "Command to execute when panel is clicked (default: evolution)");

    g_signal_connect(G_OBJECT(p_button), "clicked",
        G_CALLBACK(mdh_button_toggle_panel), &P_Time);

    label = gtk_label_new("Displays the time as defined by above strftime(3) format.");
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_widget_set_usize(GTK_WIDGET(label), 450, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox3), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
        vbox3, gtk_label_new("Time"));

    /* Mail tab */

    vbox3 = gtk_vbox_new(FALSE, 3);
    gtk_widget_show(vbox3);

    gtk_container_set_border_width(GTK_CONTAINER(vbox3), 5);

    frame = mdh_frame_new_with_title("Toolbar: Mail");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(vbox3), frame, TRUE, TRUE, 0);

    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);

    gtk_container_set_border_width(GTK_CONTAINER(vbox4), 5);

    gtk_container_add(GTK_CONTAINER(frame), vbox4);

    for(i = 0; i <= _MDH_MAX_MAIL; i++) {
        hbox = gtk_hbox_new(FALSE, 10);
        gtk_box_pack_start(GTK_BOX(vbox4), hbox, TRUE, TRUE, 3);
        gtk_widget_show(hbox);

        p_button = gtk_check_button_new_with_label("Enable");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p_button),
            P_Mail[i].enabled);
        gtk_box_pack_start(GTK_BOX(hbox), p_button, FALSE, FALSE, 0);
        gtk_widget_show(p_button);

        P_Mail[i].cfg_w[CFG_MAIL_ENABLED] = p_button;

        mdh_tooltips_set_tip(tool_tips, p_button, "Display the Mail panel");

        label = gtk_label_new(" Mailbox/dir:");
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

        entry = gtk_entry_new();
        gtk_entry_set_text(GTK_ENTRY(entry), P_Mail[i].value);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        gtk_widget_show(entry);

        P_Mail[i].cfg_w[CFG_MAIL_VALUE] = entry;

        gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Mail[i].enabled);

        mdh_tooltips_set_tip(tool_tips, entry,
            "Mailbox/dir to process (default $MAIL, $MAILDIR)");

        label = gtk_label_new(" Command:");
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

        entry = gtk_entry_new();
        gtk_entry_set_text(GTK_ENTRY(entry), P_Mail[i].command);

        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        gtk_widget_show(entry);

        P_Mail[i].cfg_w[CFG_MAIL_COMMAND] = entry;

        gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Mail[i].enabled);

        mdh_tooltips_set_tip(tool_tips, entry,
            "Command to execute when panel is clicked (default: balsa)");

        adj = gtk_adjustment_new(P_Mail[i].interval, _MAIL_I_MIN, _MAIL_I_MAX,
                                 5.0, 50.0, 0.0);
        spinner = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 0);
        gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), FALSE);
        gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, FALSE, 0);
        gtk_widget_show(spinner);

        P_Mail[i].cfg_w[CFG_MAIL_SPINNER] = spinner;

        gtk_widget_set_sensitive(GTK_WIDGET(spinner), P_Mail[i].enabled);

        mdh_tooltips_set_tip(tool_tips, spinner,
            "Set the update interval (default: 30)");

        label = gtk_label_new("seconds");
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

        g_signal_connect(G_OBJECT(p_button), "clicked",
            G_CALLBACK(mdh_button_toggle_panel), &P_Mail[i]);
    }

    hbox = gtk_hbox_new(FALSE, 10);
    gtk_box_pack_start(GTK_BOX(vbox4), hbox, TRUE, TRUE, 3);
    gtk_widget_show(hbox);

    button = gtk_check_button_new_with_label("Show Mailbox/dir Name");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), P_Mail_opt.labels);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button,
        "Display the mailbox/dir name along with count");

    p_mail_labels_w = button;

    button = gtk_check_button_new_with_label("Bold On Unread");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),P_Mail_opt.use_bold);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    p_mail_use_bold_w = button;

    mdh_tooltips_set_tip(tool_tips, button, "Bold on unread mail");

    label = gtk_label_new("Displays read/total count of above mailbox/dir(s).");
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_widget_set_usize(GTK_WIDGET(label), 450, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox3), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
        vbox3, gtk_label_new("Mail"));

#if _MDH_HAS_CPU
    /* Cpu tab */

    vbox3 = gtk_vbox_new(FALSE, 3);
    gtk_widget_show(vbox3);

    gtk_container_set_border_width(GTK_CONTAINER(vbox3), 5);

    frame = mdh_frame_new_with_title("Toolbar: Cpu");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(vbox3), frame, TRUE, TRUE, 0);

    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);

    gtk_container_set_border_width(GTK_CONTAINER(vbox4), 5);

    gtk_container_add(GTK_CONTAINER(frame), vbox4);

    hbox = gtk_hbox_new(FALSE, 10);
    gtk_box_pack_start(GTK_BOX(vbox4), hbox, FALSE, FALSE, 10);
    gtk_widget_show(hbox);

    p_button = gtk_check_button_new_with_label("Enable");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p_button), P_Cpu.enabled);
    gtk_box_pack_start(GTK_BOX(hbox), p_button, FALSE, FALSE, 0);
    gtk_widget_show(p_button);

    P_Cpu.cfg_w[CFG_CPU_ENABLED] = p_button;

    mdh_tooltips_set_tip(tool_tips, p_button, "Display the Cpu panel");

    adj = gtk_adjustment_new(P_Cpu.interval, 1.0, 30.0, 1.0, 5.0, 0.0);
    spinner = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 0);
    gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, TRUE, 0);
    gtk_widget_show(spinner);

    P_Cpu.cfg_w[CFG_CPU_SPINNER] = spinner;

    gtk_widget_set_sensitive(GTK_WIDGET(spinner), P_Cpu.enabled);

    mdh_tooltips_set_tip(tool_tips, spinner,
        "Set the update interval (default: 3)");

    label = gtk_label_new("second(s)");
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    label = gtk_label_new("Command:");
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox4), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(entry), P_Cpu.command);
    gtk_box_pack_start(GTK_BOX(vbox4), entry, FALSE, TRUE, 0);
    gtk_widget_show(entry);

    P_Cpu.cfg_w[CFG_CPU_COMMAND] = entry;

    gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Cpu.enabled);

    mdh_tooltips_set_tip(tool_tips, entry,
        "Command to execute when panel is clicked (default: xterm -e top)");

    button = gtk_check_button_new_with_label("Bold On High Load");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), P_Cpu_opt.use_bold);
    gtk_box_pack_start(GTK_BOX(vbox4), button, FALSE, FALSE, 10);
    gtk_widget_show(button);

    gtk_widget_set_sensitive(GTK_WIDGET(button), P_Cpu.enabled);

    mdh_tooltips_set_tip(tool_tips, button, "Bold when load reaches 90%");

    P_Cpu.cfg_w[CFG_CPU_BOLD] = button;

    g_signal_connect(G_OBJECT(p_button), "clicked",
        G_CALLBACK(mdh_button_toggle_panel), &P_Cpu);

    label = gtk_label_new("Displays cpu load percentage over last 60 seconds.");
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_widget_set_usize(GTK_WIDGET(label), 450, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox3), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
        vbox3, gtk_label_new("Cpu"));
#endif

#if _MDH_HAS_MEM
    /* Mem tab */

    vbox3 = gtk_vbox_new(FALSE, 3);
    gtk_widget_show(vbox3);

    gtk_container_set_border_width(GTK_CONTAINER(vbox3), 5);

    frame = mdh_frame_new_with_title("Toolbar: Mem");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(vbox3), frame, TRUE, TRUE, 0);

    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);

    gtk_container_set_border_width(GTK_CONTAINER(vbox4), 5);

    gtk_container_add(GTK_CONTAINER(frame), vbox4);

    hbox = gtk_hbox_new(FALSE, 10);
    gtk_box_pack_start(GTK_BOX(vbox4), hbox, FALSE, FALSE, 10);
    gtk_widget_show(hbox);

    p_button = gtk_check_button_new_with_label("Enable");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p_button), P_Mem.enabled);
    gtk_box_pack_start(GTK_BOX(hbox), p_button, FALSE, FALSE, 0);
    gtk_widget_show(p_button);

    P_Mem.cfg_w[CFG_MEM_ENABLED] = p_button;

    mdh_tooltips_set_tip(tool_tips, p_button, "Display the Mem panel");

    adj = gtk_adjustment_new(P_Mem.interval, 1.0, 30.0, 1.0, 5.0, 0.0);
    spinner = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 0);
    gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, TRUE, 0);
    gtk_widget_show(spinner);

    P_Mem.cfg_w[CFG_MEM_SPINNER] = spinner;

    gtk_widget_set_sensitive(GTK_WIDGET(spinner), P_Mem.enabled);

    mdh_tooltips_set_tip(tool_tips, spinner,
        "Set the update interval (default: 15)");

    label = gtk_label_new("second(s)");
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    label = gtk_label_new("Command:");
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox4), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(entry), P_Mem.command);
    gtk_box_pack_start(GTK_BOX(vbox4), entry, FALSE, TRUE, 0);
    gtk_widget_show(entry);

    P_Mem.cfg_w[CFG_MEM_COMMAND] = entry;

    gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Mem.enabled);

    mdh_tooltips_set_tip(tool_tips, entry,
        "Command to execute when panel is clicked (default: none)");

    button = gtk_check_button_new_with_label("Bold On High Usage");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), P_Mem_opt.use_bold);
    gtk_box_pack_start(GTK_BOX(vbox4), button, FALSE, FALSE, 10);
    gtk_widget_show(button);

    P_Mem.cfg_w[CFG_MEM_BOLD] = button;

    gtk_widget_set_sensitive(GTK_WIDGET(button), P_Mem.enabled);

    mdh_tooltips_set_tip(tool_tips, button, "Bold when usage reaches 85%");

    g_signal_connect(G_OBJECT(p_button), "clicked",
        G_CALLBACK(mdh_button_toggle_panel), &P_Mem);

    label = gtk_label_new("Displays amount (MB) of system memory used.");
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_widget_set_usize(GTK_WIDGET(label), 450, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox3), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
        vbox3, gtk_label_new("Mem"));
#endif

#if _MDH_HAS_DISK
    /* Disk tab */

    vbox3 = gtk_vbox_new(FALSE, 3);
    gtk_widget_show(vbox3);

    gtk_container_set_border_width(GTK_CONTAINER(vbox3), 5);

    frame = mdh_frame_new_with_title("Toolbar: Disk");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(vbox3), frame, TRUE, TRUE, 0);

    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);

    gtk_container_set_border_width(GTK_CONTAINER(vbox4), 5);

    gtk_container_add(GTK_CONTAINER(frame), vbox4);

    for(i = 0; i <= _MDH_MAX_DISK; i++) {
        hbox = gtk_hbox_new(FALSE, 10);
        gtk_box_pack_start(GTK_BOX(vbox4), hbox, TRUE, TRUE, 3);
        gtk_widget_show(hbox);

        p_button = gtk_check_button_new_with_label("Enable");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p_button),
            P_Disk[i].enabled);
        gtk_box_pack_start(GTK_BOX(hbox), p_button, FALSE, FALSE, 0);
        gtk_widget_show(p_button);

        P_Disk[i].cfg_w[CFG_DISK_ENABLED] = p_button;

        mdh_tooltips_set_tip(tool_tips, p_button, "Display the Disk panel");

        label = gtk_label_new(" FS:");
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

        entry = gtk_entry_new();
        gtk_entry_set_text(GTK_ENTRY(entry), P_Disk[i].value);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        gtk_widget_show(entry);

        P_Disk[i].cfg_w[CFG_DISK_VALUE] = entry;

        gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Disk[i].enabled);

        mdh_tooltips_set_tip(tool_tips, entry, "Filesystem to monitor");

        label = gtk_label_new(" Command:");
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

        entry = gtk_entry_new();
        gtk_entry_set_text(GTK_ENTRY(entry), P_Disk[i].command);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        gtk_widget_show(entry);

        P_Disk[i].cfg_w[CFG_DISK_COMMAND] = entry;

        gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Disk[i].enabled);

        mdh_tooltips_set_tip(tool_tips, entry,
            "Command to execute when panel is clicked (default: gmc)");

        g_signal_connect(G_OBJECT(p_button), "clicked",
            G_CALLBACK(mdh_button_toggle_panel), &P_Disk[i]);
    }

    hbox = gtk_hbox_new(FALSE, 10);
    gtk_box_pack_start(GTK_BOX(vbox4), hbox, TRUE, TRUE, 3);
    gtk_widget_show(hbox);

    button = gtk_check_button_new_with_label("Show Disk Name");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), P_Disk_opt.labels);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    p_disk_labels_w = button;

    mdh_tooltips_set_tip(tool_tips, button,
        "Display the disk name along with usage");

    button = gtk_check_button_new_with_label("Bold On High Usage");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),P_Disk_opt.use_bold);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button, "Bold when usage reaches 85%");

    p_disk_use_bold_w = button;

    label = gtk_label_new("Displays amount of disk space used.");
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_widget_set_usize(GTK_WIDGET(label), 450, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox3), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
        vbox3, gtk_label_new("Disk"));
#endif

#if _MDH_HAS_NET
    /* Net tab */

    vbox3 = gtk_vbox_new(FALSE, 3);
    gtk_widget_show(vbox3);

    gtk_container_set_border_width(GTK_CONTAINER(vbox3), 5);

    frame = mdh_frame_new_with_title("Toolbar: Net");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(vbox3), frame, TRUE, TRUE, 0);

    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);

    gtk_container_set_border_width(GTK_CONTAINER(vbox4), 5);

    gtk_container_add(GTK_CONTAINER(frame), vbox4);

    for(i = 0; i <= _MDH_MAX_NET; i++) {
        hbox = gtk_hbox_new(FALSE, 10);
        gtk_box_pack_start(GTK_BOX(vbox4), hbox, TRUE, TRUE, 3);
        gtk_widget_show(hbox);

        p_button = gtk_check_button_new_with_label("Enable");
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p_button),
            P_Net[i].enabled);
        gtk_box_pack_start(GTK_BOX(hbox), p_button, FALSE, FALSE, 0);
        gtk_widget_show(p_button);

        P_Net[i].cfg_w[CFG_NET_ENABLED] = p_button;

        mdh_tooltips_set_tip(tool_tips, p_button, "Display the Net panel");

        label = gtk_label_new(" Interface:");
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

        entry = gtk_entry_new();
#ifndef __sun__
        gtk_entry_set_text(GTK_ENTRY(entry), P_Net[i].value);
#endif
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        gtk_widget_show(entry);

        P_Net[i].cfg_w[CFG_NET_VALUE] = entry;

#ifdef __sun__
        gtk_widget_set_sensitive(GTK_WIDGET(entry), TRUE);
#else
        gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Net[i].enabled);
#endif

        mdh_tooltips_set_tip(tool_tips, entry, "Interface to monitor");

        label = gtk_label_new(" Command:");
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

        entry = gtk_entry_new();
        gtk_entry_set_text(GTK_ENTRY(entry), P_Net[i].command);
        gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
        gtk_widget_show(entry);

        P_Net[i].cfg_w[CFG_NET_COMMAND] = entry;

        gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Net[i].enabled);

        mdh_tooltips_set_tip(tool_tips, entry,
            "Command to execute when panel is clicked (default: ethereal)");

        adj = gtk_adjustment_new(P_Net[i].interval, 1.0, 30.0, 1.0, 5.0, 0.0);
        spinner = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 0);
        gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), FALSE);
        gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, TRUE, 0);
        gtk_widget_show(spinner);

        P_Net[i].cfg_w[CFG_NET_SPINNER] = spinner;

        gtk_widget_set_sensitive(GTK_WIDGET(spinner), P_Net[i].enabled);

        mdh_tooltips_set_tip(tool_tips, spinner,
            "Set the update interval (default: 3)");

        label = gtk_label_new("second(s)");
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
        gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
        gtk_widget_show(label);

        g_signal_connect(G_OBJECT(p_button), "clicked",
            G_CALLBACK(mdh_button_toggle_panel), &P_Net[i]);
    }

    hbox = gtk_hbox_new(FALSE, 10);
    gtk_box_pack_start(GTK_BOX(vbox4), hbox, TRUE, TRUE, 3);
    gtk_widget_show(hbox);

    button = gtk_check_button_new_with_label("Show Interface Name");
#ifdef __sun__
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), FALSE);
#else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), P_Net_opt.labels);
#endif
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

#ifndef __sun__
    p_net_labels_w = button;
#endif

    mdh_tooltips_set_tip(tool_tips, button,
        "Display the interface name along with usage");

    label = gtk_label_new("Displays utilization (down/up) for above interface(s).");
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_widget_set_usize(GTK_WIDGET(label), 450, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox3), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
        vbox3, gtk_label_new("Net"));
#endif

    /* Temp tab */

    vbox3 = gtk_vbox_new(FALSE, 3);
    gtk_widget_show(vbox3);

    gtk_container_set_border_width(GTK_CONTAINER(vbox3), 5);

    frame = mdh_frame_new_with_title("Toolbar: Temp");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(vbox3), frame, TRUE, TRUE, 0);

    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);

    gtk_container_set_border_width(GTK_CONTAINER(vbox4), 5);

    gtk_container_add(GTK_CONTAINER(frame), vbox4);

    hbox = gtk_hbox_new(FALSE, 10);
    gtk_box_pack_start(GTK_BOX(vbox4), hbox, FALSE, FALSE, 10);
    gtk_widget_show(hbox);

    p_button = gtk_check_button_new_with_label("Enable");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p_button), P_Temp.enabled);
    gtk_box_pack_start(GTK_BOX(hbox), p_button, FALSE, FALSE, 0);
    gtk_widget_show(p_button);

    P_Temp.cfg_w[CFG_TEMP_ENABLED] = p_button;

    mdh_tooltips_set_tip(tool_tips, p_button, "Display the Temp panel");

    adj = gtk_adjustment_new(P_Temp.interval / 3600, _WEATHER_MIN / 3600,
                             _WEATHER_MAX / 3600, 1.0, 3.0, 0.0);
    spinner = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 0);
    gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, TRUE, 0);
    gtk_widget_show(spinner);

    P_Temp.cfg_w[CFG_TEMP_SPINNER] = spinner;

    gtk_widget_set_sensitive(GTK_WIDGET(spinner), P_Temp.enabled);

    mdh_tooltips_set_tip(tool_tips, spinner,
        "Set the update interval (default: 3 hours)");

    label = gtk_label_new("hour(s)");
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    hbox = gtk_hbox_new(TRUE, 5);
    gtk_box_pack_start(GTK_BOX(vbox4), hbox, TRUE, TRUE, 0);
    gtk_widget_show(hbox);

    vbox5 = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), vbox5, FALSE, TRUE, 0);
    gtk_widget_show(vbox5);

    label = gtk_label_new("Station-ID (http://www.nws.noaa.gov/tg/siteloc.shtml):");
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox5), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    entry = gtk_entry_new();
    gtk_entry_set_max_length(GTK_ENTRY(entry), _WEATHER_LEN);
    gtk_entry_set_text(GTK_ENTRY(entry), P_Temp.value);
    gtk_box_pack_start(GTK_BOX(vbox5), entry, FALSE, FALSE, 0);
    gtk_widget_show(entry);

    P_Temp.cfg_w[CFG_TEMP_VALUE] = entry;

    gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Temp.enabled);

    mdh_tooltips_set_tip(tool_tips, entry,
        "Station-ID to use (default: KGRB -- Green Bay, WI)");

    vbox5 = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), vbox5, FALSE, TRUE, 0);
    gtk_widget_show(vbox5);

    label = gtk_label_new("Command:");
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox5), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(entry), P_Temp.command);
    gtk_box_pack_start(GTK_BOX(vbox5), entry, FALSE, TRUE, 0);
    gtk_widget_show(entry);

    P_Temp.cfg_w[CFG_TEMP_COMMAND] = entry;

    gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Temp.enabled);

    mdh_tooltips_set_tip(tool_tips, entry,
        "Command to execute when panel is clicked (default: mozilla http://www.weather.com/)");

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(vbox4), hbox, FALSE, TRUE, 10);
    gtk_widget_show(hbox);

    p_temp_proxy_enable = P_Temp_opt.proxy_enable;

    button = gtk_check_button_new_with_label("Display in Celcius");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), P_Temp_opt.celcius);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_widget_show(button);

    P_Temp.cfg_w[CFG_TEMP_CELCIUS] = button;

    gtk_widget_set_sensitive(GTK_WIDGET(button), P_Temp.enabled);

    mdh_tooltips_set_tip(tool_tips, button,
        "Display in Celcius instead of Ferenheight");

    button = gtk_check_button_new_with_label("Enable Proxy");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
        P_Temp_opt.proxy_enable);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 10);
    gtk_widget_show(button);

    P_Temp.cfg_w[CFG_TEMP_PROXY_ENABLE] = button;

    gtk_widget_set_sensitive(GTK_WIDGET(button), P_Temp.enabled);

    mdh_tooltips_set_tip(tool_tips, button, "Enable proxy support");

    g_signal_connect(G_OBJECT(button), "clicked",
        G_CALLBACK(mdh_button_toggle_temp_proxy_value), NULL);

    label = gtk_label_new("Host: ");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(entry), P_Temp_opt.proxy_host);
    gtk_box_pack_start(GTK_BOX(hbox), entry, FALSE, FALSE, 0);
    gtk_widget_show(entry);

    P_Temp.cfg_w[CFG_TEMP_PROXY_HOST] = entry;

    if(P_Temp.enabled == FALSE)
        gtk_widget_set_sensitive(GTK_WIDGET(entry), FALSE);
    else
        gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Temp_opt.proxy_enable);

    mdh_tooltips_set_tip(tool_tips, entry,
        "Set the proxy host (default: localhost)");

    label = gtk_label_new(" Port: ");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    adj = gtk_adjustment_new(P_Temp_opt.proxy_port,
              _TCP_PORT_MIN, _TCP_PORT_MAX, 1.0, 25.0, 0.0);
    spinner = gtk_spin_button_new(GTK_ADJUSTMENT(adj), 0, 0);
    gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinner), FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), spinner, FALSE, TRUE, 0);
    gtk_widget_show(spinner);

    P_Temp.cfg_w[CFG_TEMP_PROXY_PORT] = spinner;

    if(P_Temp.enabled == FALSE)
        gtk_widget_set_sensitive(GTK_WIDGET(spinner), FALSE);
    else
        gtk_widget_set_sensitive(GTK_WIDGET(spinner), P_Temp_opt.proxy_enable);

    mdh_tooltips_set_tip(tool_tips, spinner,
        "Set the proxy port (default: 8080)");

    g_signal_connect(G_OBJECT(p_button), "clicked",
        G_CALLBACK(mdh_button_toggle_panel_temp), &P_Temp);

    label = gtk_label_new("Displays temperature for above station-id. Courtesy of the NWS.");
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_widget_set_usize(GTK_WIDGET(label), 450, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox3), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
        vbox3, gtk_label_new("Temp"));

#if _MDH_HAS_UPTIME
    /* Uptime tab */

    vbox3 = gtk_vbox_new(FALSE, 3);
    gtk_widget_show(vbox3);

    gtk_container_set_border_width(GTK_CONTAINER(vbox3), 5);

    frame = mdh_frame_new_with_title("Toolbar: Uptime");
    gtk_widget_show(frame);

    gtk_box_pack_start(GTK_BOX(vbox3), frame, TRUE, TRUE, 0);

    vbox4 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox4);

    gtk_container_set_border_width(GTK_CONTAINER(vbox4), 5);

    gtk_container_add(GTK_CONTAINER(frame), vbox4);

    p_button = gtk_check_button_new_with_label("Enable");
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(p_button),
                                 P_Uptime.enabled);
    gtk_box_pack_start(GTK_BOX(vbox4), p_button, FALSE, FALSE, 10);
    gtk_widget_show(p_button);

    P_Uptime.cfg_w[CFG_UPTIME_ENABLED] = p_button;

    mdh_tooltips_set_tip(tool_tips, p_button, "Display the Uptime panel");

    label = gtk_label_new("Command:");
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox4), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    entry = gtk_entry_new();
    gtk_entry_set_text(GTK_ENTRY(entry), P_Uptime.command);
    gtk_box_pack_start(GTK_BOX(vbox4), entry, FALSE, TRUE, 0);
    gtk_widget_show(entry);

    P_Uptime.cfg_w[CFG_UPTIME_COMMAND] = entry;

    gtk_widget_set_sensitive(GTK_WIDGET(entry), P_Uptime.enabled);

    mdh_tooltips_set_tip(tool_tips, entry,
        "Command to execute when panel is clicked (default: none)");

    g_signal_connect(G_OBJECT(p_button), "clicked",
        G_CALLBACK(mdh_button_toggle_panel), &P_Uptime);

    label = gtk_label_new("Displays system uptime.");
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_widget_set_usize(GTK_WIDGET(label), 450, 0);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox3), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
        vbox3, gtk_label_new("Uptime"));
#endif

    hbox = gtk_hbox_new(TRUE, 3);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
    gtk_widget_show(hbox);

    button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_widget_show(button);

    g_signal_connect(G_OBJECT(button), "clicked",
        G_CALLBACK(mdh_window_configuration_cancel), window);

    mdh_tooltips_set_tip(tool_tips, button,
        "Cancel new (or unapplied) panel settings");

    button = gtk_button_new_from_stock(GTK_STOCK_APPLY);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_widget_show(button);

    mdh_tooltips_set_tip(tool_tips, button, "Apply new panel settings");

    g_signal_connect(G_OBJECT(button), "clicked",
        G_CALLBACK(mdh_window_configuration_apply), NULL);

    button = gtk_button_new_from_stock(GTK_STOCK_OK);
    gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
    gtk_widget_show(button);

    g_signal_connect(G_OBJECT(button), "clicked",
        G_CALLBACK(mdh_window_configuration_close), window);

    mdh_tooltips_set_tip(tool_tips, button,
        "Close and Apply new panel settings");

    gtk_widget_show(window);
}

/* if panel->value is NOT NULL and zero-length, return the name of the panel */
static gchar *mdh_toolbar_panels_have_empty_value(void)
{
    GSList *p;

    for(p = Toolbar.panels; p != NULL; p = g_slist_next(p)) {
        struct cfg_common cfg;

        mdh_toolbar_get_cfg_common(p->data, &cfg);

        if(cfg.enabled == TRUE && cfg.value != NULL && !strlen(cfg.value))
            return(PANEL(p->data)->name);
    }

    return(NULL);
}

static gboolean mdh_window_configuration_apply(void)
{
    gchar *p;

    struct cfg_common cfg;

    /* make sure the enabled panels have a value */
    if((p = mdh_toolbar_panels_have_empty_value()) != NULL) {
        gchar *text = g_strdup_printf("The %s panel is missing it's value.", p);

        mdh_window_notice(text, "Every enabled panel that requires an argument to function must have a value (eg. strftime format, mailbox, network interface). Disable the panel if you do not wish to use it.");

        g_free(text);

        return(FALSE);
    }

    mdh_toolbar_get_cfg_common(&P_Temp, &cfg);

    /* check station-id length */
    if(P_Temp.enabled == TRUE && STRLEN(cfg.value) != _WEATHER_LEN) {
        mdh_window_notice("Invalid Station-ID.",
                          "Station-ID's should be 4 characters in length.");
        return(FALSE);
    }

    mdh_toolbar_get_new();

    mdh_toolbar_reset();

    return(TRUE);
}

static void mdh_window_configuration_close_common(GtkWidget *window)
{
    mdh_toolbar_window_toggle_open(&Toolbar.config_open);

    gtk_widget_destroy(GTK_WIDGET(window));
}

static void mdh_window_configuration_close(GtkWidget *widget,
                                           GtkWidget *window)
{
    if(mdh_window_configuration_apply() == FALSE)
        return;

    mdh_window_configuration_close_common(window);
}

static void mdh_window_configuration_cancel(GtkWidget *widget,
                                            GtkWidget *window)
{
    Toolbar.geom_x = geom_x;
    Toolbar.geom_y = geom_y;

    mdh_toolbar_set_location();

    mdh_window_configuration_close_common(window);
}

static void mdh_spinner_update_geom_x(GtkWidget *widget, GtkWidget *spinner)
{
    Toolbar.geom_x = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinner));

    mdh_toolbar_set_location();
}

static void mdh_spinner_update_geom_y(GtkWidget *widget, GtkWidget *spinner)
{
    Toolbar.geom_y = gtk_spin_button_get_value(GTK_SPIN_BUTTON(spinner));

    mdh_toolbar_set_location();
}

static void mdh_button_toggle_temp_proxy_value(GtkWidget *widget,
                                               gpointer data)
{
    gboolean enabled = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));

    p_temp_proxy_enable = enabled;

    gtk_widget_set_sensitive(P_Temp.cfg_w[CFG_TEMP_PROXY_HOST], enabled);
    gtk_widget_set_sensitive(P_Temp.cfg_w[CFG_TEMP_PROXY_PORT], enabled);
}

/* configuration window: toggle button */
static void mdh_button_toggle_panel(GtkWidget *widget, struct panel_t *obj)
{
    gint i;

    gboolean enabled = gtk_toggle_button_get_active(
                           GTK_TOGGLE_BUTTON(obj->cfg_w[0]));

    for(i = 1; obj->cfg_w[i] != NULL; i++)
        gtk_widget_set_sensitive(obj->cfg_w[i], enabled);
}

static void mdh_button_toggle_panel_temp(GtkWidget *widget,
                                         struct panel_t *obj)
{
    gint i;

    gboolean enabled = gtk_toggle_button_get_active(
                           GTK_TOGGLE_BUTTON(obj->cfg_w[CFG_TEMP_ENABLED]));

    for(i = 1; i <= CFG_TEMP_PROXY_ENABLE; i++)
        gtk_widget_set_sensitive(obj->cfg_w[i], enabled);

    if(enabled == FALSE) {
        gtk_widget_set_sensitive(obj->cfg_w[CFG_TEMP_PROXY_HOST], enabled);
        gtk_widget_set_sensitive(obj->cfg_w[CFG_TEMP_PROXY_PORT], enabled);
    } else {
        gtk_widget_set_sensitive(obj->cfg_w[CFG_TEMP_PROXY_HOST],
                                 p_temp_proxy_enable);
        gtk_widget_set_sensitive(obj->cfg_w[CFG_TEMP_PROXY_PORT],
                                 p_temp_proxy_enable);
    }
}

/* configuration window: toggle button */
void mdh_button_toggle_value(GtkWidget *widget, gboolean *enabled)
{
    *enabled = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
}

static void mdh_font_selection_ok(GtkWidget *widget, GtkWidget *window)
{
    gchar *font = gtk_font_selection_dialog_get_font_name(
                      GTK_FONT_SELECTION_DIALOG(window));

    if(font == NULL) {
        gtk_widget_destroy(window);
        return;
    }

    gtk_entry_set_text(GTK_ENTRY(p_font_w), font);

    g_free(font);

    gtk_widget_destroy(window);
}

#define PREVIEW_TEXT "abcdefghijk ABCDEFGHIJK 123456789 - Goat!"

static void mdh_select_font(void)
{
    GtkWidget *window;

    const gchar *font;

    window = gtk_font_selection_dialog_new("Font Selection Dialog");

    /* default is just [a-k][A-K] */
    gtk_font_selection_dialog_set_preview_text(
        GTK_FONT_SELECTION_DIALOG(window), PREVIEW_TEXT);

    font = gtk_entry_get_text(GTK_ENTRY(p_font_w));

    /* select current font */
    if(STRLEN(font))
        gtk_font_selection_dialog_set_font_name(
            GTK_FONT_SELECTION_DIALOG(window), font);

    g_signal_connect(GTK_FONT_SELECTION_DIALOG(window)->ok_button,
        "clicked", G_CALLBACK(mdh_font_selection_ok), window);

    g_signal_connect_swapped(GTK_FONT_SELECTION_DIALOG(window)->cancel_button,
        "clicked", G_CALLBACK(gtk_widget_destroy), window);

    gtk_widget_show(window);
}

static void mdh_unselect_font(void)
{
    gtk_entry_set_text(GTK_ENTRY(p_font_w), "");
}

static GtkWidget *mdh_frame_new_with_title(const gchar *title)
{
    GtkWidget *frame;
    GtkWidget *label;

    frame = gtk_frame_new(NULL);
    gtk_frame_set_label_align(GTK_FRAME(frame), 0.0, 0.5);

    label = gtk_label_new(NULL);
    mdh_label_set_larger(GTK_LABEL(label), title);
    gtk_widget_show(label);

    gtk_frame_set_label_widget(GTK_FRAME(frame), label);

    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);

    return(frame);
}

static void mdh_toolbar_get_new(void)
{
    gint i;

    Toolbar.tooltips      = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(p_tooltips_w));

    Toolbar.on_top        = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(p_on_top_w));

    g_free(Toolbar.run_item);

    Toolbar.run_item      = g_strdup(gtk_entry_get_text(
        GTK_ENTRY(p_run_item_w)));

    g_free(Toolbar.font);

    Toolbar.font          = g_strdup(gtk_entry_get_text(GTK_ENTRY(p_font_w)));

    g_slist_foreach(Toolbar.buttons, (GFunc) mdh_toolbar_button_get_new, NULL);
    g_slist_foreach(Toolbar.panels,  (GFunc) mdh_toolbar_panel_get_new, NULL);

    P_Mail_opt.labels   = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(p_mail_labels_w));
    P_Mail_opt.use_bold = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(p_mail_use_bold_w));

    for(i = 0; i <= _MDH_MAX_MAIL; i++) {
        g_free(P_Mail[i].tooltip);
        P_Mail[i].tooltip = g_strdup_printf("Mail: %s", P_Mail[i].value);
    }

#ifdef _MDH_HAS_CPU
    P_Cpu_opt.use_bold  = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(P_Cpu.cfg_w[CFG_CPU_BOLD]));
#endif

#ifdef _MDH_HAS_MEM
    P_Mem_opt.use_bold  = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(P_Mem.cfg_w[CFG_MEM_BOLD]));
#endif

#ifdef _MDH_HAS_DISK
    P_Disk_opt.labels    = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(p_disk_labels_w));
    P_Disk_opt.use_bold  = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(p_disk_use_bold_w));

    for(i = 0; i <= _MDH_MAX_DISK; i++) {
        g_free(P_Disk[i].tooltip);
        P_Disk[i].tooltip = g_strdup_printf("Disk: %s", P_Disk[i].value);
    }
#endif

#ifdef _MDH_HAS_NET
# ifndef __sun__
    P_Net_opt.labels     = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(p_net_labels_w));

    for(i = 0; i <= _MDH_MAX_NET; i++) {
        g_free(P_Net[i].tooltip);
        P_Net[i].tooltip = g_strdup_printf("Net: %s", P_Net[i].value);
    }
# endif
#endif

    P_Temp_opt.celcius       = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(P_Temp.cfg_w[CFG_TEMP_CELCIUS]));

    P_Temp_opt.proxy_enable  = gtk_toggle_button_get_active(
        GTK_TOGGLE_BUTTON(P_Temp.cfg_w[CFG_TEMP_PROXY_ENABLE]));

    g_free(P_Temp_opt.proxy_host);
    
    P_Temp_opt.proxy_host    = g_strdup(gtk_entry_get_text(
        GTK_ENTRY(P_Temp.cfg_w[CFG_TEMP_PROXY_HOST])));
    P_Temp_opt.proxy_port    = gtk_spin_button_get_value(
        GTK_SPIN_BUTTON(P_Temp.cfg_w[CFG_TEMP_PROXY_PORT]));

    g_free(P_Temp.tooltip);

    P_Temp.tooltip = g_strdup_printf("Temp: %s", P_Temp.value);
}

static void mdh_toolbar_button_get_new(struct button_t *obj)
{
    obj->enabled = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(obj->cfg_w));
}

static void mdh_toolbar_panel_get_new(struct panel_t *obj)
{
    gboolean changed = FALSE;

    struct cfg_common cfg;

    mdh_toolbar_get_cfg_common(obj, &cfg);

    if(obj->enabled != cfg.enabled)
        changed = TRUE;

    obj->enabled = cfg.enabled;

    if(obj->value != NULL && cfg.value != NULL && !STREQ(obj->value,cfg.value)){
        changed = TRUE;

        g_free(obj->value);

        if(obj->type == _PANEL_TYPE_MAIL)
            obj->value = mdh_expand_tilde(cfg.value);
        else if(obj->type == _PANEL_TYPE_TEMP)
            obj->value = STRUP(cfg.value);
        else
            obj->value = g_strdup(g_strstrip((gchar *) cfg.value));
    }

    g_free(obj->command);

    obj->command = g_strdup(cfg.command);

    if(obj->interval != cfg.interval)
        changed = TRUE;

    if(cfg.interval > 0)
        obj->interval = cfg.interval;

    /* obj->init_close(obj); */

    if(obj->enabled == TRUE && changed == TRUE) {
        mdh_toolbar_panel_stop(obj);
        mdh_toolbar_panel_start(obj);
    }
}

static void mdh_toolbar_get_cfg_common(struct panel_t *obj,
                                       struct cfg_common *cfg)
{
    GtkWidget *e, *v, *c, *i;

    memset(cfg, '\0', sizeof(struct cfg_common));

    cfg->enabled  = FALSE;
    cfg->interval = -1;

    g_return_if_fail(obj != NULL);

    e = obj->cfg_w[0];
    v = NULL;
    c = NULL;
    i = NULL;

    switch(obj->type) {
        case _PANEL_TYPE_TIME:
            v = obj->cfg_w[CFG_TIME_VALUE];
            c = obj->cfg_w[CFG_TIME_COMMAND];
            break;
        case _PANEL_TYPE_MAIL:
            v = obj->cfg_w[CFG_MAIL_VALUE];
            c = obj->cfg_w[CFG_MAIL_COMMAND];
            i = obj->cfg_w[CFG_MAIL_SPINNER];
            break;
#if _MDH_HAS_CPU
        case _PANEL_TYPE_CPU:
            c = obj->cfg_w[CFG_CPU_COMMAND];
            i = obj->cfg_w[CFG_CPU_SPINNER];
            break;
#endif
#if _MDH_HAS_MEM
        case _PANEL_TYPE_MEM:
            c = obj->cfg_w[CFG_MEM_COMMAND];
            i = obj->cfg_w[CFG_MEM_SPINNER];
            break;
#endif
#if _MDH_HAS_DISK
        case _PANEL_TYPE_DISK:
            v = obj->cfg_w[CFG_DISK_VALUE];
            c = obj->cfg_w[CFG_DISK_COMMAND];
            break;
#endif
#if _MDH_HAS_NET
        case _PANEL_TYPE_NET:
            v = obj->cfg_w[CFG_NET_VALUE];
            c = obj->cfg_w[CFG_NET_COMMAND];
            i = obj->cfg_w[CFG_NET_SPINNER];
            break;
#endif
        case _PANEL_TYPE_TEMP:
            v = obj->cfg_w[CFG_TEMP_VALUE];
            c = obj->cfg_w[CFG_TEMP_COMMAND];
            i = obj->cfg_w[CFG_TEMP_SPINNER];
            break;
#if _MDH_HAS_UPTIME
        case _PANEL_TYPE_UPTIME:
            c = obj->cfg_w[CFG_UPTIME_COMMAND];
            break;
#endif
        default:
            break;
    }

    if(e != NULL)
        cfg->enabled  = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(e));

    if(v != NULL)
        cfg->value    = gtk_entry_get_text(GTK_ENTRY(v));

    if(c != NULL)
        cfg->command  = gtk_entry_get_text(GTK_ENTRY(c));

    if(i != NULL) {
        cfg->interval = gtk_spin_button_get_value(GTK_SPIN_BUTTON(i));

        if(obj->type == _PANEL_TYPE_TEMP)
            cfg->interval *= 3600;
    }
}
