/*
 * mdh (MailDooHicky)
 *
 * Copyright (c) 2003 Mike Hokenson <mdh at gozer dot org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "mdh.h"
#include "util.h"

#include "mdh_gen.h"

/* general event handler for misc windows, close if Escape is hit */
gint mdh_window_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    g_return_val_if_fail(GTK_IS_WIDGET(widget) || event != NULL, 0);

    switch(event->type) {
        case GDK_KEY_PRESS:
            switch(event->key.keyval) {
                case GDK_Escape:
                    gtk_widget_destroy(widget);
                    break;
                default:
                    break;
            }
            break;
        default:
            break;
    }

    return(0);
}

/* create a button with a stock icon and custom text */
GtkWidget *mdh_button_new_custom_stock(const gchar *stock_id,
                                       const gchar *text)
{
    GtkWidget *vbox;
    GtkWidget *hbox;
    GtkWidget *button;

    GtkWidget *image;
    GtkWidget *label;

    g_return_val_if_fail(stock_id != NULL && text != NULL , NULL);

    /* to center the image and text in the button */
    vbox = gtk_hbox_new(TRUE, 0);
    gtk_widget_show(vbox);

    hbox = gtk_hbox_new(FALSE, 2);
    gtk_widget_show(hbox);

    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 0);

    image = gtk_image_new_from_stock(stock_id, GTK_ICON_SIZE_BUTTON);
    gtk_widget_show(image);

    gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);

    label = gtk_label_new(text);
    gtk_widget_show(label);

    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

    button = gtk_button_new();

    gtk_container_add(GTK_CONTAINER(button), vbox);

    return(button);
}

void mdh_tooltips_set_tip(GtkTooltips *tips,
                          GtkWidget *widget,
                          const gchar *tip)
{
    g_return_if_fail(tips != NULL && widget != NULL);

    gtk_tooltips_set_tip(tips, widget, (STRLEN(tip)) ? tip : NULL, NULL);
}

void mdh_widget_set_font(GtkWidget *widget, const gchar *font)
{
    GtkWidget  *label = NULL;
    GtkRcStyle *style;

    g_return_if_fail(GTK_IS_WIDGET(widget));

    if(font == NULL)
        return;

    if(GTK_IS_BUTTON(widget)) {
        GList *list = NULL;

        list  = gtk_container_children(GTK_CONTAINER(widget));

        label = GTK_WIDGET(list->data);

        g_list_free(list);
    } else if(GTK_IS_LABEL(widget)) {
        label = widget;
    } else {
        g_print("mdh_widget_set_font(): widget is not GtkButton or GtkLabel\n");
        return;
    }

    style = gtk_rc_style_new();

    pango_font_description_free(style->font_desc);

    style->font_desc = pango_font_description_from_string(font);

    gtk_widget_modify_style(label, style);

    gtk_rc_style_unref(GTK_RC_STYLE(style));
}

void mdh_label_set_text(GtkLabel *label, const gchar *text, gboolean bold)
{
    g_return_if_fail(GTK_IS_LABEL(label) || text != NULL);

    if(bold == FALSE) {
        gtk_label_set_text(label, text);
        return;
    }

    mdh_label_set_bold(label, text);
}

void mdh_window_notice(const gchar *head, const gchar *body)
{
    GtkWidget *window;
    GtkWidget *vbox;
    GtkWidget *vbox2;
    GtkWidget *hbox;
    GtkWidget *image;
    GtkWidget *label;
    GtkWidget *button;

    g_return_if_fail(head != NULL && body != NULL);

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), "MailDooHicky: Notice");
    gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);
    gtk_widget_set_usize(GTK_WIDGET(window), 425, 0);

    gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_UTILITY);

    g_signal_connect(G_OBJECT(window), "event",
        G_CALLBACK(mdh_window_event), window);

    gtk_container_set_border_width(GTK_CONTAINER(window), 5);

    vbox = gtk_vbox_new(FALSE, 5);
    gtk_widget_show(vbox);

    gtk_container_add(GTK_CONTAINER(window), vbox);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    image = gtk_image_new_from_stock(GTK_STOCK_DIALOG_WARNING,
        GTK_ICON_SIZE_DIALOG);
    gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
    gtk_widget_show(image);

    gtk_misc_set_alignment(GTK_MISC(image), 0.5, 0.0);
    gtk_misc_set_padding(GTK_MISC(image), 0, 5);

    vbox2 = gtk_vbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(hbox), vbox2, FALSE, FALSE, 0);
    gtk_widget_show(vbox2);

    gtk_container_set_border_width(GTK_CONTAINER(vbox2), 5);

    label = gtk_label_new(NULL);
    gtk_box_pack_start(GTK_BOX(vbox2), label, TRUE, FALSE, 5);
    gtk_widget_show(label);

    gtk_widget_set_usize(GTK_WIDGET(label), 350, 0);
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    mdh_label_set_larger(GTK_LABEL(label), head);

    label = gtk_label_new(body);
    gtk_box_pack_start(GTK_BOX(vbox2), label, TRUE, FALSE, 0);
    gtk_widget_show(label);

    gtk_widget_set_usize(GTK_WIDGET(label), 350, 0);
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);

    button = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
    gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    mdh_widget_set_default(GTK_WINDOW(window), button);

    g_signal_connect_swapped(G_OBJECT(button), "clicked",
        G_CALLBACK(gtk_widget_destroy), G_OBJECT(window));

    gtk_widget_show(window);
}

void mdh_window_error(const gchar *head, const gchar *body)
{
    GtkWidget *window;
    GtkWidget *vbox;
    GtkWidget *vbox2;
    GtkWidget *hbox;
    GtkWidget *image;
    GtkWidget *label;
    GtkWidget *button;

    g_return_if_fail(head != NULL && head != NULL);

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), "MailDooHicky: Error");
    gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);
    gtk_widget_set_usize(GTK_WIDGET(window), 425, 0);

    gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_UTILITY);

    g_signal_connect(G_OBJECT(window), "event",
        G_CALLBACK(mdh_window_event), window);

    gtk_container_set_border_width(GTK_CONTAINER(window), 5);

    vbox = gtk_vbox_new(FALSE, 5);
    gtk_widget_show(vbox);

    gtk_container_add(GTK_CONTAINER(window), vbox);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    image = gtk_image_new_from_stock(GTK_STOCK_DIALOG_ERROR,
        GTK_ICON_SIZE_DIALOG);
    gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
    gtk_widget_show(image);

    gtk_misc_set_alignment(GTK_MISC(image), 0.5, 0.0);
    gtk_misc_set_padding(GTK_MISC(image), 0, 5);

    vbox2 = gtk_vbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(hbox), vbox2, FALSE, FALSE, 0);
    gtk_widget_show(vbox2);

    gtk_container_set_border_width(GTK_CONTAINER(vbox2), 5);

    label = gtk_label_new(NULL);
    gtk_box_pack_start(GTK_BOX(vbox2), label, TRUE, FALSE, 5);
    gtk_widget_show(label);

    gtk_widget_set_usize(GTK_WIDGET(label), 350, 0);
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    mdh_label_set_larger(GTK_LABEL(label), head);

    label = gtk_label_new(body);
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox2), label, TRUE, FALSE, 0);
    gtk_widget_show(label);

    gtk_widget_set_usize(GTK_WIDGET(label), 350, 0);
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);

    button = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
    gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    mdh_widget_set_default(GTK_WINDOW(window), button);

    g_signal_connect_swapped(G_OBJECT(button), "clicked",
        G_CALLBACK(gtk_widget_destroy), G_OBJECT(window));

    gtk_widget_show(window);
}

void gtk_widget_can_focus(GtkWidget *widget, gboolean value)
{
    g_return_if_fail(GTK_IS_WIDGET(widget));

    if(value == TRUE)
        GTK_WIDGET_SET_FLAGS  (GTK_WIDGET(widget), GTK_CAN_FOCUS);
    else
        GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(widget), GTK_CAN_FOCUS);
}

void mdh_widget_set_default(GtkWindow *window, GtkWidget *button)
{
    g_return_if_fail(GTK_IS_WINDOW(window) || GTK_IS_WIDGET(button));

    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);

    gtk_window_set_default(GTK_WINDOW(window), button);
}

#define _MDH_LARGE_MARKUP "<span weight=\"bold\" size=\"larger\">%s</span>"
void mdh_label_set_larger(GtkLabel *label, const gchar *text)
{
    gchar *buf = g_strdup_printf(_MDH_LARGE_MARKUP, text);

    gtk_label_set_markup(label, buf);

    g_free(buf);
}

#define _MDH_SMALL_MARKUP "<span weight=\"bold\" size=\"smaller\">%s</span>"
void mdh_label_set_smaller(GtkLabel *label, const gchar *text)
{
    gchar *buf = g_strdup_printf(_MDH_SMALL_MARKUP, text);

    gtk_label_set_markup(label, buf);

    g_free(buf);
}

#define _MDH_BOLD_MARKUP  "<span weight=\"bold\">%s</span>"
void mdh_label_set_bold(GtkLabel *label, const gchar *text)
{
    gchar *buf = g_strdup_printf(_MDH_BOLD_MARKUP, text);

    gtk_label_set_markup(label, buf);

    g_free(buf);
}

GtkWidget *mdh_label_new(const gchar *fmt, ...)
{
    gchar buf[LINE_MAX];

    va_list args;

    va_start(args, fmt);

    g_vsnprintf(buf, sizeof(buf), fmt, args);

    va_end(args);

    return(gtk_label_new(buf));
}

gboolean mdh_window_yes_no_prompt(const gchar *text)
{
    GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(Toolbar.window),
                                               GTK_DIALOG_MODAL,
                                               GTK_MESSAGE_QUESTION,
                                               GTK_BUTTONS_YES_NO,
                                               text);

    switch(gtk_dialog_run(GTK_DIALOG(dialog))) {
        case GTK_RESPONSE_YES:
            gtk_widget_destroy(dialog);
            return(TRUE);
        default:
            break;
    }

    gtk_widget_destroy(dialog);

    return(FALSE);
}

#if (!defined(GTK_CHECK_VERSION) || !GTK_CHECK_VERSION(2, 2, 0))

#define G_SLIST(x) ((GSList *) x)
#define GTK_LIST_STORE_IS_SORTED(list) (GTK_LIST_STORE (list)->sort_column_id != -2)
#define VALID_ITER(iter, list_store) (iter!= NULL && iter->user_data != NULL && list_store->stamp == iter->stamp)

/**
 * gtk_list_store_swap:
 * @store: A #GtkListStore.
 * @a: A #GtkTreeIter.
 * @b: Another #GtkTreeIter.
 *
 * Swaps @a and @b in @store. Note that this function only works with
 * unsorted stores.
 *
 * Since: 2.2
 **/
void
gtk_list_store_swap (GtkListStore *store,
                     GtkTreeIter  *a,
                     GtkTreeIter  *b)
{
  GSList *i, *prev_a = NULL, *prev_b = NULL;
  gint j, a_count = 0, b_count = 0, *order;
  GtkTreePath *path;

  g_return_if_fail (GTK_IS_LIST_STORE (store));
  g_return_if_fail (!GTK_LIST_STORE_IS_SORTED (store));
  g_return_if_fail (VALID_ITER (a, store));
  g_return_if_fail (VALID_ITER (b, store));

  if (a->user_data == b->user_data)
    return;

  if (a->user_data == store->root)
    prev_a = NULL;
  else
    {
      for (i = store->root; i; i = i->next, a_count++)
        if (i->next == a->user_data)
          {
            prev_a = i;
            break;
          }

      a_count++;
    }

  if (b->user_data == store->root)
    prev_b = NULL;
  else
    {
      for (i = store->root; i; i = i->next, b_count++)
        if (i->next == b->user_data)
          {
            prev_b = i;
            break;
          }

      b_count++;
    }

  if (!prev_a)
    store->root = b->user_data;
  else
    prev_a->next = b->user_data;

  if (!prev_b)
    store->root = a->user_data;
  else
    prev_b->next = a->user_data;

  /* think a_next inspead of a_prev here ... */
  prev_a = G_SLIST (a->user_data)->next;
  prev_b = G_SLIST (b->user_data)->next;

  G_SLIST (a->user_data)->next = prev_b;
  G_SLIST (b->user_data)->next = prev_a;

  /* update tail if needed */
  if (! G_SLIST (a->user_data)->next)
    store->tail = G_SLIST (a->user_data);
  else if (! G_SLIST (b->user_data)->next)
    store->tail = G_SLIST (b->user_data);

  /* emit signal */
  order = g_new (gint, store->length);
  for (j = 0; j < store->length; j++)
    if (j == a_count)
      order[j] = b_count;
    else if (j == b_count)
      order[j] = a_count;
    else
      order[j] = j;

  path = gtk_tree_path_new ();
  gtk_tree_model_rows_reordered (GTK_TREE_MODEL (store),
                                 path, NULL, order);
  gtk_tree_path_free (path);
  g_free (order);
}
#endif /* (!defined(GTK_CHECK_VERSION) || !GTK_CHECK_VERSION(2, 2, 0)) */
