/*
 * mdh (MailDooHicky)
 *
 * Copyright (c) 2003 Mike Hokenson <mdh at gozer dot org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/param.h>
#include <unistd.h>

#ifdef __linux__
#include <sys/vfs.h>
#endif

#ifdef BSD4_4
#include <sys/mount.h>
#endif

#ifdef __sun__
#include <sys/statvfs.h>
#endif

#include <gtk/gtk.h>

#include "mdh.h"
#include "util.h"

#include "mdh_gen.h"

#include "mdh_toolbar.h"
#include "mdh_panel_disk.h"

#if _MDH_HAS_DISK

struct disk_t {
    gdouble bytes;
    gdouble percent;
    gchar b;
};

static const gchar *mdh_toolbar_panel_disk_get_data(struct panel_t *obj);

static gboolean mdh_toolbar_panel_disk_get_real(struct panel_t *obj,
                                                struct disk_t *disk);

static GtkFunction mdh_toolbar_panel_disk_init(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_disk_init()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) FALSE); */

    if(obj->init == TRUE)
        return((GtkFunction) TRUE);

    g_return_val_if_fail(obj->value != NULL, (GtkFunction) FALSE);

    if(!STRLEN(obj->value)) {
        ADD_LOG(NULL, 0, "mdh_toolbar_panel_disk_init: empty value used");
        mdh_toolbar_panel_stop(obj);
        return((GtkFunction) FALSE);
    }

    obj->err  = _ERR_NONE;

    obj->init = TRUE;

    return((GtkFunction) TRUE);
}

static GtkFunction mdh_toolbar_panel_disk_close(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_disk_close()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    if(obj->init == FALSE)
        return((GtkFunction) FALSE);

    obj->init = FALSE;

    return((GtkFunction) FALSE);
}

static GtkFunction mdh_toolbar_panel_disk(struct panel_t *obj)
{
    const gchar *text;

    DEBUG(obj, "mdh_toolbar_panel_disk()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) obj->init); */

    if((text = mdh_toolbar_panel_disk_get_data(obj)) != NULL)
        mdh_label_set_text(GTK_LABEL(obj->widget), text, obj->bold);

    return((GtkFunction) obj->init);
}

static const gchar *mdh_toolbar_panel_disk_get_data(struct panel_t *obj)
{
    static gchar text[_DATA_MAX];

    struct disk_t disk;

    DEBUG(obj, "mdh_toolbar_panel_disk_get_data()");

    g_return_val_if_fail(obj != NULL, NULL);

    obj->bold = FALSE;

    if(obj->init == FALSE) {
        if(obj->init_open(obj) == FALSE) {
            mdh_toolbar_panel_stop(obj);
            return("N/A");
        }
    }

    if(mdh_toolbar_panel_disk_get_real(obj, &disk) == FALSE) {
        if(P_Disk_opt.labels == TRUE)
            g_snprintf(text, sizeof(text), "%s: N/A", obj->value);
        else
            g_strlcpy(text, "N/A", sizeof(text));

        return(text);
    }

    obj->err = _ERR_NONE;

    if(P_Disk_opt.use_bold == TRUE && disk.percent >= _BOLD_DISK)
        obj->bold = TRUE;

    if(P_Disk_opt.labels == TRUE)
        g_snprintf(text, sizeof(text), "%s: %2.1f %cB (%2.0f%%)",
                   obj->value, disk.bytes, disk.b, disk.percent);
    else
        g_snprintf(text, sizeof(text), "%2.1f %cB (%2.0f%%)",
                   disk.bytes, disk.b, disk.percent);

    return(text);
}

static gboolean mdh_toolbar_panel_disk_get_real(struct panel_t *obj,
                                                struct disk_t *disk)
{
    gdouble bytes, percent;

    gchar b;

#ifdef __sun__
    struct statvfs sfs;
#else
    struct statfs sfs;
#endif

    DEBUG(obj, "mdh_toolbar_panel_disk_get_real()");

    g_return_val_if_fail(obj != NULL, FALSE);

#ifdef __sun__
    if(statvfs(obj->value, &sfs) == -1) {
        PERROR(obj, errno, "mdh_toolbar_panel_disk_get_real: statvfs: %s", obj->value);
#else
    if(statfs(obj->value, &sfs) == -1) {
        PERROR(obj, errno, "mdh_toolbar_panel_disk_get_real: statfs: %s", obj->value);
#endif
        return(FALSE);
    }

#ifdef __sun__
    bytes   = (double) (sfs.f_blocks - sfs.f_bfree) * sfs.f_frsize;
#else
    bytes   = (double) (sfs.f_blocks - sfs.f_bfree) * sfs.f_bsize;
#endif

    percent = ((sfs.f_blocks - sfs.f_bfree) * 100 /
                  (sfs.f_blocks - sfs.f_bfree + sfs.f_bavail) + 0.5);

    if(bytes > 1073741824) {
        b      = 'G';
        bytes /= 1073741824;
    } else {
        b      = 'M';
        bytes /= 1048576;
    }

    disk->bytes   = bytes;
    disk->percent = percent;
    disk->b       = b;

    return(TRUE);
}

void mdh_toolbar_panel_disk_setup(struct panel_t *obj,
                                  const gchar *fs,
                                  const gchar *command,
                                  gboolean enabled)
{
    memset(obj, '\0', sizeof(struct panel_t));

    obj->init        = FALSE;
    obj->name        = "disk";
    obj->value       = (fs)      ? g_strdup(fs)      : g_strdup("/");
    obj->command     = (command) ? g_strdup(command) : g_strdup("");
    obj->tooltip     = g_strdup_printf("Disk: %s", obj->value);
    obj->interval    = 60;
    obj->enabled     = (enabled && fs && IS_DIR(fs)) ? TRUE : FALSE;
    obj->bold        = FALSE;
    obj->callback    = (GtkFunction) mdh_toolbar_panel_command_execute;
    obj->function    = (GtkFunction) mdh_toolbar_panel_disk;
    obj->init_open   = (GtkFunction) mdh_toolbar_panel_disk_init;
    obj->init_close  = (GtkFunction) mdh_toolbar_panel_disk_close;
}

#endif /* _MDH_HAS_DISK */
