/*
 * mdh (MailDooHicky)
 *
 * Copyright (c) 2003 Mike Hokenson <mdh at gozer dot org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <gtk/gtk.h>

#include "mdh.h"
#include "util.h"

#include "mdh_gen.h"

#include "mdh_toolbar.h"
#include "mdh_panel_mail.h"

/* mailbox types */
enum {
    _PANEL_MAIL_TYPE_NONE,
    _PANEL_MAIL_TYPE_BOX,
    _PANEL_MAIL_TYPE_DIR
};

static const gchar *mdh_toolbar_panel_mail_get_data(struct panel_t *obj);

static const gchar *mdh_toolbar_panel_mail_get_mail_box(struct panel_t *obj);
static const gchar *mdh_toolbar_panel_mail_get_mail_dir(struct panel_t *obj);

static gint mdh_toolbar_panel_mail_get_mail_dir_count(struct panel_t *obj,
                                                      GDir *dp);

static gboolean mdh_toolbar_panel_mail_box_message_is_internal(const gchar *str);
static gboolean mdh_toolbar_panel_mail_box_message_is_read(const gchar *str);

static GtkFunction mdh_toolbar_panel_mail_init(struct panel_t *obj)
{
    struct p_mail_t *data;

    DEBUG(obj, "mdh_toolbar_panel_mail_init()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) FALSE); */

    if(obj->init == TRUE)
        return((GtkFunction) TRUE);

    g_return_val_if_fail(obj->value != NULL, (GtkFunction) FALSE);

    if(!STRLEN(obj->value)) {
        ADD_LOG(NULL, 0, "mdh_toolbar_panel_mail_init: empty value used");
        mdh_toolbar_panel_stop(obj);
        return((GtkFunction) FALSE);
    }

    data = obj->data;

    memset(data, '\0', sizeof(struct p_mail_t));

    data->type  = _PANEL_MAIL_TYPE_NONE;
    data->last  = 0;
    data->inode = 0;

    if(IS_FILE(obj->value) == TRUE) {
        struct stat statbuf;

        stat(obj->value, &statbuf);

        /* track inode for file replacement/changes */
        data->inode = statbuf.st_ino;

        data->type  = _PANEL_MAIL_TYPE_BOX;

        if((data->fp = fopen(obj->value, "r")) == NULL) {
            PERROR(obj, errno, "mdh_toolbar_panel_mail_init: fopen: %s",
                obj->value);
            /* mdh_toolbar_panel_stop(obj); */
            return((GtkFunction) FALSE);
        }

        obj->err = _ERR_NONE;
    } else if(IS_DIR(obj->value) == TRUE) {
        data->type = _PANEL_MAIL_TYPE_DIR;

        data->md_cur = g_build_filename(obj->value, "cur", NULL);
        data->md_new = g_build_filename(obj->value, "new", NULL);

        if(IS_DIR(data->md_cur) == FALSE || IS_DIR(data->md_new) == FALSE) {
            ADD_LOG(obj, errno,
                    "mdh_toolbar_panel_mail_init: invalid maildir: %s",
                    obj->value);

            g_free(data->md_cur);
            g_free(data->md_new);

            /* mdh_toolbar_panel_stop(obj); */

            return((GtkFunction) FALSE);
        }

        if((data->dp_cur = g_dir_open(data->md_cur, 0, NULL)) == NULL) {
            PERROR(obj, errno, "mdh_toolbar_panel_mail_init: g_dir_open: %s",
                data->md_cur);

            g_free(data->md_cur);
            g_free(data->md_new);

            /* mdh_toolbar_panel_stop(obj); */

            return((GtkFunction) FALSE);
        }

        if((data->dp_new = g_dir_open(data->md_new, 0, NULL)) == NULL) {
            PERROR(obj, errno, "mdh_toolbar_panel_mail_init: g_dir_open: %s",
                data->md_new);

            g_free(data->md_cur);
            g_free(data->md_new);

            /* mdh_toolbar_panel_stop(obj); */

            return((GtkFunction) FALSE);
        }

        obj->err = _ERR_NONE;
    } else {
        ADD_LOG(obj, _ERR_EXIST,
          "mdh_toolbar_panel_mail_init: %s does not exist or is not a file/dir",
          obj->value);

        /* clients might remove the box if all messages have been removed */
        /* return((GtkFunction) FALSE); */
    }

    obj->init = TRUE;

    return((GtkFunction) TRUE);
}

static GtkFunction mdh_toolbar_panel_mail_close(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_mail_close()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    if(obj->init == FALSE)
        return((GtkFunction) FALSE);

    {
        struct p_mail_t *data = obj->data;

        if(data->type == _PANEL_MAIL_TYPE_BOX) {
            fclose(data->fp);
        } else if(data->type == _PANEL_MAIL_TYPE_DIR) {
            g_dir_close(data->dp_cur);
            g_dir_close(data->dp_new);

            g_free(data->md_cur);
            g_free(data->md_new);
        }
    }

    obj->init = FALSE;

    return((GtkFunction) FALSE);
}

static GtkFunction mdh_toolbar_panel_mail(struct panel_t *obj)
{
    const gchar *text;

    DEBUG(obj, "mdh_toolbar_panel_mail()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) obj->init); */

    if((text = mdh_toolbar_panel_mail_get_data(obj)) != NULL)
        mdh_label_set_text(GTK_LABEL(obj->widget), text, obj->bold);

    return((GtkFunction) obj->init);
}

static const gchar *mdh_basename(const gchar *name)
{
    const gchar *base;

    if((base = strrchr(name, G_DIR_SEPARATOR)) != NULL)
        return(base + 1);

    return(name);
}

static const gchar *mdh_toolbar_panel_mail_get_data(struct panel_t *obj)
{
    static gchar text[_DATA_MAX];

    struct p_mail_t *data;

    DEBUG(obj, "mdh_toolbar_panel_mail_get_data()");

    g_return_val_if_fail(obj != NULL, NULL);

    data = obj->data;

    obj->bold = FALSE;

    if(obj->init == FALSE || data->type == _PANEL_MAIL_TYPE_NONE) {
        obj->init = FALSE;

        if(obj->init_open(obj) == FALSE) {
            mdh_toolbar_panel_stop(obj);

            if(P_Mail_opt.labels == FALSE)
                return("N/A");

            g_snprintf(text, sizeof(text), "%s: N/A", mdh_basename(obj->value));

            return(text);
        }
    }

    /* mailbox or maildir? */
    if(data->type == _PANEL_MAIL_TYPE_BOX)
        return(mdh_toolbar_panel_mail_get_mail_box(obj));
    else if(data->type == _PANEL_MAIL_TYPE_DIR)
        return(mdh_toolbar_panel_mail_get_mail_dir(obj));

    ADD_LOG(obj, _ERR_EXIST,
        "mdh_toolbar_panel_mail_get_data: %s does not exist or is not a file/dir",
        obj->value);

    /* clients might remove the box if all messages have been removed */
    /* mdh_toolbar_panel_stop(obj); */

    if(P_Mail_opt.labels == FALSE)
        return("---/---");

    g_snprintf(text, sizeof(text), "%s: ---/---", mdh_basename(obj->value));

    return(text);
}

/* bits taken from gkrellm-mailwatch */
static gboolean mdh_mail_valid_from(const gchar *str)
{
    gint day;

    if(!ISSET(str))
        return(FALSE);

    if(sscanf(str, "From %*s %*s %*s %d", &day) != 1)
        if(sscanf(str, "From %*s %*s %d", &day) != 1)
            return(FALSE);

    if(day < 1 || day > 31)
        return(FALSE);

    return(TRUE);
}

static const gchar *mdh_toolbar_panel_mail_display(struct panel_t *obj,
                                                   gint new, gint total)
{
    static gchar text[_DATA_MAX];

    g_return_val_if_fail(obj != NULL, NULL);

    obj->bold = FALSE;

    if(P_Mail_opt.use_bold == TRUE && new >= _BOLD_MAIL)
        obj->bold = TRUE;

    if(P_Mail_opt.labels == TRUE)
        g_snprintf(text, sizeof(text), "%s: %3d/%3d",
                   mdh_basename(obj->value), new, total);
    else
        g_snprintf(text, sizeof(text), "%3d/%3d", new, total);

    return(text);
}

static const gchar *mdh_toolbar_panel_mail_get_mail_box(struct panel_t *obj)
{
    gint mail_total, mail_read, mail_new, last;

    struct p_mail_t *data;

    struct stat statbuf;

    DEBUG(obj, "mdh_toolbar_panel_mail_get_mail_box()");

    g_return_val_if_fail(obj != NULL, NULL);

    data = obj->data;

    stat(obj->value, &statbuf);

    /* empty file */
    if(statbuf.st_size == 0)
        return(mdh_toolbar_panel_mail_display(obj, 0, 0));

    /* inode has changed, close panel resources and re-run */
    if(data->inode != statbuf.st_ino) {
        obj->init_close(obj);
        mdh_toolbar_panel_refresh(obj);
        return(NULL);
    }

    data->inode = statbuf.st_ino;

    last = statbuf.st_mtime;

    if(statbuf.st_ctime > last)
        last = statbuf.st_ctime;

    /* see if modification time of file has changed */
    if(last <= data->last)
        return(mdh_toolbar_panel_mail_display(obj,
               data->last_new, data->last_total));

    mail_total = 0;
    mail_read  = 0;

    rewind(data->fp);
    fflush(data->fp);

    {
        gchar buf[LINE_MAX];

        gboolean is_header = FALSE;

        /* bits taken from gkrellm-mailwatch */
        while((fgets(buf, sizeof(buf), data->fp))) {
            if(buf[0] == '\n')
                is_header = FALSE;

            if(is_header == FALSE && buf[0] == 'F') {
                if(STRNEQ(buf, "From ", 5)) {
                    if(mdh_mail_valid_from(buf) == TRUE) {
                        mail_total++;
                        is_header = TRUE;
                    }
                }
            }

            if(is_header == FALSE)
                continue;

            if(mdh_toolbar_panel_mail_box_message_is_internal(buf) == TRUE)
                mail_total--;
            else if(mdh_toolbar_panel_mail_box_message_is_read(buf) == TRUE)
                mail_read++;
        }
    }

    if(mail_total < 0)
        mail_total = 0;

    if((mail_new = mail_total - mail_read) < 0)
        mail_read  = 0;

    if(mail_new < 0)
        mail_new = 0;

    /* update cached data */
    data->last       = last;
    data->last_new   = mail_new;
    data->last_total = mail_total;

    return(mdh_toolbar_panel_mail_display(obj, mail_new, mail_total));
}

static const gchar *mdh_toolbar_panel_mail_get_mail_dir(struct panel_t *obj)
{
    gint mail_total, mail_read, mail_new, last_cur, last_new;

    struct p_mail_t *data;

    DEBUG(obj, "mdh_toolbar_panel_mail_get_mail_dir()");

    g_return_val_if_fail(obj != NULL, NULL);

    data = obj->data;

    last_cur = last_new = 0;

    {
        struct stat statbuf;

        /* get modification time */
        if(stat(data->md_cur, &statbuf) == 0)
            last_cur = statbuf.st_mtime;

        if(stat(data->md_new, &statbuf) == 0)
            last_new = statbuf.st_mtime;
    }

    /* see if modification time of dirs have changed */
    if((last_new <= data->last) && (last_cur <= data->last))
        return(mdh_toolbar_panel_mail_display(obj,
               data->last_new, data->last_total));

    mail_read = mdh_toolbar_panel_mail_get_mail_dir_count(obj, data->dp_cur);
    mail_new  = mdh_toolbar_panel_mail_get_mail_dir_count(obj, data->dp_new);

    mail_total = mail_read + mail_new;

    /* update cached data */

    data->last = last_cur;

    if(last_new > last_cur)
        data->last = last_new;

    data->last_new   = mail_new;
    data->last_total = mail_total;

    return(mdh_toolbar_panel_mail_display(obj, mail_new, mail_total));
}

static gint mdh_toolbar_panel_mail_get_mail_dir_count(struct panel_t *obj,
                                                      GDir *dp)
{
    gint count = 0;

    DEBUG(obj, "mdh_toolbar_panel_mail_get_mail_dir_count()");

    /* g_return_val_if_fail(obj != NULL, -1); */

    if(dp != NULL) {
        ADD_LOG(NULL, 0, "mdh_toolbar_panel_mail_get_mail_dir_count: dir not open");
        return(0);
    }

    g_dir_rewind(dp);

    while((g_dir_read_name(dp)))
        count++;

    return(count);
}

static gboolean mdh_toolbar_panel_mail_box_message_is_internal(const gchar *str)
{
    if(!ISSET(str))
        return(FALSE);

    if(str[0] == 'F' && STRNEQ(str, "From: Mail System Internal Data", 31))
        return(TRUE);

    return(FALSE);
}

static gboolean mdh_toolbar_panel_mail_box_message_is_read(const gchar *str)
{
    if(!ISSET(str))
        return(FALSE);

    if(str[0] == 'S' && STRNEQ(str, "Status: ", 8) && strchr(str + 8, 'R'))
        return(TRUE);

    return(FALSE);
}

gchar *mdh_expand_tilde(const gchar *str)
{
    if(!ISSET(str))
        return(g_strdup(""));

    if(!STRNEQ(str, "~/", 2))
        return(g_strdup(str));

    return(g_build_filename(g_get_home_dir(), str + 2, NULL));
}

void mdh_toolbar_panel_mail_setup(struct panel_t *obj, const gchar *box)
{
    memset(obj, '\0', sizeof(struct panel_t));

    obj->init        = FALSE;
    obj->name        = "mail";
    obj->value       = mdh_expand_tilde(box);
    obj->command     = g_strdup_printf("balsa %s", obj->value);
    obj->tooltip     = g_strdup_printf("Mail: %s", obj->value);
    obj->interval    = _MAIL_I_DEF;
    obj->enabled     = (IS_FILE(obj->value) || IS_DIR(obj->value)) ? TRUE:FALSE;
    obj->bold        = FALSE;
    obj->callback    = (GtkFunction) mdh_toolbar_panel_command_execute;
    obj->function    = (GtkFunction) mdh_toolbar_panel_mail;
    obj->init_open   = (GtkFunction) mdh_toolbar_panel_mail_init;
    obj->init_close  = (GtkFunction) mdh_toolbar_panel_mail_close;

    obj->data        = g_new(struct p_mail_t, 1);
}
