/*
 * mdh (MailDooHicky)
 *
 * Copyright (c) 2003 Mike Hokenson <mdh at gozer dot org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef __linux__
#include <sys/sysinfo.h>
#endif

#ifdef __sun__
#include <sys/swap.h>
#endif

#include <gtk/gtk.h>

#include "mdh.h"
#include "util.h"

#include "mdh_gen.h"

#include "mdh_toolbar.h"
#include "mdh_panel_mem.h"

#if _MDH_HAS_MEM
struct mem_t {
    gdouble used;
    gdouble percent;
};

static const gchar *mdh_toolbar_panel_mem_get_data(struct panel_t *obj);

static gboolean mdh_toolbar_panel_mem_get_real(struct panel_t *obj,
                                               struct mem_t *mem);

static GtkFunction mdh_toolbar_panel_mem_init(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_mem_init()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) FALSE); */

    if(obj->init == TRUE)
        return((GtkFunction) TRUE);

    obj->init = TRUE;

    return((GtkFunction) TRUE);
}

static GtkFunction mdh_toolbar_panel_mem_close(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_mem_close()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    if(obj->init == FALSE)
        return((GtkFunction) FALSE);

    obj->init = FALSE;

    return((GtkFunction) FALSE);
}

static GtkFunction mdh_toolbar_panel_mem(struct panel_t *obj)
{
    const gchar *text;

    DEBUG(obj, "mdh_toolbar_panel_mem()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) obj->init); */

    if((text = mdh_toolbar_panel_mem_get_data(obj)) != NULL)
        mdh_label_set_text(GTK_LABEL(obj->widget), text, obj->bold);

    return((GtkFunction) obj->init);
}

static const gchar *mdh_toolbar_panel_mem_get_data(struct panel_t *obj)
{
    static gchar text[_DATA_MAX];

    struct mem_t mem;

    DEBUG(obj, "mdh_toolbar_panel_mem_get_data()");

    g_return_val_if_fail(obj != NULL, NULL);

    obj->bold = FALSE;

    if(obj->init == FALSE) {
        if(obj->init_open(obj) == FALSE) {
            mdh_toolbar_panel_stop(obj);
            return("N/A");
        }
    }

    if(mdh_toolbar_panel_mem_get_real(obj, &mem) == FALSE) {
        mdh_toolbar_panel_stop(obj);
        return("N/A");
    }

    if(P_Mem_opt.use_bold == TRUE && mem.percent >= _BOLD_MEM)
        obj->bold = TRUE;

    g_snprintf(text, sizeof(text), "%3.0f MB (%2.0f%%)", mem.used, mem.percent);

    return(text);
}

static gboolean mdh_toolbar_panel_mem_get_real(struct panel_t *obj,
                                               struct mem_t *mem)
{
    gulong mem_total, mem_free, swap_total, swap_free;

#ifdef __linux__
    struct sysinfo sysin;
#else /* __sun__ */
    gulong pagesize;
    struct anoninfo anon;
#endif

    DEBUG(obj, "mdh_toolbar_panel_mem_get_real()");

    g_return_val_if_fail(obj != NULL, FALSE);

#ifdef __linux__
    if(sysinfo(&sysin) == -1) {
        PERROR(NULL, 0, "mdh_toolbar_panel_mem_get_real: sysinfo");
        return(FALSE);
    }

    mem_total  = sysin.totalram;
    mem_free   = sysin.freeram;
    swap_total = sysin.totalswap;
    swap_free  = sysin.freeswap;
#else /* __sun__ */
    pagesize  = sysconf(_SC_PAGESIZE);

    /* get mem usage */
    mem_total  = sysconf(_SC_PHYS_PAGES) * pagesize;
    mem_free   = sysconf(_SC_AVPHYS_PAGES) * pagesize;

    /* get swap usage */
    if(swapctl(SC_AINFO, &anon) == -1) {
        PERROR(NULL, 0, "mdh_toolbar_panel_mem_get_real: swapctl: SC_AINFO");
        return(FALSE);
    }

    swap_total = anon.ani_max * pagesize;
    swap_free  = (anon.ani_max - anon.ani_resv) * pagesize;
#endif /* __sun__ */

    mem_total  /= 1048576;
    mem_free   /= 1048576;
    swap_total /= 1048576;
    swap_free  /= 1048576;

    mem_total  += swap_total;

    mem->used    = mem_total - (mem_free + swap_free);
    mem->percent = (mem->used / mem_total * 100) + 0.5;

    return(TRUE);
}

void mdh_toolbar_panel_mem_setup(struct panel_t *obj)
{
    memset(obj, '\0', sizeof(struct panel_t));

    obj->init        = FALSE;
    obj->name        = "mem";
    obj->command     = g_strdup("");
    obj->tooltip     = g_strdup("Mem");
    obj->interval    = 15;
    obj->enabled     = TRUE;
    obj->bold        = FALSE;
    obj->callback    = (GtkFunction) mdh_toolbar_panel_command_execute;
    obj->function    = (GtkFunction) mdh_toolbar_panel_mem;
    obj->init_open   = (GtkFunction) mdh_toolbar_panel_mem_init;
    obj->init_close  = (GtkFunction) mdh_toolbar_panel_mem_close;
}

#endif /* _MDH_HAS_MEM */
