/*
 * mdh (MailDooHicky)
 *
 * Copyright (c) 2003 Mike Hokenson <mdh at gozer dot org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <regex.h>
#include <signal.h>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <gtk/gtk.h>

#include "mdh.h"
#include "util.h"

#include "mdh_gen.h"

#include "mdh_toolbar.h"
#include "mdh_panel_temp.h"

#if (defined(HAVE_H_ERRLIST) && !defined(HAVE_H_ERRLIST_DECL))
extern gint h_nerrs;
extern gchar *h_errlist[];
#endif

#ifndef HAVE_HSTRERROR_DECL
const gchar *hstrerror(gint);
#endif

#if (defined(HAVE_H_ERRNO) && !defined(HAVE_H_ERRNO_DECL))
extern gint h_errno;
#endif

#ifndef HAVE_HSTRERROR
const gchar *hstrerror(gint herr)
{
#ifdef HAVE_H_ERRLIST
    if(herr >= 0 && herr < h_nerrs && h_errlist[herr])
        return(h_errlist[herr]);
    else
#endif
    {
        static gchar buf[100];
        g_snprintf(buf, sizeof(buf), "Host lookup error %d", herr);
        return(buf);
    }
}
#endif /* HAVE_HSTRERROR */

#define _HTTP_MAX_DAT LINE_MAX  /* I've seen ~350 bytes, this should be fine */
#define _HTTP_CONN_TM 60        /* connect() timeout */
#define _HTTP_RECV_TM 180       /* recv() timeout */

#define _WEATHER_HOST "weather.noaa.gov"
#define _WEATHER_PORT 80
#define _WEATHER_PATH "/pub/data/observations/metar/stations/"

/* failed returns from temperature gathering */
#define _WEATHER_ERR  -999      /* gethostbyname/connect/other error */
#define _WEATHER_NONE -998      /* server returned non-http 200 response */
#define _WEATHER_NA   -997      /* no/invalid METAR data returned */
#define _WEATHER_OLD  -996      /* current data (cache check) is old */
#define _WEATHER_PREV -995      /* uses previous data */

#define G_ERROR       ((g_strerror(errno)))
#define H_ERROR       ((hstrerror(h_errno)))

extern gchar *mdh_cache;

static gint sockfd;

static GIOFunc mdh_toolbar_panel_temp_get_http_handle_pipe(GIOChannel* channel,
                                                   GIOCondition condition,
                                                   struct panel_t *obj);

static struct hostent *mdh_gethostbyname(const gchar *host);

static const gchar *mdh_toolbar_panel_temp_get_data(struct panel_t *obj);

static gint mdh_toolbar_panel_temp_cache(struct panel_t *obj);
static void mdh_toolbar_panel_temp_get_http(struct panel_t *obj, time_t c_time);
static void mdh_toolbar_panel_temp_put_cache(struct panel_t *obj, gint temp);

static gchar *mdh_toolbar_panel_temp_setup_http_request(const gchar *station,
                                                        gboolean proxy);

static gboolean mdh_toolbar_panel_temp_is_status(const gchar *str);
static gboolean mdh_toolbar_panel_temp_is_metar(const gchar *str,
                                                const gchar *station);

static gint mdh_toolbar_panel_temp_parse_metar(const gchar *str);

static void mdh_toolbar_panel_temp_send_temp(const gchar *station, gint temp);

static void mdh_toolbar_panel_temp_send_err(const gchar *fmt, ...);

static gboolean mdh_is_fd(gint fd);

static gchar *mdh_regex_match(const gchar *str, const gchar *reg);

static gint mdh_toolbar_panel_temp_get_http_proxy_port(void)
{
    return((P_Temp_opt.proxy_enable) ? P_Temp_opt.proxy_port : _WEATHER_PORT);
}

static const gchar *mdh_toolbar_panel_temp_get_http_proxy_host(void)
{
    return((P_Temp_opt.proxy_enable) ? P_Temp_opt.proxy_host : _WEATHER_HOST);
}

void SIG_ALRM()
{
    const gchar *host = mdh_toolbar_panel_temp_get_http_proxy_host();
    gint port = mdh_toolbar_panel_temp_get_http_proxy_port();

    alarm(0);
    close(sockfd);

    errno = ETIMEDOUT;

    mdh_toolbar_panel_temp_send_err("connect: %s:%d: %s", host, port, G_ERROR);

    mdh_toolbar_panel_temp_send_temp(P_Temp.value, _WEATHER_ERR);

    _exit(1);
}

void SIG_CHLD()
{
    wait(NULL);

    set_signal(SIGCHLD, SIG_DFL);
}

static GtkFunction mdh_toolbar_panel_temp_init(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_temp_init()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) FALSE); */

    if(obj->init == TRUE)
        return((GtkFunction) TRUE);

    g_return_val_if_fail(obj->value != NULL, (GtkFunction) FALSE);

    if(STRLEN(obj->value) != _WEATHER_LEN) {
        ADD_LOG(NULL, 0, "mdh_toolbar_panel_temp_init: invalid station-id");

        mdh_toolbar_panel_stop(obj);

        return((GtkFunction) FALSE);
    }

    {
        struct p_temp_t *data = obj->data;

        data->last = 0;
        data->temp = _WEATHER_ERR;
    }

    obj->err  = _WEATHER_ERR;

    obj->init = TRUE;

    return((GtkFunction) TRUE);
}

static GtkFunction mdh_toolbar_panel_temp_close(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_temp_close()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    if(obj->init == FALSE)
        return((GtkFunction) FALSE);

    obj->init = FALSE;

    return((GtkFunction) FALSE);
}

static GtkFunction mdh_toolbar_panel_temp(struct panel_t *obj)
{
    const gchar *text;

    DEBUG(obj, "mdh_toolbar_panel_temp()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) obj->init); */

    if((text = mdh_toolbar_panel_temp_get_data(obj)) != NULL)
        mdh_label_set_text(GTK_LABEL(obj->widget), text, FALSE);

    return((GtkFunction) obj->init);
}

static const gchar *mdh_toolbar_panel_temp_get_data(struct panel_t *obj)
{
    gint ret;

    static gchar text[_DATA_MAX];

    struct p_temp_t *data;

    gboolean err = FALSE;

    DEBUG(obj, "mdh_toolbar_panel_temp_get_data()");

    g_return_val_if_fail(obj != NULL, NULL);

    if(obj->init == FALSE) {
        if(obj->init_open(obj) == FALSE) {
            mdh_toolbar_panel_stop(obj);
            return("N/A");
        }
    }

    data = obj->data;

    ret = mdh_toolbar_panel_temp_cache(obj);

    switch(ret) {
        case _WEATHER_ERR:
        case _WEATHER_NONE:
        case _WEATHER_NA:
        case _WEATHER_PREV:
            break;
        case _WEATHER_OLD:
        {
            gint last = data->last, temp = data->temp;

            mdh_toolbar_panel_temp_get_http(obj, time(NULL));

            if(last == 0 && temp == _WEATHER_ERR)
                return("..."); /* first run, no data */
            else
                return(NULL);  /* handle_pipe() does the work */
        }
        default:
            break;
    }

    if(obj->err == _ERR_NONE)
        data->temp = ret;
    else
        err = TRUE;

    switch(obj->err) {
        case _WEATHER_ERR:
            if(data->temp != _WEATHER_ERR) break;
            return("Error");
        case _WEATHER_NONE:
            if(data->temp != _WEATHER_ERR) break;
            return("Invalid ID");
        case _WEATHER_NA:
            if(data->temp != _WEATHER_ERR) break;
            return("N/A");
        case _WEATHER_PREV:
            return(NULL);
        default:
            break;
    }

    if(P_Temp_opt.celcius == TRUE) {
        g_snprintf(text , sizeof(text), "%dC%s",
            data->temp, (err == TRUE) ? " *" : "");

        return(text);
    }

    /* -0.4 */
    if(data->temp == -18) {
        if(err == TRUE) {
            return("0F *");
        } else {
            return("0F");
        }
    }

    g_snprintf(text, sizeof(text), "%1.0fF%s",
        (data->temp * 1.8) + 32, (err == TRUE) ? " *" : "");

    return(text);
}

static gint mdh_toolbar_panel_temp_cache_memory(struct panel_t *obj,
                                                time_t c_time)
{
    DEBUG(obj, "mdh_toolbar_panel_temp_cache_memory()");

    g_return_val_if_fail(obj != NULL, _WEATHER_ERR);

    {
        struct p_temp_t *data = obj->data;

        /* first pass or values have been reset when config window was closed */
        if(data->last == 0)
            return(_WEATHER_OLD);

        if(data->last > (c_time - obj->interval))
            return(data->temp);
    }

    return(_WEATHER_OLD);
}

static gint mdh_toolbar_panel_temp_cache_file(struct panel_t *obj,
                                              time_t c_time)
{
    gint cache_file = _WEATHER_OLD;

    gchar buf[LINE_MAX];

    FILE *fp;

    DEBUG(obj, "mdh_toolbar_panel_temp_cache_file()");

    g_return_val_if_fail(obj != NULL, _WEATHER_ERR);

    if(IS_FILE(mdh_cache) == FALSE)
        return(_WEATHER_OLD);

    if((fp = fopen(mdh_cache, "r")) == NULL) {
        PERROR(NULL, 0, "mdh_toolbar_panel_temp_cache_file: fopen: %s",
               mdh_cache);
        return(_WEATHER_ERR);
    }

    while((fgets(buf, sizeof(buf), fp))) {
        gchar entry_id[_WEATHER_LEN + 1];
        gint entry_time, entry_temp;

        if(buf[0] != obj->value[0])
            continue;

        if(sscanf(buf, "%4s:%d:%d", entry_id, &entry_time, &entry_temp) != 3)
            continue;

        if(entry_temp == _WEATHER_OLD || entry_temp == _WEATHER_PREV)
            continue;

        if(STREQ(entry_id, obj->value)) {
            if(entry_time > (c_time - obj->interval)) {
                struct p_temp_t *data = obj->data;

                cache_file = entry_temp;

                /* first load of cache? set these so memory cache will work */
                if(data->last == 0) {
                    data->last = entry_time;
                    obj->err   = entry_temp;

                    switch(entry_temp) {
                        case _WEATHER_ERR:
                        case _WEATHER_NONE:
                        case _WEATHER_NA:
                        case _WEATHER_PREV:
                            break;
                        default:
                            obj->err   = _ERR_NONE;
                            data->temp = entry_temp;
                            break;
                    }
                }
            }

            break;
        }
    }

    fclose(fp);

    return(cache_file);
}

/* checks memory cache first, then file */
static gint mdh_toolbar_panel_temp_cache(struct panel_t *obj)
{
    gint cache_memory;

    time_t current_time;

    DEBUG(obj, "mdh_toolbar_panel_temp_cache()");

    g_return_val_if_fail(obj != NULL, _WEATHER_ERR);

    current_time = time(NULL);

    cache_memory = mdh_toolbar_panel_temp_cache_memory(obj, current_time);

    switch(cache_memory) {
        case _WEATHER_OLD:
            return(mdh_toolbar_panel_temp_cache_file(obj, current_time));
            break;
        default:
            return(cache_memory);
    }

    return(_WEATHER_OLD); /* not reached */
}

static void mdh_toolbar_panel_temp_put_cache(struct panel_t *obj, gint temp)
{
    gchar buf[LINE_MAX], *mdh_cache_tmp;

    gboolean has_cache;

    FILE *fp1 = NULL, *fp2;

    DEBUG(obj, "mdh_toolbar_panel_temp_put_cache()");

    g_return_if_fail(obj != NULL);

    if((has_cache = IS_FILE(mdh_cache)) == TRUE) {
        if((fp1 = fopen(mdh_cache, "r")) == NULL) {
            PERROR(NULL, 0, "mdh_toolbar_panel_temp_put_cache: fopen: %s",
                   mdh_cache);
            return;
        }
    }

    mdh_cache_tmp = g_strdup_printf("%s.tmp", mdh_cache);

    if((fp2 = fopen(mdh_cache_tmp, "w")) == NULL) {
        PERROR(NULL, 0, "mdh_toolbar_panel_temp_put_cache: fopen: %s",
               mdh_cache_tmp);
        g_free(mdh_cache_tmp);
        return;
    }

    if(has_cache == TRUE)
        while((fgets(buf, sizeof(buf), fp1)))
            if(!STRNEQ(buf, obj->value, _WEATHER_LEN)) /* other station-ids */
                fprintf(fp2, "%s", buf);


    {
        struct p_temp_t *data = obj->data;

        fprintf(fp2, "%s:%d:%d\n", obj->value, data->last, temp);
    }

    if(has_cache == TRUE)
        fclose(fp1);

    fclose(fp2);

    if(has_cache == TRUE)
        mdh_unlink(mdh_cache);

    mdh_rename(mdh_cache_tmp, mdh_cache);

    g_free(mdh_cache_tmp);
}

static void mdh_toolbar_panel_temp_get_http(struct panel_t *obj, time_t c_time)
{
    GIOChannel *channel;

    gint pipes[2];

    pid_t pid;

    DEBUG(obj, "mdh_toolbar_panel_temp_get_http()");

    g_return_if_fail(obj != NULL);

    {
        struct p_temp_t *data = obj->data;

        /* update the timestamp regardless of success */
        data->last = c_time;
    }

    if(pipe(pipes) == -1) {
        PERROR(NULL, 0, "mdh_toolbar_panel_temp_get_http: pipe");
        return;
    }

    set_signal(SIGCHLD, SIG_CHLD);

    if((pid = fork()) == -1) {
        PERROR(NULL, 0, "mdh_toolbar_panel_temp_get_http: fork");
        return;
    }

    if(pid == 0) {
        gchar http_buf[_HTTP_MAX_DAT], line[LINE_MAX], *http_request;

        gint x, i, recv_bytes;

        const gchar *host = mdh_toolbar_panel_temp_get_http_proxy_host();
        gint port = mdh_toolbar_panel_temp_get_http_proxy_port();

        gchar *station = g_strdup(obj->value);

        struct hostent *he;
        struct sockaddr_in their_addr;

        Toolbar.debug = FALSE;

        g_slist_foreach(Toolbar.panels, (GFunc) mdh_toolbar_panel_stop, NULL);

        /*
         * memory is freed by mdh_toolbar_panel_stop(). need the station-id
         * when sending back temperature in this (child) and SIG_ALRM().
         */
        obj->value = station;

        set_signal(SIGTERM, SIG_DFL);
        set_signal(SIGQUIT, SIG_DFL);
        set_signal(SIGINT,  SIG_DFL);
        set_signal(SIGHUP,  SIG_IGN);

        set_signal(SIGCHLD, SIG_IGN);

        set_signal(SIGALRM, SIG_ALRM);

        close(pipes[0]);

        dup2(pipes[1], 0);
        dup2(pipes[1], 1);
        dup2(pipes[1], 2);

        channel = g_io_channel_unix_new(pipes[1]);

        g_usleep(125000);

        http_request = mdh_toolbar_panel_temp_setup_http_request(obj->value,
                           P_Temp_opt.proxy_enable);

        if((he = mdh_gethostbyname(host)) == NULL) {
            mdh_toolbar_panel_temp_send_temp(obj->value, _WEATHER_ERR);

            close(pipes[1]);

            _exit(1);
        }

        if((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
            mdh_toolbar_panel_temp_send_err("socket: %s", G_ERROR);

            mdh_toolbar_panel_temp_send_temp(obj->value, _WEATHER_ERR);

            close(pipes[1]);

            _exit(1);
        }

        their_addr.sin_family = AF_INET;
        their_addr.sin_port = g_htons(port);
        their_addr.sin_addr = *((struct in_addr *)he->h_addr);

        memset(&(their_addr.sin_zero), '\0', 8);

        alarm(_HTTP_CONN_TM);

        if(connect(sockfd, (struct sockaddr *)&their_addr,
            sizeof(struct sockaddr)) == -1) { 

            /* fd closed, likely from a timeout */
            if(mdh_is_fd(sockfd) == FALSE)
                errno = ETIMEDOUT;

            alarm(0);

            mdh_toolbar_panel_temp_send_err("connect: %s:%d %s",
                                            host, port, G_ERROR);

            mdh_toolbar_panel_temp_send_temp(obj->value, _WEATHER_ERR);

            close(pipes[1]);

            _exit(1);
        }

        if(send(sockfd, http_request, STRLEN(http_request), 0) == -1) {
            alarm(0);
            close(sockfd);

            g_free(http_request);

            mdh_toolbar_panel_temp_send_err("send: %s:%d %s",
                                            host, port, G_ERROR);

            mdh_toolbar_panel_temp_send_temp(obj->value, _WEATHER_ERR);

            close(pipes[1]);

            _exit(1);
        }

        g_free(http_request);

        x = 0;
        line[0] = 0;

        while((recv_bytes = recv(sockfd, http_buf, sizeof(http_buf), 0))) {
            alarm(0);
            alarm(_HTTP_RECV_TM);

            if(recv_bytes == -1) {
                alarm(0);
                close(sockfd);

                mdh_toolbar_panel_temp_send_err("recv: %s:%d %s",
                                                host, port, G_ERROR);

                mdh_toolbar_panel_temp_send_temp(obj->value, _WEATHER_ERR);

                close(pipes[1]);

                _exit(1);
            }

            if(recv_bytes > _HTTP_MAX_DAT)
                recv_bytes = _HTTP_MAX_DAT;

            http_buf[recv_bytes] = 0;

            for(i = 0; i <= recv_bytes; i++) {
                if(http_buf[i] != '\n') {
                    if(x < sizeof(line)) {
                        line[x++] = http_buf[i];
                    } else {
                        line[x++] = '\n';
                        continue;
                    }
                } else {
                    line[x] = 0;

                    if(mdh_toolbar_panel_temp_is_status(line) == TRUE) {
                        gint status = 0;

                        sscanf(line, "HTTP/1.%*d %d", &status);

                        if(status != 200) {
                            alarm(0);
                            close(sockfd);

                            line[STRLEN(line) - 1] = 0;

                            mdh_toolbar_panel_temp_send_err("server returned: %s", line);

                            mdh_toolbar_panel_temp_send_temp(obj->value,
                                (status == 404) ? _WEATHER_NONE : _WEATHER_ERR);

                            close(pipes[1]);

                            _exit(1);
                        }
                    }

                    if(mdh_toolbar_panel_temp_is_metar(line, obj->value) == TRUE) {
                        gint temp = mdh_toolbar_panel_temp_parse_metar(line);

                        if(temp == _WEATHER_NA) {
                            alarm(0);
                            close(sockfd);

                            mdh_toolbar_panel_temp_send_err("temp not found: %s", line);

                            mdh_toolbar_panel_temp_send_temp(obj->value,
                                                             _WEATHER_NA);

                            close(pipes[1]);

                            _exit(1);
                        }

                        alarm(0);
                        close(sockfd);

                        mdh_toolbar_panel_temp_send_temp(obj->value, temp);

                        close(pipes[1]);

                        _exit(0);
                    }

                    x = 0;
                    line[0] = 0;
                }
            }
        }

        alarm(0);
        close(sockfd);

        mdh_toolbar_panel_temp_send_err("temp not found");

        mdh_toolbar_panel_temp_send_temp(obj->value, _WEATHER_NA);

        close(pipes[1]);

        _exit(1);
    } else {
        close(pipes[1]);

        channel = g_io_channel_unix_new(pipes[0]);

        g_io_add_watch(channel, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
            (GIOFunc) mdh_toolbar_panel_temp_get_http_handle_pipe, obj);

        waitpid(pid, 0, WNOHANG);
    }
}

static GIOFunc mdh_toolbar_panel_temp_get_http_handle_pipe(GIOChannel* channel,
                                             GIOCondition condition,
                                             struct panel_t *obj)
{
    g_return_val_if_fail(obj != NULL, (GIOFunc) FALSE);

    if(condition & G_IO_IN) {
        gchar *text, **p;
        gint len, i;

        g_io_channel_read_to_end(channel, &text, &len, NULL);

        if(!ISSET(text))
            return((GIOFunc) TRUE);

        text[len - 1] = 0;

        p = g_strsplit(text, "\n", 0);

        g_free(text);

        for(i = 0; p[i] != NULL; i++) {
            DEBUG(obj, "mdh_toolbar_panel_temp_get_http_handle_pipe: %s", p[i]);

            if(STRNEQ(p[i], "T:", 2) && STRLEN(p[i]) > 2) {
                gint  entry_temp;
                gchar entry_id[_WEATHER_LEN + 1];

                if(sscanf(p[i], "T:%4s:%d", entry_id, &entry_temp) == 2) {
                    if(STREQ(obj->value, entry_id)) {
                        struct p_temp_t *data = obj->data;

                        obj->err = entry_temp;

                        switch(entry_temp) {
                            case _WEATHER_ERR:
                                break;
                            case _WEATHER_NONE:
                            case _WEATHER_NA:
                                mdh_toolbar_panel_temp_put_cache(obj,
                                                                 entry_temp);
                                break;
                            default:
                                obj->err = _ERR_NONE;
                                mdh_toolbar_panel_temp_put_cache(obj,
                                                                 entry_temp);
                                data->temp = entry_temp;
                                break;
                        }

                        mdh_toolbar_panel_refresh(obj);
                    } else {
                        ADD_LOG(NULL, 0,
                            "mdh_toolbar_panel_temp_get_http: station-id does not match: %s",
                                entry_id);
                    }
                } else {
                    ADD_LOG(NULL, 0,
                        "mdh_toolbar_panel_temp_get_http: invalid temperature string: %s",
                            p[i]);
                }

                wait(NULL);

                g_strfreev(p);

                g_io_channel_shutdown(channel, FALSE, NULL);
                g_io_channel_unref(channel);

                return((GIOFunc) FALSE);
             }

             ADD_LOG(NULL, 0, "%s", p[i]); /* error */
         }

         g_strfreev(p);
    }

    if(condition & G_IO_HUP || condition & G_IO_ERR || condition & G_IO_NVAL) {
        g_io_channel_shutdown(channel, FALSE, NULL);
        g_io_channel_unref(channel);

        return((GIOFunc) FALSE);
    }

    return((GIOFunc) TRUE);
}

static struct hostent *mdh_gethostbyname(const gchar *host)
{
    gint tries;

    struct hostent *he;

    g_return_val_if_fail(host != NULL && host[0] != '\0', NULL);

    tries = 0;

    while(1) {
        if((he = gethostbyname(host)) == NULL) {
            if(errno == 0) {
#ifdef TRY_AGAIN
                if(h_errno == TRY_AGAIN && tries++ <= 25) {
                    g_usleep(150000);
                    continue;
                }
#endif

                g_print("mdh_gethostbyname: %s: %s\n", host, H_ERROR);
            } else { /* errno != 0 */
                g_print("mdh_gethostbyname: %s: %s\n", host, G_ERROR);
            }

            return(NULL);
        }

        break;
    }

    return(he);
}

static gchar *mdh_toolbar_panel_temp_setup_http_request(const gchar *station,
                                                        gboolean proxy)
{
    g_return_val_if_fail(station != NULL, NULL);

    /*
     * GET [http://weather.noaa.gov]/pub/data/.../stations/KGRB.TXT HTTP/1.0
     * Host: weather.noaa.gov
     * User-Agent: MailDooHicky/X.X.XX (+http://www.gozer.org/.../mdh.php)
     * Connection: close
     */

    return(g_strconcat(
               "GET ",
                   (proxy) ? "http://"     : "",
                   (proxy) ? _WEATHER_HOST : "",
                   _WEATHER_PATH, station, ".TXT HTTP/1.0\r\n",
               "Host: ", _WEATHER_HOST, "\r\n",
               "User-Agent: MailDooHicky/", mdh_toolbar_version(),
                   " (+", _MDH_URL, ")\r\n",
               "Connection: close\r\n",
               "\r\n",
           NULL));
}

static gboolean mdh_toolbar_panel_temp_is_status(const gchar *str)
{
    g_return_val_if_fail(str != NULL && str[0] != '\0', FALSE);

    if(!STRLEN(str))
        return(FALSE);

    if(str[0] == 'H' && STRNEQ(str, "HTTP/1.", 7))
        return(TRUE);

    return(FALSE);
}

static gboolean mdh_toolbar_panel_temp_is_metar(const gchar *str,
                                                const gchar *station)
{
    g_return_val_if_fail(str != NULL && station != NULL, FALSE);

    if(!STRLEN(str) || !STRLEN(station))
        return(FALSE);

    if(str[0] == station[0] && STRNEQ(str, station, _WEATHER_LEN))
        return(TRUE);

    return(FALSE);
}

/*
 * http://weather.unisys.com/wxp/Appendices/Formats/METAR.html
 *
 *  KLAM 051950Z 14003KT 40SM SKC 05/M07 A3024
 *  KEWR 271851Z 34018G28KT 10SM CLR M08/M23 A3032 ...
 *  LSZH 271920Z 0 BR FEW003 SCT007 BKN009 04/04 Q1029
 *  KGRB 252153Z ...27010KT 10SM OVC022 M02/M07 A2996 ...
 *  KGRB 271939Z 18014KT 2SM -SN OVC014 M12/M16 A3013 ...
 *  KGRB 271914Z ... 1 1/4SM -SN OVC024 M12/M17 A3015 ...
 *
 *  KGRB 260153Z 29008KT 10SM CLR 05/ A2994 RMK AO2 ...
 *  KGRB ... 17/17 RMK ...
 */

/* regex on data from weather.noaa.gov to get temperature */
#define _WEATHER_REG  " M?[0-9][0-9]/(M?[0-9][0-9])? [AQR]"

static gint mdh_toolbar_panel_temp_parse_metar(const gchar *str)
{
    gchar *res, buf[4];

    if(!STRLEN(str) || (res = mdh_regex_match(str, _WEATHER_REG)) == NULL)
        return(_WEATHER_NA);

    g_strlcpy(buf, res + 1, sizeof(buf));

    g_free(res);

    /* M09 == -09 */
    if(buf[0] == 'M') {
        buf[0] = '-';
        buf[3] = 0;
    } else {
        buf[2] = 0;
    }

    return(ATOI(buf));
}

/* this won't work for multiple temperature panels */
static void mdh_toolbar_panel_temp_send_temp(const gchar *station, gint temp)
{
    if(!ISSET(station))
        return;

    g_print("T:%s:%d\n", station, temp);
}

static void mdh_toolbar_panel_temp_send_err(const gchar *fmt, ...)
{
    gchar buf[LINE_MAX];

    va_list args;

    va_start(args, fmt);

    g_vsnprintf(buf, sizeof(buf), fmt, args);

    va_end(args);

    g_print("mdh_toolbar_panel_temp_get_http: %s\n", buf);
}

/* returns true if fd is an open descriptor */
static gboolean mdh_is_fd(gint fd)
{
    struct stat statbuf;

    if(fstat(fd, &statbuf) == 0)
        return(TRUE);

    return(FALSE);
}

/* returns NULL if no match else newly allocated string containing match */
static gchar *mdh_regex_match(const gchar *str, const gchar *reg)
{
    gint str_len;
    gchar *match;

    regex_t preg;
    regmatch_t pmatch[2];

    g_return_val_if_fail(str != NULL && reg != NULL, NULL);

    regcomp(&preg, reg, REG_EXTENDED);

    if(regexec(&preg, str, 1, pmatch, 0)) {
        regfree(&preg);
        return(NULL);
    }

    regfree(&preg);

    str_len = STRLEN(str);

    match = g_malloc(str_len + 1);

    /* strip off the end */
    strncpy(match, str, pmatch[0].rm_eo);
    match[pmatch[0].rm_eo] = 0;

    /* skip the junk at the beginning */
    strncpy(match, match + pmatch[0].rm_so, str_len);
    match[pmatch[0].rm_eo - pmatch[0].rm_so] = 0;

    return(match);
}

void mdh_toolbar_panel_temp_proxy_setup(void)
{
    gint port = _PROXY_PORT;
    gchar *http_proxy, *host, *p;

    P_Temp_opt.proxy_enable = FALSE;
    P_Temp_opt.proxy_host   = g_strdup(_PROXY_HOST);
    P_Temp_opt.proxy_port   = port;

    if((http_proxy = (gchar *) g_getenv("http_proxy")) == NULL)
        if((http_proxy = (gchar *) g_getenv("HTTP_PROXY")) == NULL)
            return;

    if(!STRLEN(http_proxy))
        return;

    if((p = strstr(http_proxy, "://")) != NULL)
        p += 3;
    else
        p = http_proxy;

    http_proxy = p;
    host = http_proxy;

    if((p = strchr(http_proxy, ':')) != NULL) {
        port = ATOI(p + 1);

        host = http_proxy;

        while(*host++)
            if(*host == ':')
                *host = 0;

        host = http_proxy;
    }

    if(!STRLEN(host) || port < _TCP_PORT_MIN || port > _TCP_PORT_MAX)
        return;

    g_free(P_Temp_opt.proxy_host);

    P_Temp_opt.proxy_host = g_strdup(host);
    P_Temp_opt.proxy_port = port;
}

void mdh_toolbar_panel_temp_setup(struct panel_t *obj, const gchar *station)
{
    memset(obj, '\0', sizeof(struct panel_t));

    obj->init        = FALSE;
    obj->name        = "temp";
    obj->value       = (station) ? g_strdup(station) : g_strdup("KGRB");
    obj->command     = g_strdup("mozilla http://www.weather.com/");;
    obj->tooltip     = g_strdup_printf("Temp: %s", obj->value);
    obj->interval    = _WEATHER_DEF;
    obj->enabled     = FALSE;
    obj->bold        = FALSE;
    obj->callback    = (GtkFunction) mdh_toolbar_panel_command_execute;
    obj->function    = (GtkFunction) mdh_toolbar_panel_temp;
    obj->init_open   = (GtkFunction) mdh_toolbar_panel_temp_init;
    obj->init_close  = (GtkFunction) mdh_toolbar_panel_temp_close;

    obj->data        = g_new(struct p_temp_t, 1);
}
