/*
 * mdh (MailDooHicky)
 *
 * Copyright (c) 2003 Mike Hokenson <mdh at gozer dot org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef __linux__
#include <sys/sysinfo.h>
#endif

#ifdef BSD4_4
#include <sys/sysctl.h>
#endif

#ifdef __sun__
#include <kstat.h>
#endif

#include <gtk/gtk.h>

#include "mdh.h"
#include "util.h"

#include "mdh_gen.h"

#include "mdh_toolbar.h"
#include "mdh_panel_uptime.h"

#if _MDH_HAS_UPTIME

#ifdef __sun__
extern kstat_ctl_t *kc;
#endif

static const gchar *mdh_toolbar_panel_uptime_get_data(struct panel_t *obj);

static gboolean mdh_toolbar_panel_uptime_get_real(struct panel_t *obj,
                                                  gulong *seconds);

static GtkFunction mdh_toolbar_panel_uptime_init(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_uptime_init()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) FALSE); */

#ifdef __sun__ /* FIXME */
    if(P_Net.init == TRUE || P_Uptime.init == TRUE)
        return((GtkFunction) TRUE);
#else
    if(obj->init == TRUE)
        return((GtkFunction) TRUE);
#endif

#ifdef __sun__
    if((kc = kstat_open()) == NULL) {
        PERROR(NULL, 0, "mdh_toolbar_panel_uptime_init: kstat_open");
        return((GtkFunction) FALSE);
    }
#endif

#ifdef __sun__ /* FIXME */
    P_Net.init = P_Uptime.init = TRUE;
#else
    obj->init  = TRUE;
#endif

    return((GtkFunction) TRUE);
}

static GtkFunction mdh_toolbar_panel_uptime_close(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_uptime_close()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

#ifdef __sun__ /* FIXME */
    if(P_Net.init == FALSE || P_Uptime.init == FALSE)
        return((GtkFunction) FALSE);
#else
    if(obj->init == FALSE)
        return((GtkFunction) FALSE);
#endif

#ifdef __sun__
    if(kstat_close(kc) == -1)
        PERROR(NULL, 0, "mdh_toolbar_panel_uptime_close: kstat_close");
#endif

#ifdef __sun__ /* FIXME */
    P_Net.init = P_Uptime.init = FALSE;
#else
    obj->init  = FALSE;
#endif

    return((GtkFunction) FALSE);
}

static GtkFunction mdh_toolbar_panel_uptime(struct panel_t *obj)
{
    const gchar *text;

    DEBUG(obj, "mdh_toolbar_panel_uptime()");

    g_return_val_if_fail(obj != NULL, (GtkFunction) FALSE);

    /* if(obj->enabled == FALSE)
        return((GtkFunction) obj->init); */

    if((text = mdh_toolbar_panel_uptime_get_data(obj)) != NULL)
        mdh_label_set_text(GTK_LABEL(obj->widget), text, FALSE);

    return((GtkFunction) obj->init);
}

static const gchar *mdh_toolbar_panel_uptime_get_data(struct panel_t *obj)
{
    gulong seconds;
    guint hours, minutes, days;

    static gchar text[_DATA_MAX];

    DEBUG(obj, "mdh_toolbar_panel_uptime_get_data()");

    g_return_val_if_fail(obj != NULL, NULL);

    if(obj->init == FALSE) {
        if(obj->init_open(obj) == FALSE) {
            mdh_toolbar_panel_stop(obj);
            return("N/A");
        }
    }

    if(mdh_toolbar_panel_uptime_get_real(obj, &seconds) == FALSE) {
        mdh_toolbar_panel_stop(obj);
        return("N/A");
    }

    minutes  = seconds / 60;
    hours    = minutes / 60;
    minutes %= 60;
    days     = hours / 24;
    hours   %= 24;

    g_snprintf(text, sizeof(text), "%dd %2d:%02d", days, hours, minutes);

    return(text);
}

static gboolean mdh_toolbar_panel_uptime_get_real(struct panel_t *obj,
                                                  gulong *seconds)
{
    gulong sec;

#ifdef __linux__
    struct sysinfo sysin;
#endif

#ifdef __sun__
    kstat_t *ksp;
    kstat_named_t *knp;
#endif

#ifdef BSD4_4
    static gint mib[] = { CTL_KERN, KERN_BOOTTIME };
    struct timeval btime;
    size_t size = sizeof(btime);
#endif

    DEBUG(obj, "mdh_toolbar_panel_uptime_get_real()");

    g_return_val_if_fail(obj != NULL, FALSE);

#ifdef __linux__
    if(sysinfo(&sysin) == -1) {
        PERROR(NULL, 0, "mdh_toolbar_panel_uptime_get_real: sysinfo");
        return(FALSE);
    }

    sec = sysin.uptime;
#endif

#ifdef __sun__
    if(kstat_chain_update(kc) == -1) {
        PERROR(NULL, 0, "mdh_toolbar_panel_uptime_get_real: kstat_chain_update");
        return(FALSE);
    }

    ksp = kstat_lookup(kc, "unix", -1, "system_misc");

    if(ksp != NULL && kstat_read(kc, ksp, NULL) >= 0)
        if((knp = kstat_data_lookup(ksp, "boot_time")) != NULL)
            sec = time(NULL) - knp->value.ui32 + 30;
#endif

#ifdef BSD4_4
    if(sysctl(mib, 2, &btime, &size, NULL, 0) != -1 && btime.tv_sec != 0)
        sec = (time(NULL) - btime.tv_sec) + 30;
#endif

    *seconds = sec;

    return(TRUE);
}

void mdh_toolbar_panel_uptime_setup(struct panel_t *obj)
{
    memset(obj, '\0', sizeof(struct panel_t));

    obj->init        = FALSE;
    obj->name        = "uptime";
    obj->command     = g_strdup("");
    obj->tooltip     = g_strdup("Uptime");
    obj->interval    = 60;
    obj->enabled     = TRUE;
    obj->bold        = FALSE;
    obj->callback    = (GtkFunction) mdh_toolbar_panel_command_execute;
    obj->function    = (GtkFunction) mdh_toolbar_panel_uptime;
    obj->init_open   = (GtkFunction) mdh_toolbar_panel_uptime_init;
    obj->init_close  = (GtkFunction) mdh_toolbar_panel_uptime_close;
}

#endif /* _MDH_HAS_UPTIME */
