/*
 * mdh (MailDooHicky)
 *
 * Copyright (c) 2003 Mike Hokenson <mdh at gozer dot org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#include <gtk/gtk.h>

#include "mdh.h"
#include "util.h"
#include "version.h"

#include "mdh_cfg.h"
#include "mdh_gen.h"
#include "mdh_err.h"
#include "mdh_pad.h"
#include "mdh_run.h"

#include "mdh_toolbar.h"
#include "mdh_custom.h"
#include "mdh_panel_time.h"

/* from confuse.c */
extern gchar confuse_version[];
extern gchar confuse_author[];

static gint mdh_toolbar_panel_event(GtkWidget *widget,
                                    GdkEvent *event,
                                    struct panel_t *obj);

static gint mdh_toolbar_custom_event(GtkWidget *widget,
                                     GdkEvent *event,
                                     struct custom_t *obj);

static gint mdh_toolbar_button_event(GtkWidget *widget,
                                     GdkEvent *event,
                                     struct button_t *obj);

static void mdh_toolbar_panel_toggle(struct panel_t *obj);
static void mdh_toolbar_panel_show(struct panel_t *obj);
static void mdh_toolbar_panel_hide(struct panel_t *obj);
static void mdh_toolbar_panel_reset(struct panel_t *obj);

static void mdh_toolbar_button_toggle(struct button_t *obj);
static void mdh_toolbar_button_show(struct button_t *obj);
static void mdh_toolbar_button_hide(struct button_t *obj);
static void mdh_toolbar_button_reset(struct button_t *obj);

static void mdh_toolbar_custom_toggle(struct custom_t *obj);
static void mdh_toolbar_custom_show(struct custom_t *obj);
static void mdh_toolbar_custom_hide(struct custom_t *obj);
static void mdh_toolbar_custom_reset(struct custom_t *obj);

static void mdh_toolbar_panel_timeout_add(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_timeout_add()");

    g_return_if_fail(obj != NULL);

    if(obj->handler > 0)
        return;

    /* so screwed up timers don't cause mdh to go crazy */
    if(obj->interval < 1) {
        DEBUG(obj, "mdh_toolbar_panel_timeout_add(): warning, interval < 1");
        obj->interval = 3;
    }

    obj->handler = gtk_timeout_add(obj->interval * 1000, obj->function, obj);
}

static void mdh_toolbar_panel_timeout_del(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_timeout_del()");

    g_return_if_fail(obj != NULL);

    if(obj->handler == 0)
        return;

    gtk_timeout_remove(obj->handler);

    obj->handler = 0;
}

void mdh_toolbar_button_add(GtkWidget *box, struct button_t *obj)
{
    mdh_toolbar_button_new(obj);
    gtk_box_pack_start(GTK_BOX(box), obj->box, FALSE, FALSE, 0);
    gtk_widget_show(obj->box);
}

void mdh_toolbar_custom_add(GtkWidget *box, struct custom_t *obj)
{
    mdh_toolbar_custom_new(obj);
    gtk_box_pack_start(GTK_BOX(box), obj->box, FALSE, FALSE, 0);
    gtk_widget_show(obj->box);
}

void mdh_toolbar_panel_add(GtkWidget *box, struct panel_t *obj,
                           gint type, gint id)
{
    g_return_if_fail(obj != NULL);

    mdh_toolbar_panel_new(obj);
    gtk_box_pack_start(GTK_BOX(box), obj->event_box, FALSE, FALSE, 0);
    gtk_widget_show(obj->event_box);

    obj->type = type;
    obj->id   = id;
}

/* make sure one button or panel is displayed. if none, show menu button */
void mdh_toolbar_check_all(void)
{
    GSList *p;

    DEBUG(NULL, "mdh_toolbar_check_all()");

    for(p = Toolbar.buttons; p != NULL; p = g_slist_next(p))
        if(BUTTON(p->data)->hidden == FALSE)
            return;

    for(p = Toolbar.custom; p != NULL; p = g_slist_next(p))
        if(CUSTOM(p->data)->hidden == FALSE)
            return;

    for(p = Toolbar.panels; p != NULL; p = g_slist_next(p))
        if(PANEL(p->data)->hidden == FALSE)
            return;

    DEBUG(NULL, "mdh_toolbar_check_all(): all items disabled, enabling menu");

    mdh_toolbar_button_show(&B_Menu);
}

void mdh_toolbar_panel_start(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_start()");

    g_return_if_fail(obj != NULL);

    /* if(obj->enabled == FALSE)
        return; */

    mdh_toolbar_panel_timeout_add(obj);

    obj->function(obj);

    /* if(obj->init == FALSE && obj->init_open)
        obj->init_open(obj); */
}

/* stop the timer and close the resource(s) relating to the panel */
void mdh_toolbar_panel_stop(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_stop()");

    g_return_if_fail(obj != NULL);

    /* if(obj->enabled == FALSE)
        return; */

    mdh_toolbar_panel_timeout_del(obj);

    if(obj->init == TRUE && obj->init_close)
        obj->init_close(obj);
}

void mdh_toolbar_panel_new(struct panel_t *obj)
{
    GtkWidget *frame;

    g_return_if_fail(obj != NULL);

    obj->box = gtk_hbox_new(FALSE, 0);

    obj->event_box = gtk_event_box_new();
    gtk_container_add(GTK_CONTAINER(obj->event_box), obj->box);
    gtk_widget_show(obj->event_box);

    mdh_tooltips_set_tip(Toolbar.tips, obj->event_box, obj->tooltip);

    if(obj->callback != NULL)
        g_signal_connect(G_OBJECT(obj->event_box), "button_release_event",
            G_CALLBACK(mdh_toolbar_panel_event), obj);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_widget_show(frame);

    obj->widget = gtk_label_new(obj->value);

    gtk_misc_set_alignment(GTK_MISC(obj->widget), 0.5, 0.5);
    gtk_misc_set_padding(GTK_MISC(obj->widget), 4, 3);

    gtk_container_add(GTK_CONTAINER(frame), obj->widget);

    /* mdh_widget_set_font(obj->widget, Toolbar.font); */

    gtk_box_pack_start(GTK_BOX(obj->box), frame, FALSE, FALSE, 0);

    obj->handler = 0;

    obj->hidden = TRUE;

    Toolbar.panels = g_slist_append(Toolbar.panels, obj);
}

void mdh_toolbar_custom_new(struct custom_t *obj)
{
    g_return_if_fail(obj != NULL);

    obj->box = gtk_hbox_new(FALSE, 0);

    obj->widget = gtk_button_new_with_label(obj->value);

    gtk_widget_can_focus(GTK_WIDGET(obj->widget), FALSE);

    /* mdh_widget_set_font(obj->widget, Toolbar.font); */

    gtk_box_pack_start(GTK_BOX(obj->box), obj->widget, FALSE, FALSE, 0);

    mdh_tooltips_set_tip(Toolbar.tips, obj->widget, obj->tooltip);

    g_signal_connect(G_OBJECT(obj->widget), "button_release_event",
        G_CALLBACK(mdh_toolbar_custom_event), obj);

    obj->hidden = TRUE;

    Toolbar.custom = g_slist_append(Toolbar.custom, obj);
}

void mdh_toolbar_button_new(struct button_t *obj)
{
    g_return_if_fail(obj != NULL);

    obj->box = gtk_hbox_new(FALSE, 0);

    obj->widget = gtk_button_new_with_label(obj->name);
    /* gtk_button_set_relief(GTK_BUTTON(obj->widget), GTK_RELIEF_NONE); */
    gtk_widget_can_focus(GTK_WIDGET(obj->widget), FALSE);

    /* mdh_widget_set_font(obj->widget, Toolbar.font); */

    gtk_box_pack_start(GTK_BOX(obj->box), obj->widget, FALSE, FALSE, 0);

    mdh_tooltips_set_tip(Toolbar.tips, obj->widget, obj->tooltip);

    g_signal_connect(G_OBJECT(obj->widget), "button_release_event",
        G_CALLBACK(mdh_toolbar_button_event), obj);

    obj->hidden = TRUE;

    Toolbar.buttons = g_slist_append(Toolbar.buttons, obj);
}

static void mdh_toolbar_panel_toggle(struct panel_t *obj)
{
    g_return_if_fail(obj != NULL);

    if(obj->enabled == FALSE)
        return;

    if(obj->hidden == TRUE)
        mdh_toolbar_panel_show(obj);
    else
        mdh_toolbar_panel_hide(obj);
}

static void mdh_toolbar_panel_show(struct panel_t *obj)
{
    g_return_if_fail(obj != NULL);

    gtk_widget_show_all(obj->box);

    obj->hidden = FALSE;
}

static void mdh_toolbar_panel_hide(struct panel_t *obj)
{
    g_return_if_fail(obj != NULL);

    gtk_widget_hide_all(obj->box);

    obj->hidden = TRUE;
}

static void mdh_toolbar_panel_reset(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_reset_panel()");

    g_return_if_fail(obj != NULL);

    mdh_widget_set_font(obj->widget, Toolbar.font);

    mdh_tooltips_set_tip(Toolbar.tips, obj->event_box, obj->tooltip);

    if(obj->enabled == TRUE) {
        mdh_toolbar_panel_start(obj);
        mdh_toolbar_panel_show(obj);
    } else {
        mdh_toolbar_panel_stop(obj);
        mdh_toolbar_panel_hide(obj);
    }
}

static void mdh_toolbar_button_toggle(struct button_t *obj)
{
    g_return_if_fail(obj != NULL);

    if(obj->enabled == FALSE || obj->can_toggle == FALSE)
        return;

    if(obj->hidden == TRUE)
        mdh_toolbar_button_show(obj);
    else
        mdh_toolbar_button_hide(obj);
}

static void mdh_toolbar_button_show(struct button_t *obj)
{
    g_return_if_fail(obj != NULL);

    gtk_widget_show_all(obj->box);

    obj->hidden = FALSE;
}

static void mdh_toolbar_button_hide(struct button_t *obj)
{
    g_return_if_fail(obj != NULL);

    gtk_widget_hide_all(obj->box);

    obj->hidden = TRUE;
}

static void mdh_toolbar_button_reset(struct button_t *obj)
{
    g_return_if_fail(obj != NULL);

    mdh_widget_set_font(obj->widget, Toolbar.font);

    if(obj->enabled == TRUE)
        mdh_toolbar_button_show(obj);
    else
        mdh_toolbar_button_hide(obj);
}

static void mdh_toolbar_custom_toggle(struct custom_t *obj)
{
    g_return_if_fail(obj != NULL);

    if(obj->enabled == FALSE)
        return;

    if(obj->hidden == TRUE)
        mdh_toolbar_custom_show(obj);
    else
        mdh_toolbar_custom_hide(obj);
}

static void mdh_toolbar_custom_show(struct custom_t *obj)
{
    g_return_if_fail(obj != NULL);

    gtk_widget_show_all(obj->box);

    obj->hidden = FALSE;
}

static void mdh_toolbar_custom_hide(struct custom_t *obj)
{
    g_return_if_fail(obj != NULL);

    gtk_widget_hide_all(obj->box);

    obj->hidden = TRUE;
}

static void mdh_toolbar_custom_reset(struct custom_t *obj)
{
    g_return_if_fail(obj != NULL);

    gtk_button_set_label(GTK_BUTTON(obj->widget), obj->value);

    mdh_tooltips_set_tip(Toolbar.tips, obj->widget, obj->tooltip);

    mdh_widget_set_font(obj->widget, Toolbar.font);

    if(obj->enabled == TRUE)
        mdh_toolbar_custom_show(obj);
    else
        mdh_toolbar_custom_hide(obj);
}

void mdh_toolbar_toggle_all(void)
{
    DEBUG(NULL, "mdh_toolbar_toggle_all()");

    /* hiding the panel while the config window is open can cause problems */
    if(Toolbar.config_open == TRUE || Toolbar.custom_open == TRUE)
        return;

    if(Toolbar.hidden == TRUE) {
        gtk_button_set_label(GTK_BUTTON(B_Hide.widget), "<");
        mdh_widget_set_font(B_Hide.widget, Toolbar.font);

        Toolbar.hidden = FALSE;
    } else {
        gtk_button_set_label(GTK_BUTTON(B_Hide.widget), ">");
        mdh_widget_set_font(B_Hide.widget, Toolbar.font);

        Toolbar.hidden = TRUE;
    }

    g_slist_foreach(Toolbar.buttons, (GFunc) mdh_toolbar_button_toggle, NULL);
    g_slist_foreach(Toolbar.custom,  (GFunc) mdh_toolbar_custom_toggle, NULL);
    g_slist_foreach(Toolbar.panels,  (GFunc) mdh_toolbar_panel_toggle, NULL);
}

void mdh_toolbar_panel_refresh(struct panel_t *obj)
{
    DEBUG(obj, "mdh_toolbar_panel_refresh()");

    g_return_if_fail(obj != NULL);

    if(obj->enabled == FALSE)
        return;

    if(obj->handler == 0 && obj->init_close) {
        mdh_toolbar_panel_stop(obj);
        mdh_toolbar_panel_start(obj);
    } else if(obj->interval >= 5) {
        obj->function(obj);
    }
}

void mdh_toolbar_panel_refresh_all(void)
{
    DEBUG(NULL, "mdh_toolbar_panel_refresh_all()");

    g_slist_foreach(Toolbar.panels, (GFunc) mdh_toolbar_panel_refresh, NULL);
}

#define _MDH_TOOLBAR_EVENT_NONE     0
#define _MDH_TOOLBAR_EVENT_CALLBACK 1
#define _MDH_TOOLBAR_EVENT_MENU     2

static gint mdh_toolbar_event_common(GdkEvent *event)
{
    g_return_val_if_fail(event != NULL, 0);

    switch(event->type) {
        case GDK_BUTTON_RELEASE:
        {
            GdkEventButton *bevent = (GdkEventButton *) event; 

            switch(bevent->button) {
                case 1:
                case 2:
                    return(_MDH_TOOLBAR_EVENT_CALLBACK);
                case 3:
                    return(_MDH_TOOLBAR_EVENT_MENU);
                default:
                    return(_MDH_TOOLBAR_EVENT_NONE);
            }

            break;
        }

        default:
            break;
    }

    return(_MDH_TOOLBAR_EVENT_NONE);
}

static gint mdh_toolbar_panel_event(GtkWidget *widget,
                                    GdkEvent *event,
                                    struct panel_t *obj)
{
    g_return_val_if_fail(obj != NULL, 0);

    switch(mdh_toolbar_event_common(event)) {
        case _MDH_TOOLBAR_EVENT_CALLBACK:
            obj->callback(obj);
            break;
        case _MDH_TOOLBAR_EVENT_MENU:
            mdh_toolbar_menu_show();
            break;
        default:
            break;
    }

    return(0);
}

static gint mdh_toolbar_custom_event(GtkWidget *widget,
                                     GdkEvent *event,
                                     struct custom_t *obj)
{
    g_return_val_if_fail(obj != NULL, 0);

    switch(mdh_toolbar_event_common(event)) {
        case _MDH_TOOLBAR_EVENT_CALLBACK:
            obj->callback(obj);
            break;
        case _MDH_TOOLBAR_EVENT_MENU:
            mdh_toolbar_menu_show();
            break;
        default:
            break;
    }

    return(0);
}

static gint mdh_toolbar_button_event(GtkWidget *widget,
                                     GdkEvent *event,
                                     struct button_t *obj)
{
    g_return_val_if_fail(obj != NULL, 0);

    switch(mdh_toolbar_event_common(event)) {
        case _MDH_TOOLBAR_EVENT_CALLBACK:
            obj->callback(obj);
            break;
        case _MDH_TOOLBAR_EVENT_MENU:
            mdh_toolbar_menu_show();
            break;
        default:
            break;
    }

    return(0);
}

static void mdh_toolbar_calendar(void)
{
    GtkWidget *window;
    GtkWidget *vbox;
    GtkWidget *calendar;
    GtkWidget *button;

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), "MailDooHicky: Calendar");
    gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);

    g_signal_connect(G_OBJECT(window), "event",
        G_CALLBACK(mdh_window_event), NULL);

    gtk_container_set_border_width(GTK_CONTAINER(window), 5);

    vbox = gtk_vbox_new(FALSE, 5);
    gtk_widget_show(vbox);

    gtk_container_add(GTK_CONTAINER(window), vbox);

    calendar = gtk_calendar_new();
    gtk_widget_show(calendar);

    gtk_box_pack_start(GTK_BOX(vbox), calendar, FALSE, FALSE, 0);

    button = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
    gtk_box_pack_start(GTK_BOX(vbox), button, TRUE, TRUE, 0);
    gtk_widget_show(button);

    mdh_widget_set_default(GTK_WINDOW(window), button);

    g_signal_connect_swapped(G_OBJECT(button), "clicked",
        G_CALLBACK(gtk_widget_destroy), G_OBJECT(window));

    gtk_widget_show(window);
}

static void mdh_toolbar_quit(void)
{
    if(mdh_window_yes_no_prompt("Are sure you want to quit?") == TRUE)
        gtk_main_quit();
}

static void mdh_toolbar_about(void)
{
    GtkWidget *window;
    GtkWidget *vbox;
    GtkWidget *vbox2;
    GtkWidget *hbox;
    GtkWidget *image;
    GtkWidget *label;
    GtkWidget *hsep;
    GtkWidget *button;

    gchar *text;

    /*
     * http://www.geocrawler.com/mail/msg.php3?msg_id=6021322&list=522
     *   char buf[3] = { 0xc2, 0xa9, '\0' };
     *   char buf[7]; int len = g_unichar_to_utf8 (176, buf); buf[len] = '\0';
     */

    gchar COPYRIGHT[3] = { 0xc2, 0xa9, 0 };
    gchar TRADEMARK[4] = { 0xe2, 0x84, 0xa2, 0 };

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), "MailDooHicky: About");
    gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, FALSE);

    g_signal_connect(G_OBJECT(window), "event",
        G_CALLBACK(mdh_window_event), NULL);

    gtk_container_set_border_width(GTK_CONTAINER(window), 5);

    vbox = gtk_vbox_new(FALSE, 2);
    gtk_widget_show(vbox);

    gtk_container_add(GTK_CONTAINER(window), vbox);
    
    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);
    
    image = gtk_image_new_from_stock(GTK_STOCK_DIALOG_INFO,
        GTK_ICON_SIZE_DIALOG);
    gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
    gtk_widget_show(image);
    
    gtk_misc_set_alignment(GTK_MISC(image), 0.5, 0.0);
    gtk_misc_set_padding(GTK_MISC(image), 0, 5);
    
    vbox2 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox2);

    gtk_container_set_border_width(GTK_CONTAINER(vbox2), 5);
    
    gtk_container_add(GTK_CONTAINER(hbox), vbox2);
    
    label = gtk_label_new(NULL);
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox2), label, FALSE, FALSE, 5);
    gtk_widget_show(label);

    text = g_strdup_printf("MailDooHicky v%s", mdh_toolbar_version());

    mdh_label_set_larger(GTK_LABEL(label), text);

    g_free(text);
    
    label = gtk_label_new(_MDH_URL);
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox2), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    label = mdh_label_new("%s 2003 Mike Hokenson <%s>", COPYRIGHT, _MDH_EMAIL);
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox2), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    hsep = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox2), hsep, FALSE, TRUE, 10);
    gtk_widget_show(hsep);

    label = mdh_label_new("libConfuse v%s", confuse_version);
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox2), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    label = gtk_label_new("http://savannah.nongnu.org/projects/confuse");
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox2), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    label = mdh_label_new("%s 2002-2003 %s", COPYRIGHT, confuse_author);
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox2), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    hsep = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox2), hsep, FALSE, TRUE, 10);
    gtk_widget_show(hsep);

    label = gtk_label_new(NULL);
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox2), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    text = g_strdup_printf("\"We got you covered like a jimmy hat.\" %s",
        TRADEMARK);

    mdh_label_set_smaller(GTK_LABEL(label), text);

    g_free(text);

    button = gtk_button_new_from_stock(GTK_STOCK_OK);
    gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
    gtk_widget_show(button);

    mdh_widget_set_default(GTK_WINDOW(window), button);
    
    g_signal_connect_swapped(G_OBJECT(button), "clicked",
        G_CALLBACK(gtk_widget_destroy), G_OBJECT(window));
                 
    gtk_widget_show(window);
}   

static GtkWidget *mdh_toolbar_menu_create(void)
{
    GtkWidget *menu;
    GtkWidget *menu_sub;
    GtkWidget *menu_item;

    DEBUG(NULL, "mdh_toolbar_menu_create()");

    menu = gtk_menu_new();
    gtk_widget_show(menu);

    menu_sub = gtk_menu_new();
    gtk_widget_show(menu_sub);

    menu_item = gtk_menu_item_new_with_label("MailDooHicky");
    gtk_menu_append(GTK_MENU(menu), menu_item);
    gtk_widget_show(menu_item);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_item), menu_sub);

    menu_item = gtk_menu_item_new();
    gtk_menu_append(GTK_MENU(menu), menu_item);
    gtk_widget_show(menu_item);

        menu_item = gtk_menu_item_new_with_label("About");
        gtk_menu_append(GTK_MENU(menu_sub), menu_item);
        gtk_widget_show(menu_item);

        g_signal_connect(G_OBJECT(menu_item), "activate",
            G_CALLBACK(mdh_toolbar_about), NULL);

        menu_item = gtk_menu_item_new();
        gtk_menu_append(GTK_MENU(menu_sub), menu_item);
        gtk_widget_show(menu_item);

        menu_item = gtk_menu_item_new_with_label("Configure");
        gtk_menu_append(GTK_MENU(menu_sub), menu_item);
        gtk_widget_show(menu_item);

        if(Toolbar.config_open == TRUE)
            gtk_widget_set_sensitive(GTK_WIDGET(menu_item), FALSE);

        g_signal_connect(G_OBJECT(menu_item), "activate",
            G_CALLBACK(mdh_window_configuration), NULL);

        menu_item = gtk_menu_item_new_with_label("Custom Buttons");
        gtk_menu_append(GTK_MENU(menu_sub), menu_item);
        gtk_widget_show(menu_item);

        if(Toolbar.custom_open == TRUE)
            gtk_widget_set_sensitive(GTK_WIDGET(menu_item), FALSE);

        g_signal_connect(G_OBJECT(menu_item), "activate",
            G_CALLBACK(mdh_toolbar_custom), NULL);

        menu_item = gtk_menu_item_new_with_label("Toggle Toolbar");
        gtk_menu_append(GTK_MENU(menu_sub), menu_item);
        gtk_widget_show(menu_item);

        /* Menu and Hide buttons are the only two visible when toggled */
        if(B_Menu.enabled == FALSE && B_Hide.enabled == FALSE)
            gtk_widget_set_sensitive(GTK_WIDGET(menu_item), FALSE);

        g_signal_connect(G_OBJECT(menu_item), "activate",
            G_CALLBACK(mdh_toolbar_toggle_all), NULL);

        menu_item = gtk_menu_item_new();
        gtk_menu_append(GTK_MENU(menu_sub), menu_item);
        gtk_widget_show(menu_item);

        menu_item = gtk_menu_item_new_with_label("Error Log");
        gtk_menu_append(GTK_MENU(menu_sub), menu_item);
        gtk_widget_show(menu_item);

        if(Toolbar.error_open == TRUE)
            gtk_widget_set_sensitive(GTK_WIDGET(menu_item), FALSE);

        g_signal_connect(G_OBJECT(menu_item), "activate",
            G_CALLBACK(mdh_toolbar_error_log), NULL);

    menu_sub = gtk_menu_new();
    gtk_widget_show(menu_sub);

    menu_item = gtk_menu_item_new_with_label("Custom Buttons");
    gtk_menu_append(GTK_MENU(menu), menu_item);
    gtk_widget_show(menu_item);

    gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_item), menu_sub);

    {
        gint i;

        for(i = 0; i <= _MDH_MAX_CUST; i++) {
            struct custom_t *obj = &C_User[i];

            gchar *title = (STRLEN(obj->tooltip)) ? obj->tooltip : obj->value;

            menu_item = gtk_menu_item_new_with_label(title);
            gtk_menu_append(GTK_MENU(menu_sub), menu_item);
            gtk_widget_show(menu_item);

            g_signal_connect_swapped(G_OBJECT(menu_item), "activate",
                G_CALLBACK(obj->callback), obj);
        }
    }

    menu_item = gtk_menu_item_new();
    gtk_menu_append(GTK_MENU(menu), menu_item);
    gtk_widget_show(menu_item);

    menu_item = gtk_menu_item_new_with_label("Calendar");
    gtk_menu_append(GTK_MENU(menu), menu_item);
    gtk_widget_show(menu_item);

    g_signal_connect(G_OBJECT(menu_item), "activate",
        G_CALLBACK(mdh_toolbar_calendar), NULL);

    menu_item = gtk_menu_item_new_with_label("Scratch-Pad");
    gtk_menu_append(GTK_MENU(menu), menu_item);
    gtk_widget_show(menu_item);

    if(Toolbar.pad_open == TRUE)
        gtk_widget_set_sensitive(GTK_WIDGET(menu_item), FALSE);

    g_signal_connect(G_OBJECT(menu_item), "activate",
        G_CALLBACK(mdh_toolbar_scratch_pad), NULL);

    menu_item = gtk_menu_item_new_with_label("Run...");
    gtk_menu_append(GTK_MENU(menu), menu_item);
    gtk_widget_show(menu_item);

    if(Toolbar.run_open == TRUE)
        gtk_widget_set_sensitive(GTK_WIDGET(menu_item), FALSE);

    g_signal_connect(G_OBJECT(menu_item), "activate",
        G_CALLBACK(mdh_toolbar_run), NULL);

    menu_item = gtk_menu_item_new();
    gtk_menu_append(GTK_MENU(menu), menu_item);
    gtk_widget_show(menu_item);

    menu_item = gtk_menu_item_new_with_label("Quit");
    gtk_menu_append(GTK_MENU(menu), menu_item);
    gtk_widget_show(menu_item);

    g_signal_connect(G_OBJECT(menu_item), "activate",
        G_CALLBACK(mdh_toolbar_quit), NULL);

    return(menu);
}

#define MDH_TOOLBAR_IS_PANEL_MENU(menu) (menu != NULL && GTK_IS_MENU(menu))

void mdh_toolbar_menu_destroy(void)
{
    DEBUG(NULL, "mdh_toolbar_menu_destroy()");

    if(!MDH_TOOLBAR_IS_PANEL_MENU(Toolbar.menu))
        return;

    gtk_widget_unparent(GTK_WIDGET(Toolbar.menu));
    gtk_widget_destroy(GTK_WIDGET(Toolbar.menu));
    gtk_object_sink(GTK_OBJECT(Toolbar.menu));

    Toolbar.menu = NULL;
}

GtkFunction mdh_toolbar_menu_show(void)
{
    if(!MDH_TOOLBAR_IS_PANEL_MENU(Toolbar.menu))
        Toolbar.menu = mdh_toolbar_menu_create();

    gtk_menu_popup(GTK_MENU(Toolbar.menu), NULL, NULL, NULL, NULL, 0, 0);

    return((GtkFunction) FALSE);
}

void mdh_toolbar_panel_command_execute(struct panel_t *obj)
{
    g_return_if_fail(obj != NULL);

    if(!STRLEN(obj->command))
        return;

    mdh_toolbar_exec(obj->command);
}

void mdh_toolbar_custom_command_execute(struct custom_t *obj)
{
    g_return_if_fail(obj != NULL);

    mdh_toolbar_exec(obj->command);
}

/* show/hide widgets, setup timers */
void mdh_toolbar_reset(void)
{
    DEBUG(NULL, "mdh_toolbar_reset()");

    mdh_toolbar_set_location();

    mdh_toolbar_set_layer_on_top(Toolbar.on_top);

    if(Toolbar.tooltips == TRUE)
        gtk_tooltips_enable(Toolbar.tips);
    else
        gtk_tooltips_disable(Toolbar.tips);

    if(Toolbar.hidden == TRUE)
        mdh_toolbar_toggle_all();

    g_slist_foreach(Toolbar.buttons, (GFunc) mdh_toolbar_button_reset, NULL);
    g_slist_foreach(Toolbar.custom,  (GFunc) mdh_toolbar_custom_reset, NULL);
    g_slist_foreach(Toolbar.panels,  (GFunc) mdh_toolbar_panel_reset, NULL);

    /* make sure at least one panel is shown, otherwise display menu button */
    mdh_toolbar_check_all();

    mdh_toolbar_menu_destroy();
}

/* just custom buttons; for OK/Apply in custom button editor */
void mdh_toolbar_reset_custom_buttons(void)
{
    DEBUG(NULL, "mdh_toolbar_reset_custom_buttons()");

    /* if(Toolbar.hidden == TRUE)
        mdh_toolbar_toggle_all(); */

    g_slist_foreach(Toolbar.custom, (GFunc) mdh_toolbar_custom_reset, NULL);

    /* make sure at least one panel is shown, otherwise display menu button */
    mdh_toolbar_check_all();
}

#define _ERR_LOG_MAX 128  /* maximum number of entries in the error log */

/* adds an entry to the error log */
void mdh_toolbar_add_error_log_entry(struct panel_t *obj, gint err,
                                     const gchar *fmt, ...)
{
    gchar buf[LINE_MAX];

    va_list args;

    if(obj != NULL) {
        if(obj->err == err)
            return;

        obj->err = err;
    }

    /* would be better to remove the first entries, but this is easier :I */
    if(g_slist_length(Toolbar.errors) > _ERR_LOG_MAX) {
        g_slist_foreach(Toolbar.errors, (GFunc) g_free, NULL);
        g_slist_free(Toolbar.errors);
        Toolbar.errors = NULL;

        ADD_LOG(NULL, 0, "MailDooHicky error log cleared");
    }

    va_start(args, fmt);

    g_vsnprintf(buf, sizeof(buf), fmt, args);

    va_end(args);

    Toolbar.errors = g_slist_append(Toolbar.errors,
        g_strdup_printf("%s - %s", mdh_strftime("%m/%d - %l:%M %p"), buf));
}

const gchar *mdh_toolbar_version(void)
{
    return(version);
}

/* returns true upon successful command execution */
gboolean mdh_toolbar_exec(const gchar *command)
{
    GError *err = NULL;

    DEBUG(NULL, "mdh_toolbar_exec(): %s", command);

    g_return_val_if_fail(command != NULL, TRUE);

    if(g_spawn_command_line_async(command, &err) == FALSE) {
        ADD_LOG(NULL, 0, "mdh_toolbar_exec: %s", err->message);
        mdh_window_error("Execution failed.", err->message);
        g_error_free(err);
        return(FALSE);
    }

    return(TRUE);
}
